#ifndef QCALC_OPERATOR_ITEM_H
#define QCALC_OPERATOR_ITEM_H

#include "qcalcCommon.h"
#include "qcalcCalcItem.h"

// Virtual operator item. Has both inputs and outputs. May be a black box, ie a container for a sub-graph of VCalcItems (including other black boxes)
class VOperatorItem: public VCalcItem
{
	public:
		// Creates a 2-inputs-1-output additive item, by default. To be changed afterwards
		VOperatorItem(const QCalcOperation& op=QCALC_ADD, const QString& pluginName="");
		
		// Also removes the GUI
		virtual ~VOperatorItem();

		
		// Returns the associated GUI, if it exists, or NULL
		virtual OperatorItem* getGUI();
		
		// if there is a GUI associated, removes it from the scene, and deletes it
		void removeGUI();
		
		void setOperation(QCalcOperation operation, const QString& pluginName="");

		// Recomputes the values of the outputs
		virtual void reevaluateOutSlots();
		
		void setErrMessage(bool err=false, const QString& message="");
		
		// Makes it a black box created with the bb list of VCalcItems.
		void setAsBlackBox(const QList<VCalcItem*>& bb, const QString& newName="Black Box");

        void setUpPlugin(const QString& pluginName);

        bool isPlugin();

        void getPluginInfo(QString& io, QString& help, QString& html);

        virtual QDomElement toXMLElement(QDomDocument&) const;

        virtual void fromXMLElement(QDomElement& el, QGraphicsScene* scene);
        		QCalcOperation operation;
		static int counter;
		int operator_id;
		virtual bool hasGui();
		virtual void setSelected(bool sel=true);
		QList<int> blackBoxIDList; // Used when loading from XML
		
		bool unwrap(QGraphicsScene* scene, const QPointF& pos);
		QList<VCalcItem*> getBBList(){return blackBoxList;};
		
	protected:
		OperatorItem* gui;
		QList<VCalcItem*> blackBoxList;
		void setNumSlots(int nIn, int nOut);

		
		QEvaluator* evaluator;
        QString pluginName;
};

// GUI item associated to a VOperatorItem. Has two states: normal, and maximised
class OperatorItem: public CalcItem
{
	public:
		// The VInSlots and VOutSlots are given via the constructor. They can be changed with resetSlots.
		OperatorItem(VOperatorItem* parentItem, const QVector<VInSlot*>& inputs, const QVector<VOutSlot*>& outputs, QGraphicsItem* parent=NULL);
		
		// deletes the slots
		virtual ~OperatorItem();
		
		
		virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget=NULL);
		
		inline virtual int type() const { return Type;}
		enum { Type = TYPE_OPERATORITEM };
		
		virtual void setDisplayedString(const QString& s)
		{
			displayedString=s;
			update();
		}
		
		void expand();
		
		void collapse();
		
		inline bool isExpanded(){return expanded;}
        
		void setOperation(QCalcOperation op, const QString& pluginName="");
		VOperatorItem* parentItem;
		
		// Deletes the InSlots and OutSlots. Must be called before deleting slots from VOperatorItem
		void clearSlots();
		
		// Creates a new set of virtual slots for this operator. Must be called after deleting slots from VOperatorItem
		void resetSlots(const QVector<VInSlot*>& inputs, const QVector<VOutSlot*>& outputs);
		
		// creates the connectors for the input and output slots that are virtually connected but not graphically. This happens after creating a black box:
		// it deletes the GUI items, thus the connectors, but not the underlying connections, it creates a new operator based on the virtual slots, recreates graphic slots for these virtual slots, and thus requires to recreate the connectors.
		void resetConnectors();
		
		virtual QDomElement toXMLElement(QDomDocument& doc) const;

        virtual void fromXMLElement(QDomElement& el);
		
		bool unwrap();
		virtual int page(){return parentItem->page;}
	protected:
		void paintExpanded(QPainter* painter) const;
		void paintCollapsed(QPainter* painter) const;
		virtual QVariant itemChange(GraphicsItemChange change, const QVariant& value);
		ExpandButton* expandButton;
		QString displayedString;
		bool expanded;
		QMatrix matSave;
        
};

#endif
