#include "qcalcSlots.h"
#include "qcalcScene.h"
#include "qcalcOperatorItem.h"
#include "qcalcEvaluator.h"
#include <QGraphicsScene>
#include "qbu.h"

VInSlot::VInSlot(VCalcItem* parent, const QString& name, const QString& description)
{
	this->_parent=parent;
	_outSlot=NULL;
	gui=NULL;
	this->name=name;
	this->description=description;

}

VInSlot::~VInSlot()
{

	if(_outSlot!=NULL)
	{
		_outSlot->removeConnectionTo(this);
	}
}

VCalcItem* VInSlot::connectedItem() const
{
	if(_outSlot==NULL) return NULL;
	return _outSlot->parent();
}

InSlot* VInSlot::getGUI()
{
	return gui;
}

void VInSlot::removeGUI()
{
	if(gui!=NULL)
	{
		gui=NULL;
	}
}

void VInSlot::setName(const QString& name)
{
	this->name=name;
	resetToolTips();
}

void VInSlot::setDescription(const QString& description)
{
	this->description=description;
	resetToolTips();
}

void VInSlot::resetToolTips()
{
	if(gui!=NULL)
	{
		gui->setToolTip(name+"\n"+description);
	}
}

/////////////////////////////////////

VOutSlot::VOutSlot(VCalcItem* parent, const QString& name, const QString& description)
{
	this->_parent=parent;
	_inSlots.clear();
	gui=NULL;
	value = new DvObject();
	reevaluationRequired=true;
	valueOK=false;
	this->name=name;
	this->description=description;
	openSlot = false; // used for Black Box outputs
	
}

VOutSlot::~VOutSlot()
{

	foreach(VInSlot* in, _inSlots)
	{
		removeConnectionTo(in);
	}
}

QList<VInSlot*> VOutSlot::connectedSlots() const
{
	return _inSlots;
}

QList<VCalcItem*> VOutSlot::connectedItems() const
{
	QList<VCalcItem*> res;
	res.clear();
	VCalcItem* temp;
	foreach(VInSlot* in, _inSlots)
	{
		temp=in->parent();
		if(!res.contains(temp)) res.push_back(temp);
	}
	return res;
}

void VOutSlot::connectTo(VInSlot* slot)
{
	if(!_inSlots.contains(slot))
	{
		_inSlots.push_back(slot);
		slot->_outSlot=this;
		slot->parent()->flagOutSlotsForReevaluation();
	}

}

void VOutSlot::removeConnectionTo(VInSlot* slot)
{
	if(_inSlots.contains(slot))
	{
		_inSlots.removeAll(slot);
		slot->_outSlot=NULL;
		slot->parent()->flagOutSlotsForReevaluation();
	}
}

OutSlot* VOutSlot::getGUI()
{
	return gui;
}

void VOutSlot::removeGUI()
{

	if(gui!=NULL)
	{
		gui=NULL;
	}
}

DvObject_var VOutSlot::getValue(bool& ok)
{
    if(reevaluationRequired)
	{
		_parent->reevaluateOutSlots();
	}
	ok=(valueOK && value.is_ok());
    
    if(valueOK && value.is_nil()){
        DataBrowseWindow * browser;
        DvObject_var nilObj = new DvObject();
        browser = QbuLaunchBrowse(nilObj, NULL, "Error Browser", value, getName().toLatin1().data());
       
        browser->move(300,300);
        browser->setAttribute(Qt::WA_DeleteOnClose,true);
    }
    
	return value;
}

DvObject_var VOutSlot::getNotJoinedValue(bool& ok)
{
	DvObject_var value = _parent->getNotJoinedVal();

	ok=value.is_ok();
	return value;
}

void VOutSlot::setName(const QString& name)
{
	this->name=name;
	resetToolTips();
}

void VOutSlot::setDescription(const QString& description)
{
	this->description=description;
	resetToolTips();
}

void VOutSlot::resetToolTips()
{
	if(gui!=NULL)
	{
		gui->setToolTip(name+"\n"+description);
	}
}

InSlot::InSlot(CalcItem* parent, VInSlot* slot):
	QGraphicsEllipseItem(parent)
{
	this->slot=slot;
	
	setRect(-SLOT_SIZE/2., -SLOT_SIZE/2., SLOT_SIZE, SLOT_SIZE);
	
	setFlag(QGraphicsItem::ItemIsMovable, false);
	setFlag(QGraphicsItem::ItemIsSelectable, false);
	hasMouseFocus=false;
	setAcceptHoverEvents(true);
	connector=NULL;
	setZValue(1.1);

	slot->setGUI(this);
}

InSlot::~InSlot()
{
	slot->removeGUI();
	if(connector!=NULL) delete connector;
}

void InSlot::paint(QPainter* painter, const QStyleOptionGraphicsItem* , QWidget*)
{
	QColor color1;
	QColor color2;
	if(hasMouseFocus)
	{
		color2=((CalcItem*)parentItem())->color1;
		color1=((CalcItem*)parentItem())->outlineColor;
	}
	else
	{
		color1=((CalcItem*)parentItem())->color1;
		color2=((CalcItem*)parentItem())->outlineColor;
	}
	
	painter->setRenderHints(QPainter::Antialiasing);
	QPen pen(Qt::SolidLine);
	pen.setWidthF(1.);
	pen.setColor(color2);
	painter->setPen(pen);
	QBrush brush(color2);
	painter->setBrush(brush);
	
	painter->drawEllipse(boundingRect());
	
	brush.setColor(color1);
	pen.setColor(color1);
	painter->setPen(pen);
	painter->setBrush(brush);
	
	OperatorItem* parentOperator=qgraphicsitem_cast<OperatorItem*>(parentItem());
	if(parentOperator!=NULL)
	{
		if(parentOperator->isExpanded())
		{
			QPointF points[3] = {
				QPointF(-0.25*SLOT_SIZE, 0.433*SLOT_SIZE),
					 QPointF(-0.25*SLOT_SIZE, -0.433*SLOT_SIZE),
							 QPointF(0.5*SLOT_SIZE, 0.)
			};
			painter->drawConvexPolygon(points, 3);
		}
		else
		{
			QPointF points[3] = {
				QPointF(0.433*SLOT_SIZE, -0.25*SLOT_SIZE),
					QPointF(-0.433*SLOT_SIZE, -0.25*SLOT_SIZE),
							QPointF(0., 0.5*SLOT_SIZE)
			};
			painter->drawConvexPolygon(points, 3);
		}
	}
	else
	{
		QPointF points[3] = {
			QPointF(0.433*SLOT_SIZE, -0.25*SLOT_SIZE),
				QPointF(-0.433*SLOT_SIZE, -0.25*SLOT_SIZE),
					 QPointF(0., 0.5*SLOT_SIZE)
		};
		painter->drawConvexPolygon(points, 3);
	}
	
	
}

void InSlot::hoverEnterEvent(QGraphicsSceneHoverEvent*)
{
	hasMouseFocus=true;
	((CalcItem*)parentItem())->setDisplayedString(slot->getName()+"  "+slot->getDescription());
	update();
}

void InSlot::hoverLeaveEvent(QGraphicsSceneHoverEvent*)
{
	hasMouseFocus=false;
	((CalcItem*)parentItem())->setDisplayedString("");
	update();
}

void InSlot::removeConnector(Connector* c)
{
	if(connector==c)
	{
		connector=NULL;
		OutSlot* fromSlot=c->getFromSlot(); // Cannot be NULL
		VOutSlot* from=fromSlot->slot;
		from->removeConnectionTo(slot);
	}
}

void InSlot::setConnector(Connector* c)
{
	if(connector!=NULL)
	{
		delete connector;
	}
	connector=c;
	// Establish actual connection
	OutSlot* fromSlot=c->getFromSlot(); // Cannot be NULL
	VOutSlot* from=fromSlot->slot;
	from->connectTo(slot);
}

void InSlot::updateConnectorPosition()
{
	if(connector!=NULL) connector->updatePosition();
}

//////////////////////////////////////////////////////////////

OutSlot::OutSlot(CalcItem* parent, VOutSlot* slot):
		QGraphicsEllipseItem(parent)
{
	this->slot=slot;
	
	width=10.;
	height=10.;
	setRect(-SLOT_SIZE/2., -SLOT_SIZE/2., SLOT_SIZE, SLOT_SIZE);
	
	setFlag(QGraphicsItem::ItemIsMovable, false);
	setFlag(QGraphicsItem::ItemIsSelectable, false);
	
	hasMouseFocus=false;
	setAcceptHoverEvents(true);
	connectors.clear();
	
	setZValue(1.1);
	slot->setGUI(this);
}

OutSlot::~OutSlot()
{
	slot->removeGUI();
	foreach(Connector* c, connectors)
	{
		delete c;
	}
}

void OutSlot::paint(QPainter* painter, const QStyleOptionGraphicsItem* , QWidget*)
{
	QColor color1;
	QColor color2;
	if(hasMouseFocus)
	{
		color2=((CalcItem*)parentItem())->color1;
		color1=((CalcItem*)parentItem())->outlineColor;
	}
	else
	{
		color1=((CalcItem*)parentItem())->color1;
		color2=((CalcItem*)parentItem())->outlineColor;
	}
		
	painter->setRenderHints(QPainter::Antialiasing);
	QPen pen(Qt::SolidLine);
	pen.setWidthF(1.);
	pen.setColor(color2);
	painter->setPen(pen);
	QBrush brush(color2);
	painter->setBrush(brush);
	
	painter->drawEllipse(boundingRect());
	
	brush.setColor(color1);
	pen.setColor(color1);
	painter->setPen(pen);
	painter->setBrush(brush);
	
	OperatorItem* parentOperator=qgraphicsitem_cast<OperatorItem*>(parentItem());
	if(parentOperator!=NULL)
	{
		if(parentOperator->isExpanded())
		{
			QPointF points[3] = {
				QPointF(-0.25*SLOT_SIZE, 0.433*SLOT_SIZE),
					 QPointF(-0.25*SLOT_SIZE, -0.433*SLOT_SIZE),
							 QPointF(0.5*SLOT_SIZE, 0.)
			};
			painter->drawConvexPolygon(points, 3);
		}
		else
		{
			QPointF points[3] = {
				QPointF(0.433*SLOT_SIZE, -0.25*SLOT_SIZE),
					QPointF(-0.433*SLOT_SIZE, -0.25*SLOT_SIZE),
							QPointF(0., 0.5*SLOT_SIZE)
			};
			painter->drawConvexPolygon(points, 3);
		}
	}
	else
	{
		QPointF points[3] = {
			QPointF(0.433*SLOT_SIZE, -0.25*SLOT_SIZE),
				QPointF(-0.433*SLOT_SIZE, -0.25*SLOT_SIZE),
					 QPointF(0., 0.5*SLOT_SIZE)
		};
		painter->drawConvexPolygon(points, 3);
	}

}

void OutSlot::hoverEnterEvent(QGraphicsSceneHoverEvent*)
{
	hasMouseFocus=true;
	((CalcItem*)parentItem())->setDisplayedString(slot->getName()+"  "+slot->getDescription());
	update();
}

void OutSlot::hoverLeaveEvent(QGraphicsSceneHoverEvent*)
{
	hasMouseFocus=false;
	((CalcItem*)parentItem())->setDisplayedString("");
	update();
}

void OutSlot::removeConnector(Connector* c)
{
	connectors.removeAll(c);
}

void OutSlot::addConnector(Connector* c)
{
	connectors.removeAll(c); // to ensure one and only one reference to a connector
	connectors.push_back(c);
}

void OutSlot::updateConnectorsPositions()
{
	foreach(Connector* c, connectors)
	{
		c->updatePosition();
	}
}


