#ifndef QCALC_SLOTS_H
#define QCALC_SLOTS_H

#include "qcalcCommon.h"

// Virtual input slot
// Connects to a virtual output slot
// If not in a black box, has an InputSlot for gui
// If in a black box, it CAN have an InputSlot for GUI, if this slot is not an inner one
class VInSlot
{
	friend class VOutSlot;
	
	public:
		// Name and description appear when hovering the gui associated, or when the graphic item is maximised
		VInSlot(VCalcItem* parent, const QString& name="", const QString& description="");
		
		// Deleting it also removes the possible connection from a VOutSlot
		virtual ~VInSlot();
		
		// Returns the connected virtual output slot if any, or NULL
		inline VOutSlot* connectedSlot() const {return _outSlot;}
		
		// Returns the virtual item connected to this slot, or NULL if none
		VCalcItem* connectedItem() const;
		
		// Returns the virtual item this slot _directly_ belongs to (not via a black box)
		inline VCalcItem* parent() const{return _parent;}
		
		// Returns the InSlot representing this virtual slot (either the one of the item it belongs to, or the one of the black box it belongs to). Does NOT create the GUI element
		InSlot* getGUI();
		
		// Removes the association to a GUI, but does NOT delete the GUI
		void removeGUI();
		
		// Associates to a GUI
		inline void setGUI(InSlot* gui){this->gui=gui;}
		
		inline bool hasGUI(){return (gui!=NULL);}
		
		inline QString getName(){return name;}
		inline QString getDescription(){return description;}
		
		void setName(const QString& name);
		void setDescription(const QString& description);
		void resetToolTips();
		
	protected:
		VOutSlot* _outSlot;
		VCalcItem* _parent;
		InSlot* gui;
		QString name;
		QString description;
};

// Virtual output slot
// Connects to (a) virtual input slot(s)
// If not in a black box, has an OutputSlot for gui
// If in a black box, it CAN have an OutputSlot for GUI, if this slot is not an inner one
class VOutSlot
{
	public:
		// Name and description appear when hovering the gui associated, or when the graphic item is maximised
		VOutSlot(VCalcItem* parent, const QString& name="", const QString& description="");
		
		// Deleting it also removes the possible connection from a VOutSlot
		virtual ~VOutSlot();
		
		// Returns the (possibly empty) list of VInSlots it is connected to
		QList<VInSlot*> connectedSlots() const;
		
		// Returns the (possibly empty) list of elements it is connected to (via its connectedSlots())
		QList<VCalcItem*> connectedItems() const;
		
		// The item it _directly_ belongs to (not via a black box)
		inline VCalcItem* parent() const{return _parent;}
		
		// Creates a connection to a virtual input slot. This will remove the possibly existing connections to this input slot
		void connectTo(VInSlot* slot);
		
		void removeConnectionTo(VInSlot* slot);
		
		// The GUI associated to this element, if it exists (not _inside_ a black box)
		OutSlot* getGUI();
		
		// Removes the association to a GUI, but does NOT delete the GUI
		void removeGUI();
		
		inline void setGUI(OutSlot* gui){this->gui=gui;}
		
		void setValue(DvObject_var value, bool ok)
		{
			this->value=value;
			this->valueOK=ok;
			reevaluationRequired=false;
		}
		
		DvObject_var getValue(bool& ok);
		DvObject_var getNotJoinedValue(bool& ok);
		
		// Flags the output for reevaluation. Else, it won't be reevaluated in the next computation involving it
		void setReevaluationFlag(bool b){reevaluationRequired=b;}
		
		inline QString getName(){return name;}
		inline QString getDescription(){return description;}
		
		void setName(const QString& name);
		void setDescription(const QString& description);
		void resetToolTips();
		
		inline bool hasGUI(){return (gui!=NULL);}
		
		bool openSlot; // identifies black box output slots
		
	protected:
		QList<VInSlot*> _inSlots;
		VCalcItem* _parent;
		OutSlot* gui;
		DvObject_var value;
		bool valueOK;
		bool reevaluationRequired;
		QString name;
		QString description;
};

// GUI element associated to a virtual Input Slot (VInSlot)
class InSlot: public QGraphicsEllipseItem
{
	public:
		InSlot(CalcItem* parent, VInSlot* slot);
		
		// Removes the GUI from the associated virtual slot, deletes the possible connector coming into it.
		virtual ~InSlot();
		
		// for qgraphicsitem_cast
		inline virtual int type() const { return Type;}
		enum { Type = TYPE_INSLOT };
		
		virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget=NULL);
		
		// removes connector and the underlying virtual connection
		void removeConnector(Connector* c);
		
		// Sets a connection (removing the possible previous one), connects the underlying virtual slots
		void setConnector(Connector* c);
		
		// Refreshes the connector position
		void updateConnectorPosition();

		Connector* getConnector(){return connector;}
		
		VInSlot* slot;
		
	protected:
		Connector* connector;
		void hoverEnterEvent(QGraphicsSceneHoverEvent*);
		void hoverLeaveEvent(QGraphicsSceneHoverEvent*);
		bool hasMouseFocus;
};

// GUI element associated to a virtual Output Slot (VOutSlot)
class OutSlot: public QGraphicsEllipseItem
{
	public:
		OutSlot(CalcItem* parent, VOutSlot* slot);
		
		// Removes the GUI from the associated virtual slot, deletes the possible connectors issued from it.
		virtual ~OutSlot();
		
		// for qgraphicsitem_cast
		inline virtual int type() const { return Type;}
		enum { Type = TYPE_OUTSLOT };
		
		virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget=NULL);
		
		// Removes connector, but does not manage the underlying connection
		void removeConnector(Connector* c);
		// Adds connector, but does not manage the underlying connection. It is done when the connector is actually connected to an InSlot.
		void addConnector(Connector* c);

		// Refreshes the connectors positions
		void updateConnectorsPositions();

		VOutSlot* slot;
		double width, height;
		
	protected:
		void hoverEnterEvent(QGraphicsSceneHoverEvent*);
		void hoverLeaveEvent(QGraphicsSceneHoverEvent*);
		bool hasMouseFocus;
		QList<Connector*> connectors;
};


#endif
