#ifndef TRIANGULATION_H
#define TRIANGULATION_H

#include <qthread.h>

#include "qdutil.h"
#include <iostream>
#include <float.h>



class Point2D
{
    public:
        
        double x, y, val, area, norm2;
        
        Point2D(double in_x=0., double in_y=0., double in_val=0., double in_area=0.): x(in_x), y(in_y), val(in_val), area(in_area), norm2(in_x*in_x+in_y*in_y)
        {}
        
        Point2D(const Point2D& p):x(p.x), y(p.y), val(p.val), area(p.area), norm2(p.norm2)
        {}

        inline double distance(const Point2D& p) const
        {
            return sqrt((p.x-x)*(p.x-x)+(p.y-y)*(p.y-y));
        }

        //Square of distance
        inline double distance2(const Point2D& p) const
        {
            return ((p.x-x)*(p.x-x)+(p.y-y)*(p.y-y));
        }
        
        inline double norm() const
        {
            return sqrt(norm2);
        }
        
        inline bool operator<(const Point2D& p) const
        {
            return (norm2<p.norm2);
        }
        
        inline bool operator==(const Point2D& p) const
        {
            return (p.x==x && p.y==y);
        }
        
        inline bool operator!=(const Point2D& p) const
        {
            return (p.x!=x || p.y!=y);
        }
        
        inline Point2D operator+(const Point2D& p) const
        {
            return Point2D(x+p.x, y+p.y);
        }
        
        inline Point2D operator-(const Point2D& p) const
        {
            return Point2D(x-p.x, y-p.y);
        }
        
        inline void opposite()
        {
            x=-x;
            y=-y;
        }
        
        inline Point2D translation(double x, double y)
        {
            return Point2D(this->x+x, this->y+y, val, area);
        }
        
        inline Point2D rotation(double angle)
        {
            double c(cos(angle)), s(sin(angle));
            return Point2D(c*x-s*y, s*x+c*y, val, area);
        }
};

class Point3D
{
    public:
        
        double x, y, z, val, area, norm2, orig_x, orig_y, orig_z;
        
        Point3D(double in_x=0., double in_y=0., double in_z=0., double in_val=0., double in_area=0.): x(in_x), y(in_y), z(in_z), val(in_val), area(in_area), norm2(in_x*in_x+in_y*in_y+in_z*in_z)
        {}
        
        Point3D(const Point2D& p, double in_z):x(p.x), y(p.y), z(in_z), val(p.val), area(p.area), norm2(p.x*p.x+p.y*p.y+in_z*in_z)
        {}
        
        Point3D(const Point3D& p):x(p.x), y(p.y), z(p.z), val(p.val), area(p.area), norm2(p.norm2), orig_x(p.orig_x), orig_y(p.orig_y), orig_z(p.orig_z)
        {}
        
        inline double distance(const Point3D& p) const
        {
            return sqrt((p.x-x)*(p.x-x)+(p.y-y)*(p.y-y)+(p.z-z)*(p.z-z));
        }

        // square of distance
        inline double distance2(const Point3D& p) const
        {
            return ((p.x-x)*(p.x-x)+(p.y-y)*(p.y-y)+(p.z-z)*(p.z-z));
        }
        
        inline double norm() const
        {
            return sqrt(norm2);
        }
        
        inline bool operator<(const Point3D& p) const
        {
            return (norm2<p.norm2);
        }
        
        inline bool operator==(const Point3D& p) const
        {
            return (p.x==x && p.y==y && p.z==z);
        }
        
        inline bool operator!=(const Point3D& p) const
        {
            return (p.x!=x || p.y!=y || p.z!=z);
        }
        
        inline Point3D operator+(const Point3D& p) const
        {
            return Point3D(x+p.x, y+p.y, z+p.z);
        }
        
        inline Point3D operator-(const Point3D& p) const
        {
            return Point3D(x-p.x, y-p.y, z-p.z);
        }
        
        inline Point3D operator*(const double d) const
        {
            return Point3D(x*d, y*d, z*d);
        }
        
        inline Point3D operator/(const double d) const
        {
            return Point3D(x/d, y/d, z/d);
        }
        
        inline double dotProd(const Point3D& p) const
        {
            return x*p.x+y*p.y+z*p.z;
        }
        
        inline Point3D crossProd(const Point3D& p) const
        {
            return Point3D(y*p.z-z*p.y, z*p.x-x*p.z, x*p.y-y*p.x);
        }
        
        inline void opposite()
        {
            x=-x;
            y=-y;
            z=-z;
        }
        
        inline Point3D translation(double x, double y, double z)
        {
            return Point3D(this->x+x, this->y+y, this->z+z, val, area);
        }
        
        inline Point3D rotationX(double angle)
        {
            double c(cos(angle)), s(sin(angle));
            return Point3D(x, c*y-s*z, s*y+c*z, val, area);
        }
        
        inline Point3D rotationY(double angle)
        {
            double c(cos(angle)), s(sin(angle));
            return Point3D(c*x+s*z, y, -s*x+c*z, val, area);
        }
        
        inline Point3D rotationZ(double angle)
        {
            double c(cos(angle)), s(sin(angle));
            return Point3D(c*x-s*y, s*x+c*y,z, val, area);
        }
};

class Point4D
{
    public:
        
        double x, y, z, t, val, area, norm2;
        double orig_x, orig_y, orig_z;
		static long long next_p_ID;
		long long p_ID;
		bool doneAlready;
        
        Point4D(double in_x=0., double in_y=0., double in_z=0., double in_t=0.,  double in_val=0., double in_area=0.)
        {
			x=in_x;
			y=in_y;
			z=in_z;
			t=in_t;
			val=in_val;
			area = in_area;
			norm2=in_x*in_x+in_y*in_y+in_z*in_z+in_t*in_t;
			orig_x=in_x;
			orig_y=in_y;
			orig_z=in_z;
			
			p_ID = Point4D::next_p_ID;
			Point4D::next_p_ID++;
			doneAlready = false;
		}
        
        Point4D(const Point3D& p, double in_t)
        {
			x=p.x;
			y=p.y;
			z=p.z;
			t=in_t;
			val=p.val;
			area = p.area;
			norm2=p.x*p.x+p.y*p.y+p.z*p.z+in_t*in_t;
			orig_x=p.orig_x;
			orig_y=p.orig_y;
			orig_z=p.orig_z;

			p_ID = Point4D::next_p_ID;
			Point4D::next_p_ID++;
			doneAlready = false;
		}
        
        Point4D(const Point4D& p)
        {
			x=p.x;
			y=p.y;
			z=p.z;
			t=p.t;
			val=p.val;
			area=p.area;
			norm2=p.norm2;
			orig_x=p.orig_x;
			orig_y=p.orig_y;
			orig_z=p.orig_z;
			
			p_ID = Point4D::next_p_ID;
			Point4D::next_p_ID++;
			doneAlready = false;
		}
        
        inline double distance(const Point4D& p) const
        {
            return sqrt((p.x-x)*(p.x-x)+(p.y-y)*(p.y-y)+(p.z-z)*(p.z-z)+(p.t-t)*(p.t-t));
        }

        // Square of distance
        inline double distance2(const Point4D& p) const
        {
            return ((p.x-x)*(p.x-x)+(p.y-y)*(p.y-y)+(p.z-z)*(p.z-z)+(p.t-t)*(p.t-t));
        }
        
        inline double norm() const
        {
            return sqrt(norm2);
        }
        
        inline bool operator<(const Point4D& p) const
        {
            return (norm2<p.norm2);
        }
        
        inline bool operator==(const Point4D& p) const
        {
			 return (p.p_ID == this->p_ID);
			
           // return (p.x==x && p.y==y && p.z==z && p.t==t);
        }
        
        inline bool operator!=(const Point4D& p) const
        {
			return (p.p_ID != this->p_ID);
           // return (p.x!=x || p.y!=y || p.z!=z || p.t!=t);
        }
        
        inline Point4D operator+(const Point4D& p) const
        {
            return Point4D(x+p.x, y+p.y, z+p.z, t+p.t);
        }
        
        inline Point4D operator-(const Point4D& p) const
        {
            return Point4D(x-p.x, y-p.y, z-p.z, t-p.t);
        }
        
        inline double dotProd(const Point4D& p) const
        {
            return x*p.x+y*p.y+z*p.z+t*p.t;
        }
        
        inline Point4D crossProd(const Point4D& v, const Point4D& w) const
        {
            return Point4D( -(y*(v.z*w.t-v.t*w.z)-v.y*(z*w.t-t*w.z)+w.y*(z*v.t-t*v.z)),
            (x*(v.z*w.t-v.t*w.z)-v.x*(z*w.t-t*w.z)+w.x*(z*v.t-t*v.z)),
            -(x*(v.y*w.t-v.t*w.y)-v.x*(y*w.t-t*w.y)+w.x*(y*v.t-t*v.y)),
            (x*(v.y*w.z-v.z*w.y)-v.x*(y*w.z-z*w.y)+w.x*(y*v.z-z*v.y))
                           );
        }
        
        inline void opposite()
        {
            x=-x;
            y=-y;
            z=-z;
            t=-t;
        }
};


class Element2D
{
    public:
        
        double points[3][2];
        double val[3];
        double area[3];
        Point2D circumCentre;
        double distToOrigin, dist1, dist2, dist3;
        double circumRadius2, elementArea;
        
        Element2D(){}
        
        Element2D(Point3D* p1, Point3D* p2, Point3D* p3)
        {
            points[0][0]=p1->x;  points[0][1]=p1->y; val[0]=p1->val; area[0]=p1->area;
            points[1][0]=p2->x;  points[1][1]=p2->y; val[1]=p2->val; area[1]=p2->area;
            points[2][0]=p3->x;  points[2][1]=p3->y; val[2]=p3->val; area[2]=p3->area;
            
            // Computing circumcircle parameters
            double a((p3->x-p2->x)*(p3->x-p2->x)+(p3->y-p2->y)*(p3->y-p2->y)), b((p3->x-p1->x)*(p3->x-p1->x)+(p3->y-p1->y)*(p3->y-p1->y)), c((p2->x-p1->x)*(p2->x-p1->x)+(p2->y-p1->y)*(p2->y-p1->y));
            double coeff1(a*(-a+b+c)), coeff2(b*(a-b+c)), coeff3(c*(a+b-c));
            double norm(coeff1+coeff2+coeff3);
            coeff1/=norm;
            coeff2/=norm;
            coeff3/=norm;
            
            circumCentre=Point2D(coeff1*p1->x+coeff2*p2->x+coeff3*p3->x, coeff1*p1->y+coeff2*p2->y+coeff3*p3->y);
            
            circumRadius2=(circumCentre.x-p1->x)*(circumCentre.x-p1->x)+(circumCentre.y-p1->y)*(circumCentre.y-p1->y);//circumCentre.distance2(*p1);
            
            elementArea=sqrt((a+b+c)*(-a+b+c)*(a-b+c)*(a+b-c))/4.;
            
            dist1=points[0][0]*points[0][0]+points[0][1]*points[0][1];//p1.norm2;
            dist2=points[1][0]*points[1][0]+points[1][1]*points[1][1];//p2.norm2;
            dist3=points[2][0]*points[2][0]+points[2][1]*points[2][1];//p3.norm2;
            
            if(dist1>dist2 && dist1>dist3)
            {
                distToOrigin=dist1;
            }
            else
            {
                if(dist2>dist1 && dist2>dist3)
                {
                    distToOrigin=dist2;
                }
                else
                {
                    distToOrigin=dist3;
                }
            }
        }
        
        bool findBarycentricCoordinates(const Point2D& p, double& c1, double& c2, double& c3) const
        {
            if(circumCentre.distance2(p)>circumRadius2)
            {
                return false;
            }

            double  A(points[0][0]-points[2][0]), B(points[1][0]-points[2][0]), C(points[2][0]-p.x),
                     D(points[0][1]-points[2][1]), E(points[1][1]-points[2][1]), F(points[2][1]-p.y);

            if(A==0. && B==0.)
            {
                double temp=A;
                A=D; D=temp;
                temp=B; B=E; E=temp;
                temp=C; C=F; F=temp;
                if(A==0. && B==0.) return false;
            }
            
            c1=(B*F-C*E)/(A*E-B*D);
            
            if(1.+c1<1.-1000.*DBL_EPSILON) return false;
            
            c2=(A*F-C*D)/(B*D-A*E);
            c3=1.-c1-c2;
            
            if(1.+c2<1.-1000.*DBL_EPSILON || 1.+c3<1.-1000.*DBL_EPSILON) return false;
            
            return true;
        }
        
        inline bool operator<(const Element2D& e) const
        {
            return distToOrigin<e.distToOrigin;
        }
};

class Facet3D;

// A 3D ridge is the segment of line between two adjacent facets
class Ridge3D
{
    public:
        Point3D *p1, *p2;
        Facet3D* f1;
        Facet3D* f2;
        
        Ridge3D(Point3D* _p1, Point3D* _p2): f1(NULL), f2(NULL)
        {
            if(*_p1<*_p2) // Notion of order to make comparisons easier
            {
                p1=_p1;
                p2=_p2;
            }
            else
            {
                p2=_p1;
                p1=_p2;
            }
			
        }
        
        inline bool operator==(const Ridge3D& r) const
        {
            return (r.p1==p1 && r.p2==p2);
        }
        
        inline bool operator!=(const Ridge3D& r) const
        {
            return (r.p1!=p1 || r.p2!=p2);
        }
        
        inline bool operator<(const Ridge3D& r) const
        {
            return (*p1<*r.p1 || (p1==r.p1 && *p2<*r.p2));
        }
};

// A 3D facet is a triangle, defined as its 3 ridges (edges)
class Facet3D
{
    public:
        
        static long long currentStatus;
        
        Facet3D():status(0){}
        
        Facet3D(Ridge3D* _r1, Ridge3D* _r2, Ridge3D* _r3):
                r1(_r1), r2(_r2), r3(_r3), p1(_r1->p1), p2(_r1->p2), status(0)
        {
            if(_r2->p1 != p1 && _r2->p1 != p2)
                p3=_r2->p1;
            else
                p3=_r2->p2;
            centre.x=(p1->x+p2->x+p3->x)/3.;
            centre.y=(p1->y+p2->y+p3->y)/3.;
            centre.z=(p1->z+p2->z+p3->z)/3.;
        }

        // Neighbour via ridge 1
        inline Facet3D* neighbour1() const
        {
            if(r1->f2==this)
                return r1->f1;
            return r1->f2;
        }

        // Neighbour via ridge 2
        inline Facet3D* neighbour2() const 
        {
            if(r2->f2==this)
                return r2->f1;
            return r2->f2;
        }

        // Neighbour via ridge 3
        inline Facet3D* neighbour3() const
        {
            if(r3->f2==this)
                return r3->f1;
            return r3->f2;
        }
        
        // Defines the normal, where p is outside the facet
        void setOuterNormal(Point3D* p)
        {
            outerNormal=Point3D((p2->y-p1->y)*(p3->z-p1->z)-(p2->z-p1->z)*(p3->y-p1->y), (p2->z-p1->z)*(p3->x-p1->x)-(p2->x-p1->x)*(p3->z-p1->z), (p2->x-p1->x)*(p3->y-p1->y)-(p2->y-p1->y)*(p3->x-p1->x));
            
            if(outerNormal.dotProd(*p-centre)>0.)
            {
                outerNormal.opposite();
            }
            // just to make sure the outer normal vector is not ridiculously large or small, to improve subsequent floating point computations for visibility
            double norm=outerNormal.norm();
            outerNormal.x/=norm;
            outerNormal.y/=norm;
            outerNormal.z/=norm;
        }

        // Distance of the facet to a point
        inline double distance(Point3D* p)
        {
            return outerNormal.x*(p->x-centre.x)+outerNormal.y*(p->y-centre.y)+outerNormal.z*(p->z-centre.z);
        }
        
        std::list<Point3D*> halfspace; // Set of points in the outer half-space
        
        Point3D outerNormal;
        Point3D centre;
        Point3D *p1, *p2, *p3;
        Ridge3D *r1, *r2, *r3;

        long long status;
};

// Tetrahedron
class Element3D
{
    public:
        
        double points[4][3];
        double val[4];
        double area[4];
        Point3D circumCentre;
        
        double distToOrigin, dist1, dist2, dist3, dist4;
        double circumRadius2, elementArea, A, B, C, E, F, G, I, J, K;
        
        Element3D(){};
        
        Element3D(Point4D* p1, Point4D* p2, Point4D* p3, Point4D* p4)
        {
            points[0][0]=p1->orig_x;  points[0][1]=p1->orig_y; points[0][2]=p1->orig_z; val[0]=p1->val; area[0]=p1->area;
            points[1][0]=p2->orig_x;  points[1][1]=p2->orig_y; points[1][2]=p2->orig_z; val[1]=p2->val; area[1]=p2->area;
            points[2][0]=p3->orig_x;  points[2][1]=p3->orig_y; points[2][2]=p3->orig_z; val[2]=p3->val; area[2]=p3->area;
            points[3][0]=p4->orig_x;  points[3][1]=p4->orig_y; points[3][2]=p4->orig_z; val[3]=p4->val; area[3]=p4->area;
            
            dist1=points[0][0]*points[0][0]+points[0][1]*points[0][1]+points[0][2]*points[0][2];
            dist2=points[1][0]*points[1][0]+points[1][1]*points[1][1]+points[1][2]*points[1][2];
            dist3=points[2][0]*points[2][0]+points[2][1]*points[2][1]+points[2][2]*points[2][2];
            dist4=points[3][0]*points[3][0]+points[3][1]*points[3][1]+points[3][2]*points[3][2];
            
            double d12((points[1][0]-points[0][0])*(points[1][0]-points[0][0])+(points[1][1]-points[0][1])*(points[1][1]-points[0][1])+(points[1][2]-points[0][2])*(points[1][2]-points[0][2])), d13((points[2][0]-points[0][0])*(points[2][0]-points[0][0])+(points[2][1]-points[0][1])*(points[2][1]-points[0][1])+(points[2][2]-points[0][2])*(points[2][2]-points[0][2])), d14((points[3][0]-points[0][0])*(points[3][0]-points[0][0])+(points[3][1]-points[0][1])*(points[3][1]-points[0][1])+(points[3][2]-points[0][2])*(points[3][2]-points[0][2]));
            
            //             circumCentre=((P2-P1).crossProd(P3-P1)*d14*d14+(P4-P1).crossProd(P2-P1)*d13*d13+(P3-P1).crossProd(P4-P1)*d12*d12)/(2.*(P2-P1).dotProd((P3-P1).crossProd(P4-P1)));
            // Previous (commented out) computation unrolled for better performance
            
            double factor=1./(2.*(
                            (points[1][0]-points[0][0])*
                                    ((points[2][1]-points[0][1])*(points[3][2]-points[0][2])-(points[2][2]-points[0][2])*(points[3][1]-points[0][1]))+
                            (points[1][1]-points[0][1])*
                                    ((points[2][2]-points[0][2])*(points[3][0]-points[0][0])-(points[2][0]-points[0][0])*(points[3][2]-points[0][2]))+
                            (points[1][2]-points[0][2])*
                                    ((points[2][0]-points[0][0])*(points[3][1]-points[0][1])-(points[2][1]-points[0][1])*(points[3][0]-points[0][0]))));
            double x=   (((points[1][1]-points[0][1])*(points[2][2]-points[0][2])-(points[1][2]-points[0][2])*(points[2][1]-points[0][1]))*d14+
                        ((points[3][1]-points[0][1])*(points[1][2]-points[0][2])-(points[3][2]-points[0][2])*(points[1][1]-points[0][1]))*d13+
                        ((points[2][1]-points[0][1])*(points[3][2]-points[0][2])-(points[2][2]-points[0][2])*(points[3][1]-points[0][1]))*d12)
                        *factor;
            
            double y=   (((points[1][2]-points[0][2])*(points[2][0]-points[0][0])-(points[1][0]-points[0][0])*(points[2][2]-points[0][2]))*d14+
                        ((points[3][2]-points[0][2])*(points[1][0]-points[0][0])-(points[3][0]-points[0][0])*(points[1][2]-points[0][2]))*d13+
                        ((points[2][2]-points[0][2])*(points[3][0]-points[0][0])-(points[2][0]-points[0][0])*(points[3][2]-points[0][2]))*d12)
                        *factor;
            
            double z=   (((points[1][0]-points[0][0])*(points[2][1]-points[0][1])-(points[1][1]-points[0][1])*(points[2][0]-points[0][0]))*d14+
                        ((points[3][0]-points[0][0])*(points[1][1]-points[0][1])-(points[3][1]-points[0][1])*(points[1][0]-points[0][0]))*d13+
                        ((points[2][0]-points[0][0])*(points[3][1]-points[0][1])-(points[2][1]-points[0][1])*(points[3][0]-points[0][0]))*d12)
                    *factor;
                        
            circumCentre=Point3D(x, y, z);
                        
            circumRadius2=circumCentre.x*circumCentre.x+circumCentre.y*circumCentre.y+circumCentre.z*circumCentre.z;
            
            circumCentre.x+=p1->x;
            circumCentre.y+=p1->y;
            circumCentre.z+=p1->z;
                    
            //distToOrigin=circumCentre.norm2;
            if(dist1>=dist2 && dist1>=dist3 && dist1>=dist4)
            {
                distToOrigin=dist1;
            }
            else
            {
                if(dist2>=dist1 && dist2>=dist3 && dist2>=dist4)
                {
                    distToOrigin=dist2;
                }
                else
                {
                    if(dist3>=dist1 && dist3>=dist2 && dist3>=dist4)
                    {
                        distToOrigin=dist3;
                    }
                    else
                    {
                        distToOrigin=dist4;
                    }
                }
            }
            
            A=points[0][0]-points[3][0];
            B=points[1][0]-points[3][0];
            C=points[2][0]-points[3][0];
            E=points[0][1]-points[3][1];
            F=points[1][1]-points[3][1];
            G=points[2][1]-points[3][1];
            I=points[0][2]-points[3][2];
            J=points[1][2]-points[3][2];
            K=points[2][2]-points[3][2];

        }

        // Volume of the tetrahedron made of p, and points of indexes i, j and k of this tetrahedron. Used to compute barycentric coordinates
        inline double volume(const Point3D& p, long i, long j, long k)
        {
            //VOLUME = |a.(bxc)|/6, with a=(A-D), b=(B-D), c=(C-D)
            double ax(points[i][0]-p.x), ay(points[i][1]-p.y), az(points[i][2]-p.z),
                      bx(points[j][0]-p.x), by(points[j][1]-p.y), bz(points[j][2]-p.z),
                         cx(points[k][0]-p.x), cy(points[k][1]-p.y), cz(points[k][2]-p.z);
            
                         return fabs(ax*(by*cz-bz*cy)+ay*(bz*cx-bx*cz)+az*(bx*cy-by*cx))/6.;
        }
        
        inline double volume()
        {
            //VOLUME = |a.(bxc)|/6, with a=(A-D), b=(B-D), c=(C-D)
            double ax(points[0][0]-points[3][0]), ay(points[0][1]-points[3][1]), az(points[0][2]-points[3][2]),
                      bx(points[1][0]-points[3][0]), by(points[1][1]-points[3][1]), bz(points[1][2]-points[3][2]),
                         cx(points[2][0]-points[3][0]), cy(points[2][1]-points[3][1]), cz(points[2][2]-points[3][2]);
            
                         return fabs(ax*(by*cz-bz*cy)+ay*(bz*cx-bx*cz)+az*(bx*cy-by*cx))/6.;
        }

        // Gets the barycentric coordinates of p in the frame defined by the tetrahedron
        bool findBarycentricCoordinates(const Point3D& p, double& c1, double& c2, double& c3, double& c4)
        {
            if(circumCentre.distance2(p)>circumRadius2)
            {
                return false;
            }

            c1=volume(p, 1, 2, 3);
            c2=volume(p, 0, 2, 3);
            c3=volume(p, 0, 1, 3);
            c4=volume(p, 0, 1, 2);
            double total=volume();
            if(total*1.01<c1+c2+c3+c4) return false;

            total=c1+c2+c3+c4;
            c1/=total;
            c2/=total;
            c3/=total;
            c4/=total;
            return true;
        }
        
        inline bool operator<(const Element3D& e) const
        {
            return distToOrigin<e.distToOrigin;
        }
        

};

class Facet4D;


// A ridge is the space shared by 2 adjacent facets. In 4D, this a triangle (facets are tetrahedra)
class Ridge4D
{
    public:
        Point4D *p1, *p2, *p3;

        Facet4D* f1;
        Facet4D* f2;

		bool seenAlready;
		
        Ridge4D(Point4D* p1, Point4D* p2, Point4D* p3): f1(NULL), f2(NULL)
        {
			seenAlready = false;
            if(p1<p2) // Notion of order to make comparisons easier
            {
                if(p2<p3)
                {
                    this->p1=p1;
                    this->p2=p2;
                    this->p3=p3;
                }
                else
                {
                    if(p3<p1)
                    {
                        this->p1=p3;
                        this->p2=p1;
                        this->p3=p2;
                    }
                    else
                    {
                        this->p1=p1;
                        this->p2=p3;
                        this->p3=p2;
                    }
                }
            }
            else
            {
                //p2<p1
                if(p3<p2)
                {
                    this->p1=p3;
                    this->p2=p2;
                    this->p3=p1;
                }
                else
                {
                    if(p1<p3)
                    {
                        this->p1=p2;
                        this->p2=p1;
                        this->p3=p3;
                    }
                    else
                    {
                        this->p1=p2;
                        this->p2=p3;
                        this->p3=p1;
                    }
                }
            }
        }
        
        inline bool operator==(const Ridge4D& r) const
        {
            return (*r.p1==*p1 && *r.p2==*p2 && *r.p3==*p3);
        }
        
        inline bool operator!=(const Ridge4D& r) const
        {
            return (r.p1!=p1 || r.p2!=p2 || r.p3!=p3);
        }
        
        inline bool operator<(const Ridge4D& r) const
        {
            return (*p1<*r.p1 || (*p1==*r.p1 && *p2<*r.p2) || (*p1==*r.p1 && *p2==*r.p2 && *p3<*r.p3));
        }
		
        inline bool equivalent(Point4D* ep1, Point4D* ep2, Point4D* ep3) const
        {
			return (ep1 == this->p1  && ep2 == this->p2 && ep3 == this->p3);
			 
        }
};

// In 4D, a facet is a tetrahedron
class Facet4D
{
    public:
        
        static long long currentStatus;
		bool doneAlready; 
        
        Facet4D():status(0){}
        
        Facet4D(Ridge4D* _r1, Ridge4D* _r2, Ridge4D* _r3, Ridge4D* _r4):
                r1(_r1), r2(_r2), r3(_r3), r4(_r4), p1(_r1->p1), p2(_r1->p2), p3(_r1->p3), status(0)
                {
 
                    if(_r2->p1 != _r1->p1 && _r2->p1 != _r1->p2 && _r2->p1 != _r1->p3)
                    {
                        p4=_r2->p1;
                    }
                    else
                    {
                        if(_r2->p2 != _r1->p1 && _r2->p2 != _r1->p2 && _r2->p2 != _r1->p3)
                        {
                            p4=_r2->p2;
                        }
                        else
                        {
                            p4=_r2->p3;
                        }
                    }

                    
                	centre.x=(p1->x+p2->x+p3->x+p4->x)*.25;
                    centre.y=(p1->y+p2->y+p3->y+p4->y)*.25;
                    centre.z=(p1->z+p2->z+p3->z+p4->z)*.25;
                    centre.t=(p1->t+p2->t+p3->t+p4->t)*.25;
					doneAlready = false;
                }

                // Neighbour via ridge 1
                inline Facet4D* neighbour1() const
                {
					
                    if(r1->f2==this)
                        return r1->f1;
                    return r1->f2;
					
                }

                // Neighbour via ridge 2
                inline Facet4D* neighbour2() const 
                {
                    if(r2->f2==this)
                        return r2->f1;
                    return r2->f2;
                }

                // Neighbour via ridge 3
                inline Facet4D* neighbour3() const
                {
                    if(r3->f2==this)
                        return r3->f1;
                    return r3->f2;
                }

                // Neighbour via ridge 4
                inline Facet4D* neighbour4() const
                {
                    if(r4->f2==this)
                        return r4->f1;
                    return r4->f2;
                }
        
                // Neighbour via ridge 1
                void setOuterNormal(Point4D* p)
                {
                    outerNormal=(*p2-*p1).crossProd(*p3-*p1, *p4-*p1);
                    if(outerNormal.dotProd(*p-centre)>0.)
                    {
                        outerNormal.opposite();
                    }
                    // just to make sure the outer normal vector is not ridiculously large or small, to improve floating point computations
                    double norm=outerNormal.norm();
                    outerNormal.x/=norm;
                    outerNormal.y/=norm;
                    outerNormal.z/=norm;
                    outerNormal.t/=norm;
                }
        
                inline double distance(Point4D* p)
                {
                    return outerNormal.dotProd(*p-centre);
                }
        
                std::list<Point4D*> halfspace; // List of points in the outer half space ("Visibility list")
        
                Point4D outerNormal;
                Point4D centre;
                Point4D *p1, *p2, *p3, *p4;
                Ridge4D *r1, *r2, *r3, *r4;

                long long status;
        
};


class Ridge4Dlist: public list<Ridge4D*>
{
    public:
		inline void addUnique(Ridge4D* pt){
			if(pt->seenAlready) return;
			push_front(pt);
			pt->seenAlready = true;
		}
		
		inline bool find(Ridge4D* pt){
			for(list<Ridge4D*>::iterator iter=this->begin(); iter!=this->end(); iter++)
        	{
				if(*iter == pt) return true;
			}
			return false;
		}
		
		void reset(){
			for(list<Ridge4D*>::iterator iter=this->begin(); iter!=this->end(); iter++)
        	{
				(*iter)->seenAlready = false;
			}
		}

		 
	
};


class Facet4Dlist: public list<Facet4D*>
{
    public:
		
		void reset(){
			for(list<Facet4D*>::iterator iter=this->begin(); iter!=this->end(); iter++)
        	{
				(*iter)->doneAlready = false;
			}
		}

		 
	
};


class Point4Dlist: public list<Point4D*>
{
    public:
		
		void reset(){
			for(list<Point4D*>::iterator iter=this->begin(); iter!=this->end(); iter++)
        	{
				(*iter)->doneAlready = false;
			}
		}
		
		void makeUnique(){
			Point4Dlist newList;

			for(list<Point4D*>::iterator iter=this->begin(); iter!=this->end(); iter++)
        	{
				if(!(*iter)->doneAlready){
					(*iter)->doneAlready = true;
					newList.push_front(*iter);
				}
			}
			newList.reset(); // set the doneAlready flags to false for next call
			this->clear();
			this->splice(this->begin(),newList); 
			
		}
	
};



void quickHull(list<Point4D*>& points, list<Element3D>& delaunay);
void quickHull(list<Point3D*>& points, list<Element2D>& delaunay);

void triangulation(std::list<Point3D>& in, std::list<Element3D>& out);
void interpolation(list<Point3D>& points, list<Element3D>& hull);
void triangulation(std::list<Point2D>& in, std::list<Element2D>& out);
void interpolation(list<Point2D>& points, list<Element2D>& hull);


void test_integration_oversampling();
void test_integration_subsampling();
void test2();
void test2D();
void test3D();
void test3Dsphere();
void testRealConditions3D();
void testRealConditions2D(bool with_plot);

#endif
