/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	Q4SC_Normal.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS 4 SC Determination of Boundary Normal and Velocity wrapper
*
* PURPOSE:      Calculate the normal and boundary velocity based on the
*               crossing of the boundary (assumed planar) by 4 spacecraft.
*               Based on the methd in Ch 10.4.3 of the ISSI book.
*
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*   INPUT   ANY_DATA_OBJECT    dr1  pos of s/c1 as const vec or time ser
    INPUT   ANY_DATA_OBJECT    dr2 ,
    INPUT   ANY_DATA_OBJECT    dr3 ,
    INPUT   ANY_DATA_OBJECT    dr4 ,
    INPUT   TIME               tidr, centre time used if dr's are time ser
    INPUT   ANY_DATA_OBJECT    ti1 , event time as s/c 1 in secs or Time Int Obj
    INPUT   ANY_DATA_OBJECT    ti2 ,
    INPUT   ANY_DATA_OBJECT    ti3 ,
    INPUT   ANY_DATA_OBJECT    ti4 ,
    INPUT   ANY_DATA_OBJECT    taus , string array with 4 elements, the 4 crossing durations

    Out:

    OUTPUT  MATRIX             n_cva, n_cta,   
                                     normal vectors with same frame as dr's
                                     n_x forced positive
    OUTPUT  FLOAT              Vn_cva, Vn_cta,
                                      velocity along normal

* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: Hacked from Boundary Normal routine
*
* HISTORY: November 2003 Steve Schwartz 
*          January 2004 AJA added error traps
*          28 April, 2004 SEH : added CTA part
*          01 Jun, 2004 SEH : added combined CTA/CVA 
*          25 Aug, 2006 AJA : added to QSAS distribution
*          2 June, 2015 AJA : migrated to use DVOS core
*
*- mode: C++ ; c-file-style: "stroustrup" */
/***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include "qplug_if.h"
#include "qrtn.h"

#include "qdutil.h"
#include "DvObj.h"

#include <qstring.h>
#include <qdatetime.h>

//const double Re = 6.3712e6; //Earth radius [m]  - a global	

using namespace QSAS;  

// local prototypes

extern "C" QplugReturnStatus DvCVA_CTA_Calculator(vector <DvObject_var>&, vector <DvObject_var>&);

QplugReturnStatus Q4SCPosition(RtnVector *vecp, DvObject_var pos_in, DvTime tidr);

void RtnVector_diag_print(RtnVector *vecp);

DvObject_var str2Vec(DvString s);

DvObject_var CTA_calc(double R[4][3], double dt[4], DvString taustring, 
		       DvObject_var &vn_cta, DvObject_var &d_cta, 
		       DvObject_var &vn_com, DvObject_var &d_com, DvObject_var &n_com, 
		       DvObject_var &v_cva, DvObject_var &d_cva, DvObject_var &n_cva, 
		       DvObject_var &v_cva_curve,  DvObject_var &v_cta_curve, DvObject_var &v_com_curve, 
		       DvTime tidr, DvObject_var &tau);

double vecsum(DvObject_var V);

 
#define MAT_INV_EPS 10*DBL_EPSILON

typedef enum MAT_INV_RET_E{
MAT_INV_OK,
MAT_INV_SINGULAR,
MAT_INV_UNKNOWN
} MAT_INV_RET;

MAT_INV_RET mat_inv(double a[3][3], double inverse[3][3]);
double determinant(double mat[3][3]);
void multiply_mat(double a[3][3],double b[3][3],double prod[3][3]);
void multiply_mat_vec(double a[3][3],double b[3],double prod[3]);
void print_mat(double a[3][3]);
 
  
/**********************************************************************/
QplugReturnStatus DvCVA_CTA_Calculator ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
/*!
  \brief Perform a 4SC boundary normal calculation
  \param call_list - pointer to a list of input objects
*/
/**********************************************************************/
{
// define some C variables

int i,j,sign_nx;
double dt[4];
double  A[3][3], A_INV[3][3], B[3];
double  XX[3],n[3];
double drvec[4][3], signedmodX;
MAT_INV_RET ret_mat_inv;

// call-list arguments
 
DvObject_var   drin[4];        //  0-3
DvTime         tidr;           //  4
DvObject_var   ti_tvar[4];    //  5-8 
DvObject_var   n_var;          //  10
DvObject_var   Vn_var;         //  12

// define dvos data objects

RtnVector  dr[4];

QplugReturnStatus ret_status;

if(QuSlotDebug()){
   printf("Q4SC Plugin Diagnostics On; undefine in .cc file and recompile\n");
   fflush(stdout);
}

// Ensure no input slots are empty

 if(inputs.size() != 10 ){
    QplugAppendTextDisplay("At least one input slot is empty");
    return QPLUG_FAILURE;
 }
 
// ...4 items are the s/c positions

for (i=0; i<4; i++){
  drin[i] = inputs[i];
  if (drin[i].is_nil()){
    char pos_err_string[200]; 
    sprintf(pos_err_string,"Q4SC>> Error getting pos dr[%d] object\n",i+1);
    QplugAppendTextDisplay(pos_err_string);
  }
	if(QuSlotDebug()){
		printf("Q4SC got input s/c %d position\n",i);
		fflush(stdout);
	}
}

if( inputs[4].is_nil() ) {
  QplugAppendTextDisplay ("Q4SC>> Time object for tidr invalid.\n");
  return QPLUG_FAILURE;
}
tidr = inputs[4]->time();


if(QuSlotDebug()){
   printf("Q4SC got input time interval %d\n",i);
   fflush(stdout);
}


for (i=0; i<4; i++){
  ret_status = Q4SCPosition(&(dr[i]),drin[i], tidr);
  if (ret_status != QPLUG_SUCCESS){
    char pos_err_string[200]; 
    sprintf(pos_err_string,"Q4SC>> Q4SCPosition failed for sc[%d]\n",i);
    QplugAppendTextDisplay(pos_err_string);
    return QPLUG_FAILURE;
  }
  if(QuSlotDebug()){
     printf("Q4SC calculated s/c %d position...\n",i);
	 fflush(stdout);
  }
  // dr holds single values for each S/C
  drvec[i][0] = dr[i].ptr->getVectorComponent(DV_X)->dbl(0);  
  drvec[i][1] = dr[i].ptr->getVectorComponent(DV_Y)->dbl(0);
  drvec[i][2] = dr[i].ptr->getVectorComponent(DV_Z)->dbl(0);
	if(QuSlotDebug()){
		printf("and filled drvec = (%f, %f,%f)\n",drvec[i][0],drvec[i][1],drvec[i][2]);
		fflush(stdout);
	}
}

// dr[i]'s now have a sequence with 1 element, the position
// of the crossing at s/c i with one timetag, and drvecs are C-arrays 

for (i=0; i<4; i++){
    ti_tvar[i] = inputs[i+5];
    if (ti_tvar[i].is_ok() ){
        if(ti_tvar[i]->is_dbl()) {
            dt[i] = ti_tvar[i]->dbl();
        }
        else if(ti_tvar[i]->is_time()){
            dt[i] = ti_tvar[i]->time() - tidr;
        }
        else if(ti_tvar[i]->is_event()) {
            dt[i] = ti_tvar[i]->event().duration();
        }
        else{
            dt[i] = ti_tvar[i]->asDouble();
        }
    } 
    else {
        char time_err_string[200];
        sprintf(time_err_string,"Q4SC ERROR>> ti[%d] neither scalar nor time interval\n",i+1);
        QplugAppendTextDisplay(time_err_string);
        return QPLUG_FAILURE;
    }
	if(QuSlotDebug()){
		printf("Q4SC calculated s/c %d dt as %g\n",i,dt[i]);
		fflush(stdout);
	}
}

// dt[i] now contains times in secs for all 4 s/c

// Now fill dr matrix and right hand sides; s/c 4 is the reference
// [it doesn't matter which s/c is the ref; there are only 3 indep pairs]

for(i=0; i<3; i++){
  for(j=0; j<3; j++){
    A[i][j] = drvec[i][j] - drvec[3][j];
    printf(" %10.2f",drvec[i][j]);
	fflush(stdout);
  }
  printf("\n");
  fflush(stdout);
  B[i]      = dt[i] - dt[3];
}

if(QuSlotDebug()){
   printf("Q4SC A matrix is\n");
   fflush(stdout);
   print_mat(A);
}
ret_mat_inv = mat_inv(A,A_INV);
switch (ret_mat_inv) {
   case MAT_INV_OK:
     break;
   case MAT_INV_SINGULAR:
     QplugAppendTextDisplay("Q4SC ERROR>> dr matrix singular\n");
     return QPLUG_FAILURE;
     break;
   case MAT_INV_UNKNOWN: default:
     QplugAppendTextDisplay("Q4SC ERROR>> Matrix Inv unknown error\n");
     return QPLUG_FAILURE;
     break;
}
if(QuSlotDebug()){
   printf("Q4SC A inv matrix found successfully as\n");
   printf("Q4SC B       is (%g, %g, %g)\n", B[0],B[1],B[2]);
   print_mat(A_INV);
   fflush(stdout);
}


multiply_mat_vec(A_INV, B, XX);
sign_nx = (XX[0] < 0 ) ? -1 : 1;  

signedmodX = (float)sign_nx*sqrt(XX[0]*XX[0] + XX[1]*XX[1] + XX[2]*XX[2]);
Vn_var = new DvObject((double)(1.0/signedmodX));

n_var  = new DvObject(0.0, DIM3);
n_var->copy_xrefs_from( dr[3].ptr); // copy  frame, etc.
char crossingt_str[200];
char outstr[200];
sprintf(crossingt_str,"%s", dr[3].t->time(0).getISOstring().c_str());
sprintf(outstr,"Q4SC>> Crossing at %s\n\n",  crossingt_str);
QplugAppendTextDisplay(outstr);


// added by AJA Feb 2004
n_var->change_xref(DEPEND_0, dr[3].t); // copy  time tag

for (i=0; i<3; i++) {
  n[i]  = XX[i]/signedmodX;
  n_var->dbl(i) = n[i];
}
   //  -- SEH, 28 April 2004 - CTA and COM calculations ------

   vector <size_t> dim4;
   dim4.push_back(4);
  
   DvObject_var n_cta  = new DvObject(0.0, DIM3);
   DvObject_var n_com  = new DvObject(0.0, DIM3);
   DvObject_var vn_cta = new DvObject(0.0, DIM4);
   DvObject_var vn_com = new DvObject(0.0, DIM4);
   DvObject_var d_cta  = new DvObject(0.0, DIM4);
   DvObject_var d_com  = new DvObject(0.0, DIM4);
   DvObject_var d_cva  = new DvObject(0.0, DIM4);
   DvObject_var tau    = new DvObject(0.0, DIM4);
   DvObject_var v_cta_curve =  new DvObject(0.0, 120);
   DvObject_var v_cva_curve =  new DvObject(0.0, 120);
   DvObject_var v_com_curve =  new DvObject(0.0, 120);
 
   DvObject_var tauS = inputs[9];
   
   if (tauS.is_ok()) {
   	DvString tau_string = tauS->asStr();
   	n_cta = CTA_calc(drvec,dt,tau_string,
					      vn_cta, d_cta, 
					      vn_com, d_com, n_com, 
					      Vn_var, d_cva, n_var,
					      v_cva_curve,v_cta_curve,v_com_curve,
					      tidr,tau);
   }
   printf("d_cta = %f\n",d_cta->dbl(0));fflush(stdout);
   
   n_cta->change_xref(FIELDNAM, "CTA_normal");
   n_cta->change_xref(SI_CONVERSION, "1>()");

   //  -- End of CTA and COM calculations ------


if(QuSlotDebug()){
   printf("Q4SC normal velocity is %g\n", Vn_var->asDouble() );

   printf("Q4SC B       is (%g, %g, %g)\n", B[0],B[1],B[2]);
   printf("Q4SC XX      is (%g, %g, %g)\n", XX[0],XX[1],XX[2]);
   printf("Q4SC norm    is %g\n",signedmodX);
   printf("Q4SC normal  is (%g, %g, %g)\n", n[0],n[1],n[2]);
   fflush(stdout);
}

// set the attributes for the output objects

n_var->change_xref(FIELDNAM, "CVA_normal");
Vn_var->change_xref(FIELDNAM, "Vn_CVA");

if(QuSlotDebug()){
   printf("Q4SC FIELDNAMs set\n");
   fflush(stdout);
}

n_var->change_xref(SI_CONVERSION, "1>()");
char si_string[200];
sprintf(si_string, "%s s^-1",drin[3]->getXrefText(SI_CONVERSION).c_str());
Vn_var->change_xref(SI_CONVERSION, si_string);

if(QuSlotDebug()){
   printf("Q4SC SI_conversions set\n");
   fflush(stdout);
}

n_var->change_xref(UNITS, " ");
char units_str[200];
sprintf(units_str,"%s/s",drin[3]->getXrefText(UNITS).c_str());
Vn_var->change_xref(UNITS, units_str);

if(QuSlotDebug()){
   printf("Q4SC UNITS set\n");
   fflush(stdout);
}

Vn_var->change_xref("Crossing Time", crossingt_str);
n_var->change_xref("Crossing Time", crossingt_str);

// --- SEH CTA ---
n_cta->change_xref("Crossing Time", crossingt_str);

char frame_str[200];
sprintf(frame_str,"%s", drin[3]->getXrefText(FRAME).c_str());
n_var->change_xref(FRAME, frame_str);
n_cta->change_xref(FRAME, frame_str);

v_cta_curve->change_xref(UNITS, units_str);
v_cva_curve->change_xref(UNITS, units_str);
v_com_curve->change_xref(UNITS, units_str);

outputs.push_back( n_var);     
// --- SEH CTA ---
outputs.push_back( n_cta);     
outputs.push_back( Vn_var);     
outputs.push_back( v_cta_curve);     
outputs.push_back( v_cva_curve);     
outputs.push_back( v_com_curve);     


// --- SEH 28 April : modifications to display CTA results ---
// put some results to plug-in text window
  sprintf(outstr,"n_cva : [%10.4g  %10.4g  %10.4g] ", n[0], n[1], n[2]);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"v_cva :  %10.1f %s\n", Vn_var->dbl(), Vn_var->getXrefText(UNITS).c_str());
  QplugAppendTextDisplay(outstr);
  double d_cva_ave=0.00;
  double v_cta_ave=0.00;
  double v_com_ave=0.00;
  double d_com_ave=0.00;

  for (i=0; i<4; i++) {
    d_cva_ave += 0.25 * d_cva->dbl(i);
    v_cta_ave += 0.25 * vn_cta->dbl(i);
    v_com_ave += 0.25 * vn_com->dbl(i);
    d_com_ave += 0.25 * d_com->dbl(i);
  }
  sprintf(outstr,"d_cva  :  %10.1f  %10.1f  %10.1f  %10.1f  (%10.1f) km\n\n",
	  d_cva->dbl(0),d_cva->dbl(1),d_cva->dbl(2),d_cva->dbl(3),d_cva_ave);
  QplugAppendTextDisplay(outstr);
 

  sprintf(outstr,"n_cta  : [%10.4g  %10.4g  %10.4g]\n",n_cta->dbl(0),n_cta->dbl(1),n_cta->dbl(2));
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"v_cta  :  %10.2f  %10.2f  %10.2f  %10.2f  (%10.2f) km/s\n",
	  vn_cta->dbl(0),vn_cta->dbl(1),vn_cta->dbl(2),vn_cta->dbl(3),v_cta_ave);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"d_cta  : %10.2f  km\n\n",d_cta->dbl(0));
  QplugAppendTextDisplay(outstr);
  
  sprintf(outstr,"n_com  : [%10.4g  %10.4g  %10.4g]\n",n_com->dbl(0),n_com->dbl(1),n_com->dbl(2));
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"v_com  :  %10.2f  %10.2f  %10.2f  %10.2f  (%10.2f) km/s\n",
	  vn_com->dbl(0),vn_com->dbl(1),vn_com->dbl(2),vn_com->dbl(3),v_com_ave);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"d_com  :  %10.1f  %10.1f  %10.1f  %10.1f  (%10.1f) km\n\n",
	  d_com->dbl(0),d_com->dbl(1),d_com->dbl(2),d_com->dbl(3),d_com_ave);
  QplugAppendTextDisplay(outstr);
   

return QPLUG_SUCCESS;

}

/**********************************************************************/
QplugReturnStatus Q4SCPosition(	RtnVector *vecp,
                              	DvObject_var pos_in,
			      				DvTime tidr){
/**********************************************************************/

QplugReturnStatus ret_status;

// sort out position of crossing. Input can either be a time series or
// a constant vector, so this a bit complicated. Separate routines
// return in vecp a sequence of 1 cartesian vector whose timetag
// is the centre of the time interval time_int, and whose frame is
// that of pos_in. 

if (!pos_in->isThreeVector()){
   QplugAppendTextDisplay("Q4SC>> Const pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}

DvObject_var tt = new DvObject(tidr);

vecp->t = tt;
vecp->n_recs = 1;
vecp->in_frame = vecp->QrGetFrameFromAttr(pos_in);
if (vecp->in_frame == UNKNOWN){
  QplugAppendTextDisplay("Q4SC>> Sequence pos_in unknown frame\n");
  return QPLUG_FAILURE;
}

if( pos_in->seqSize() > 1) {
 
	// linearly interpolate crossing position from bracketing time tags
	// Note that Qrtn routines only give correct answers if given x,y,z representation
	vecp->ptr = pos_in->interpAt(tidr);
} 
else {
	vecp->ptr = pos_in;
} 

return QPLUG_SUCCESS;
}



/**************************************************************************/
void RtnVector_diag_print(RtnVector *vecp){
  printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
        vecp->IsPositionVector,vecp->in_frame,vecp->out_frame, vecp->n_recs);
  printf("First Vector: (%9.4f, %9.4f, %9.4f) at %s\n",
     vecp->ptr->dbl(0), vecp->ptr->dbl(1), vecp->ptr->dbl(2),
     vecp->t->time(0).getISOstring().c_str());
	 fflush(stdout);
  return;
}



MAT_INV_RET mat_inv(double a[3][3], double inverse[3][3]){

double det;


det = determinant(a);
if (fabs(det) < MAT_INV_EPS) return MAT_INV_SINGULAR;

inverse[0][0] =  ( a[1][1]*a[2][2] - a[1][2]*a[2][1] )/det;
inverse[0][1] = -( a[0][1]*a[2][2] - a[0][2]*a[2][1] )/det;
inverse[0][2] =  ( a[0][1]*a[1][2] - a[0][2]*a[1][1] )/det;
inverse[1][0] = -( a[1][0]*a[2][2] - a[1][2]*a[2][0] )/det;
inverse[1][1] =  ( a[0][0]*a[2][2] - a[0][2]*a[2][0] )/det;
inverse[1][2] = -( a[0][0]*a[1][2] - a[0][2]*a[1][0] )/det;
inverse[2][0] =  ( a[1][0]*a[2][1] - a[1][1]*a[2][0] )/det;
inverse[2][1] = -( a[0][0]*a[2][1] - a[0][1]*a[2][0] )/det;
inverse[2][2] =  ( a[0][0]*a[1][1] - a[0][1]*a[1][0] )/det;

return MAT_INV_OK;


}



double determinant(double m[3][3]){

double d;

d  = m[0][0]*m[1][1]*m[2][2] + m[0][1]*m[1][2]*m[2][0] +m [0][2]*m[1][0]*m[2][1];
d -= m[2][0]*m[1][1]*m[0][2] + m[2][1]*m[1][2]*m[0][0] + m[2][2]*m[1][0]*m[0][1];

return d;
}

void multiply_mat(double a[3][3],double b[3][3], double prod[3][3]){
int i,j,k;

for (i=0; i<3; i++){
  for (j=0; j<3; j++){
    for (prod[i][j]=0.0,k=0; k<3; k++) prod[i][j]+=a[i][k]*b[k][j];
  }
}

return;

}

void multiply_mat_vec(double a[3][3],double b[3], double prod[3]){
int i,j;

for (i=0; i<3; i++){
    for (prod[i]=0.0,j=0; j<3; j++) prod[i]+=a[i][j]*b[j];
}


return;

}

void print_mat(double a[3][3]){

int i,j;

for (i=0; i<3; i++){
   for (j=0; j<3; j++) printf("%12.9f ", a[i][j]);
   printf("\n");
   fflush(stdout);
}
return;
}


// -------------------------------------------------------------------
/*!
  \brief Parses a text and converts text to a 4 element double array
*/
// -------------------------------------------------------------------
DvObject_var str2Vec(DvString s)
{
    DvObject_var v = new DvObject(0.0, DIM4);
    s.strip('[');
    s.strip(']');
    s=s.trimmed(); // remove surrounding white space
    v->dbl(0) = s.before(' ').toDouble();
	s=s.after(' ');
	s=s.trimmed();
    v->dbl(1) = s.before(' ').toDouble();
	s=s.after(' ');
	s=s.trimmed();
    v->dbl(2) = s.before(' ').toDouble();
	s=s.after(' ');
	s=s.trimmed();
    v->dbl(3) = s.toDouble();
    return(v);
}





// -------------------------------------------------------------------
/*!
  \brief CTA calculations
  \return a CTA normal and a velocity for each of the 4 SC
  \param A - 3 x 3 matrix with separation distances
  \param B - 3 x 1 vector with relative timing
  \param taustring - string with the 4 Tau values
  \param vn - 4 x 1 matrix with normal velocities for each of the 4 SC
  \param d_cta - 4 x 1 matrix with thicknesses for each of the 4 SC
  \param n_cva - CVA normal - used for combined method
  \param v_cta - velocity curve from CTA
  \param v_ctam - velocity curve from combined CTA, CVA
  \param tidr - time (used to set timetags on velocity curves)

  Calculates a CTA (Constant Thickness Approach) boundary normal, normal
  velcoity and thickness. See Haaland et al, Annales Geophysicae, no 4, 2004
  for details.
*/
//
// SEH  May, 2004 
//
// -------------------------------------------------------------------
DvObject_var CTA_calc(double dr[4][3], double dt[4], DvString taustring, 
		       DvObject_var &vn_cta, DvObject_var &d_cta, 
		       DvObject_var &vn_com, DvObject_var &d_com, DvObject_var &n_com, 
		       DvObject_var &vn_cva, DvObject_var &d_cva, DvObject_var &n_cva, 
		       DvObject_var &v_cva_curve, DvObject_var &v_cta_curve, DvObject_var &v_com_curve, 
		       DvTime tidr, DvObject_var &tau)
{
  int i,j;
  DvObject_var n_cta =  new DvObject(0.0, DIM3);
  DvObject_var vn_instant =  new DvObject(0.0, DIM4); 
  double tc[4] = {dt[0],dt[1],dt[2],dt[3]};
  double A[3][3];
  double B[3];
  int SC_id[4] = {1,2,3,4};
  char outstr[200];

   tau =  str2Vec(taustring);

if(QuSlotDebug()){
  printf(" ---------------------------------------------------------------------------\n");
  printf("SCID before : %d  %d  %d  %d\n",SC_id[0], SC_id[1], SC_id[2], SC_id[3]);
  printf("tc Before   : %10.2f  %10.2f  %10.2f  %10.2f\nR :\n", tc[0],tc[1],tc[2],tc[3]);
  for (i=0; i<3; i++) printf("%10.2f  %10.2f  %10.2f  %10.2f\n",dr[0][i],dr[1][i], dr[2][i], dr[3][i]);
  printf("Tau before  : %10.2f  %10.2f  %10.2f  %10.2f\n\n", tau->dbl(0), tau->dbl(1), tau->dbl(2),tau->dbl(3));
  fflush(stdout);
}
 
  //  -- step 1 - sort in ascending order of crossing t0=0, t1 < t2 < t3 -----
 

  double tmp;
  double tmpa[3];
  long tmpID;
  for (i=1; i<4; i++) {
    for (j=3; j>=i; j--) {
      printf("Sort : j= %d\n",j);
	  fflush(stdout);
      if (tc[j-1] > tc[j]) {
	tmp = tc[j];
	tc[j] = tc[j-1];
	tc[j-1] = tmp;

       tmp=tau->dbl(j);
       tau->dbl(j) = tau->dbl(j-1);
       tau->dbl(j-1) = tmp;
	
       tmpID =  SC_id[j];
       SC_id[j] =  SC_id[j-1];    
       SC_id[j-1] = tmpID;
       
       for (int k=0; k<3; k++) {
	 	 tmpa[k] = dr[j][k];
	 	 dr[j][k] = dr[j-1][k];
	 	 dr[j-1][k] = tmpa[k];
       }
      }
    }
  }


if(QuSlotDebug()){
  printf("SCID after : %d  %d  %d  %d\n",SC_id[0], SC_id[1], SC_id[2], SC_id[3]);
  printf("tc After   : %10.2f  %10.2f  %10.2f  %10.2f\nR:\n", tc[0],tc[1],tc[2],tc[3]);
  for (i=0; i<3; i++) printf("%10.2f  %10.2f  %10.2f  %10.2f\n",dr[0][i],dr[1][i], dr[2][i], dr[3][i]);
  printf("tau after  : %10.2f  %10.2f  %10.2f  %10.2f\n", tau->dbl(0), tau->dbl(1), tau->dbl(2),tau->dbl(3));
  printf(" -----------------------------------------------------------------------------------\n\n");
  fflush(stdout);
}
 
  sprintf(outstr,"SC_id :  %10d    %10d    %10d    %10d\n", 
	  SC_id[0], SC_id[1], SC_id[2], SC_id[3]);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"tc     :   %10.2f  %10.2f  %10.2f  %10.2f\n", 
	  tc[0], tc[1], tc[2], tc[3]);
  QplugAppendTextDisplay(outstr);
  sprintf(outstr,"2Tau  :  %10.2f  %10.2f  %10.2f  %10.2f\n---------------------------------------------------------", 
	  tau->dbl(0), tau->dbl(1), tau->dbl(2), tau->dbl(3));
  QplugAppendTextDisplay(outstr);
 

  DvObject_var overtau = new DvObject(0.0, DIM4);
  DvObject_var Avec  = new DvObject(0.0, DIM4);
  DvObject_var TI  = new DvObject(1.0, DIM44);
  DvObject_var invTI  = new DvObject(0.0, DIM44);
  DvObject_var RdotM =  new DvObject(0.0, DIM3);
  DvObject_var R =  new DvObject(0.0, DIM33);
  DvObject_var invR =  new DvObject(0.0, DIM33);
  DvObject_var m =  new DvObject(0.0, DIM3);
  DvObject_var n =  new DvObject(0.0, DIM3);

  // -- step 2 fix separation R matrix and invert this ----

  for (i=1; i<4; i++) 
    for (j=0; j<3; j++) R->dbl(0,i-1,j) = dr[i][j] - dr[0][j];  // R = transposed of A
  
  for (i=4; i>=0; i--) tc[i] -= tc[0];   
  printf("tc After2   : %10.2f  %10.2f  %10.2f  %10.2f\nR:\n", tc[0],tc[1],tc[2],tc[3]);
  fflush(stdout);

  invR = R->inverse();


  //  -s step 3 fill in TI matrix and invert ------


  for (j=0; j<4; j++){
    TI->dbl(0,1,j) = tc[j];
    TI->dbl(0,2,j) = pow(tc[j],2.0) + pow(tau->dbl(j),2.0)/3.0; 
    TI->dbl(0,3,j) = pow(tc[j],3) + (tc[j]* pow(tau->dbl(j),2)); 
  }
     
  invTI = TI->inverse();
  for (i=0; i<4; i++) overtau->dbl(i) = 1.0/(2.0*tau->dbl(i));


  Avec = overtau * invTI;

  // --- calculate R dot N -- se eq (4) in Haaland et al, Ang 2, 2004 ---

  for (i=0; i<3; i++){
    RdotM->dbl(i) = Avec->dbl(0)*tc[i+1] + 
                  Avec->dbl(1)*(pow(tc[i+1],2)/2.0) + 
                  Avec->dbl(2)*(pow(tc[i+1],3)/3.0) + 
		  Avec->dbl(3)*(pow(tc[i+1],4)/4.0);
  }

  for (i=0; i<3; i++) 
    for (j=0; j<3; j++) 
      m->dbl(i) += invR->dbl(0,i,j) * RdotM->dbl(j);

  double sign = (m->dbl(0) < 0 ) ? -1 : 1;  

  double d = 1.0/vecsum(m);
  for (i=0; i<3; i++) 
    n->dbl(i) = sign * m->dbl(i) * d;

  vn_instant->dbl(0) = d * Avec->dbl(0);
  vn_cta->dbl(0)  =  sign * (vn_instant->dbl(0) + d*Avec->dbl(2)*pow(tau->dbl(0),2)/3.);
  for (j=1; j<4; j++) {
    vn_instant->dbl(j) = d * (Avec->dbl(0) + Avec->dbl(1)*tc[j] + Avec->dbl(2)*pow(tc[j],2) + Avec->dbl(3)* pow(tc[j],3));
    vn_cta->dbl(j)  = sign * (vn_instant->dbl(j) + d*Avec->dbl(2)*pow(tau->dbl(j),2)/3. + d*Avec->dbl(3)*tc[j]*pow(tau->dbl(j),2));
  }


if(QuSlotDebug()){
  printf("TI = \n");
  fflush(stdout);
  for (j=0; j<4; j++){
    printf("%10.2f  %10.2f  %10.2f  %10.2f\n",
	   TI->dbl(0,0,j),
 	   TI->dbl(0,1,j),
 	   TI->dbl(0,2,j),
 	   TI->dbl(0,3,j));

  }

  printf("invTI = \n");
  for (j=0; j<4; j++){
    printf("%10.4f  %10.4f  %10.4f  %10.4f\n",
	   invTI->dbl(0,0,j),
 	   invTI->dbl(0,1,j),
 	   invTI->dbl(0,2,j),
 	   invTI->dbl(0,3,j) );
  }

  printf("R = \n");
  for (j=0; j<3; j++){
    printf("         %10.2f  %10.2f  %10.2f\n",
	   R->dbl(0,0,j),
 	   R->dbl(0,1,j),
 	   R->dbl(0,2,j) );
  }

  printf("invR1R2 = \n");
  for (j=0; j<3; j++){
    printf("         %10.6f  %10.6f  %10.6f\n",
	   invR->dbl(0,0,j),
 	   invR->dbl(0,1,j),
 	   invR->dbl(0,2,j) );
  }

  printf("OverTau_CTA = \n");
  for (j=0; j<4; j++) printf("%10.4f\n",overtau->dbl(j));
  printf("\n");

  printf("Avec_CTA = \n");
  for (j=0; j<4; j++) printf("%12.8f\n",Avec->dbl(j));
  printf("\n");

  printf("RdotM_CTA = \n");
  for (j=0; j<3; j++) printf("%12.8f\n",RdotM->dbl(j));
  printf("\n");

  printf("m_CTA = \n");
  for (j=0; j<3; j++) printf("%12.8f\n",m->dbl(j));
  printf("\n");

  printf("d_CTA = %10.4f\n",d);

  printf("n_CTA = \n");
  for (j=0; j<3; j++) printf("%12.8f\n",n->dbl(j));
  printf("\n");

  printf("V_CTA = \n");
  for (j=0; j<4; j++) printf("%12.8f\n",vn_instant->dbl(j));
  printf("\n");

  printf("VaveCTA = \n");
  for (j=0; j<4; j++) printf("%12.8f\n",vn_cta->dbl(j));
  printf("\n");
  fflush(stdout);
}

  for (i=0; i<4; i++) d_cta->dbl(i) = d;

  // --- velocit curves - start a few steps before, and end a few steps later

  int Ntags = 120;
  double tstep = (fabs(tc[3])) / (double) (Ntags-20);
  double t_curvestart = -10*tstep;
  double t = t_curvestart;
  
  DvObject_var tt = new DvObject(tidr+t, Ntags);
  tt->assign_regular(tstep );  

  for (i=0; i<Ntags; i++) {
    v_cta_curve->dbl(i) =     sign * d * (Avec->dbl(0) + Avec->dbl(1)*t + Avec->dbl(2)*pow(t,2) + Avec->dbl(3)* pow(t,3));
    v_cva_curve->dbl(i) =     vn_cva->dbl();
    t+= tstep;
  }

  // check that d is really constant ---

//   double di = 0;
//   double ti[4] = {0.0, B[0], B[1], B[2]};
//   for (i=0; i<4; i++) {
//     di = 2*(*tau)[i] * ((*Avec)[0] + (*Avec)[1]*ti[i] + (*Avec)[2]*pow(ti[i],2) + (*Avec)[3]* pow(ti[i],3) + ((*Avec)[2]*pow((*tau)[i],2))/3.0 + (*Avec)[3]*ti[i]*pow((*tau)[i],2));
//     printf("i = %d, di = %f\n",i,di);
//   }

  printf("Sign = %f\n",sign);
  fflush(stdout);
  v_cta_curve->change_xref(DEPEND_0, tt);
  v_cva_curve->change_xref(DEPEND_0, tt);


  // -------------------------------------------------------------------------------------------------
  //
  //      combined CTA/CVA as proposed by BS, 26 May, 2004 
  //
  // -------------------------------------------------------------------------------------------------


  //  double theta = acos((*n_cva)[0] * (*n)[0] + (*n_cva)[1] * (*n)[1] +(*n_cva)[2] * (*n)[2]);
  
  for (i=0; i<3; i++) n_com->dbl(i) = (n->dbl(i) + n_cva->dbl(i)) / 2.0;
  double normlen=0.00;
    for (i=0; i<3; i++) normlen += n_com->dbl(i) * n_com->dbl(i);
    for (i=0; i<3; i++) n_com->dbl(i) /= sqrt(normlen);

  double k[4] = {0.0,  0.0,  0.0,  0.0};
  double aveTau=0.0, aveU=0.00, u[4];
  double ti[4] = {tc[0], tc[1], tc[2], tc[3]};

  for (i=0; i<4; i++) {
    u[i] = tau->dbl(i) *  (1.0 
	       - (2.0 * ti[i]*(1.0/ti[1] + 1.0/ti[2] + 1.0/ti[3]))
	       + (3.0 * (pow(ti[i],2) + pow(tau->dbl(i),2)/3.0) * (1.0/(ti[1]*ti[2]) + 1.0/(ti[1]*ti[3]) + 1.0/(ti[2]*ti[3]))) 
	       - (4.0 * (pow(ti[i],3) + ti[i]*pow(tau->dbl(i),2)) / (ti[1]*ti[2]*ti[3])));
    aveU += 0.25 * u[i];
    aveTau += 0.25 * tau->dbl(i);   
    printf("i = %d,  ui= %f\n",i,u[i]);
	fflush(stdout);
  }    

  double ave_Tau3     = 0.00;
  double ave_t_Tau    = 0.00;
  double ave_t_Tau3   = 0.00;
  double ave_t2_Tau   = 0.00;
  double ave_t3_Tau   = 0.00;
  double ave_U_Tau    = 0.00;
  double ave_U_Tau3   = 0.00;
  double ave_U_t_Tau  = 0.00;
  double ave_U_t_Tau3 = 0.00;
  double ave_U_t2_Tau = 0.00;
  double ave_U_t3_Tau = 0.00;

  for (i=0; i<4; i++) {
    ave_Tau3      += 0.25 * (pow(tau->dbl(i),3.0));
    ave_t_Tau     += 0.25 * (ti[i] * tau->dbl(i));
    ave_t2_Tau    += 0.25 * (pow(ti[i],2) * tau->dbl(i));
    ave_t3_Tau    += 0.25 * (pow(ti[i],3) * tau->dbl(i));
    ave_t_Tau3    += 0.25 * (ti[i] * pow(tau->dbl(i),3));
    ave_U_Tau     += 0.25 * (u[i] * tau->dbl(i));
    ave_U_Tau3    += 0.25 * (u[i] * pow(tau->dbl(i),3));
    ave_U_t_Tau   += 0.25 * (u[i] * ti[i] * tau->dbl(i));
    ave_U_t_Tau3  += 0.25 * (u[i] * ti[i] * pow(tau->dbl(i),3));
    ave_U_t2_Tau  += 0.25 * (u[i] * pow(ti[i],2) * tau->dbl(i));
    ave_U_t3_Tau  += 0.25 * (u[i] * pow(ti[i],3) * tau->dbl(i));
  }
  
  printf(" ave_Tau3     = %f\n", ave_Tau3 );
  printf(" ave_t_Tau    = %f\n", ave_t_Tau );
  printf(" ave_t_Tau3   = %f\n", ave_t_Tau3 );
  printf(" ave_U_Tau    = %f\n", ave_U_Tau );
  printf(" ave_U_Tau3   = %f\n", ave_U_Tau3 );
  printf(" ave_U_t_Tau  = %f\n", ave_U_t_Tau );
  printf(" ave_U_t2_Tau = %f\n", ave_U_t2_Tau );
  printf(" ave_U_t3_Tau = %f\n", ave_U_t3_Tau );

  k[0] = 1.0/aveU * (ave_U_Tau - aveU * aveTau);
  k[1] = 1.0/aveU * (ave_U_t_Tau - aveU * ave_t_Tau);
  k[2] = 1.0/aveU * (ave_U_t2_Tau - aveU * ave_t2_Tau + ave_U_Tau3/3.0 - (aveU * ave_Tau3)/3.0);
  k[3] = 1.0/aveU * (ave_U_t3_Tau - aveU * ave_t3_Tau + ave_U_t_Tau3 - aveU * ave_t_Tau3);

  printf("Avetau = %f\n",aveTau);
  printf("AveU = %f\n",aveU);
  printf("k = \n");
  for (j=0; j<4; j++) printf("%15.10f;\n",k[j]);
  printf("\n");
  fflush(stdout);

  DvObject_var MTI  = new DvObject(0.0, DIM44);
  DvObject_var invMTI  = new DvObject(0.0, DIM44);
  DvObject_var RdotN =  new DvObject(0.0, DIM4);


  for (j=0; j<3; j++) {
    MTI->dbl(0,0,j) = ti[j+1];
    MTI->dbl(0,1,j) = pow(ti[j+1],2) / 2.0;
    MTI->dbl(0,2,j) = pow(ti[j+1],3) / 3.0;
    MTI->dbl(0,3,j) = pow(ti[j+1],4) / 4.0;
  }

  for (i=0; i<4; i++) MTI->dbl(0,i,3) = k[i];

  invMTI = MTI->inverse();


  // --- calculate 4 elem R dot N --------

  for (j=0; j<3; j++){
    RdotN->dbl(j) = R->dbl(0,j,0)* n_com->dbl(0) +  R->dbl(0,j,1)* n_com->dbl(1) + R->dbl(0,j,2)*n_com->dbl(2);
  }
   
  printf("Rdot n = [%g  %g  %g]\n",RdotN->dbl(0), RdotN->dbl(1),RdotN->dbl(2) ) ;
fflush(stdout);
  Avec = RdotN * invMTI;


  for (i=0; i<4; i++) {
    d_com->dbl(i) = sign * 2 * tau->dbl(i) * (Avec->dbl(0) + Avec->dbl(1)*ti[i] + Avec->dbl(2)*pow(ti[i],2) + Avec->dbl(3)* pow(ti[i],3) + (Avec->dbl(2)*pow(tau->dbl(i),2))/3.0 + Avec->dbl(3)*ti[i]*pow(tau->dbl(i),2));
    printf("i = %d, di = %f\n", i, d_com->dbl(i));
	fflush(stdout);
    d_cva->dbl(i) = sign * 2.0*tau->dbl(i) * vn_cva->dbl();
  }

  // --- The 4 V_com velocitiese ----
  vn_instant->dbl(0) = Avec->dbl(0);
  vn_com->dbl(0)  =  (vn_instant->dbl(0) + Avec->dbl(2)*pow(tau->dbl(0),2)/3.);
  for (j=1; j<4; j++) {
    vn_instant->dbl(j) = (Avec->dbl(0) + Avec->dbl(1)*tc[j-1] + Avec->dbl(2)*pow(tc[j-1],2) + Avec->dbl(3)* pow(tc[j-1],3));
    vn_com->dbl(j)  = (vn_instant->dbl(j) + Avec->dbl(2)*pow(tau->dbl(j),2)/3. + Avec->dbl(3)*tc[j-1]*pow(tau->dbl(j),2));
  }


if(QuSlotDebug()){
  printf("MTI = \n");
  for (j=0; j<4; j++){
    printf("%14.4f  %14.4f  %14.4f  %14.4f\n",
	   MTI->dbl(0,0,j),
 	   MTI->dbl(0,1,j),
 	   MTI->dbl(0,2,j),
 	   MTI->dbl(0,3,j) );
  }

  printf("invMTI = \n");
  for (j=0; j<4; j++){
    printf("%14.12f  %14.12f  %14.12f  %14.12f\n",
	   invMTI->dbl(0,0,j),
 	   invMTI->dbl(0,1,j),
 	   invMTI->dbl(0,2,j),
 	   invMTI->dbl(0,3,j) );
  }
  printf("AvecCOM = \n");
  for (j=0; j<4; j++) printf("%14.10f;\n",Avec->dbl(j));
  printf("\n");

  printf("n_combined = \n");
  for (j=0; j<3; j++) printf("%14.10f\n",n_com->dbl(j));
  printf("\n");
  fflush(stdout);
}
  
  // --- V_com curve ----
  t = t_curvestart;
  for (i=0; i<Ntags; i++) {
    v_com_curve->dbl(i) =    (Avec->dbl(0) + Avec->dbl(1)*t + Avec->dbl(2)*pow(t,2) + Avec->dbl(3)* pow(t,3));
    t+= tstep;
  }
  tt->change_xref(DEPEND_0, v_com_curve);
  printf("at end : tau 0 = %f\n",tau->dbl(0) );
  fflush(stdout);

  // ---- output to /tmp/qsas_cvacta.txt ----
  FILE *f2;

    DvObject_var tt_out = v_cta_curve->getTimeTags();

    QplugAppendTextDisplay("Log file written to /tmp/qsas_cva_cta.txt");
    f2=fopen("/tmp/qsas_cva_cta.txt","w"); 
    if(f2 == NULL) f2 = fopen("/dev/stdout", "w");

    QDateTime today = QDateTime::currentDateTime();
    QString logTxt;
    logTxt += "#QSAS CVA_CTA_COM run : "+today.toString() + "\n#\n";
    sprintf(outstr,"#SC_id   %10d    %10d    %10d    %10d\n",  SC_id[0], SC_id[1], SC_id[2], SC_id[3]);
    logTxt += QString(outstr);
    sprintf(outstr,"#tc      %10.2f  %10.2f  %10.2f  %10.2f\n", tc[0], tc[1], tc[2], tc[3]);
    logTxt += QString(outstr);
    sprintf(outstr,"#tstep      %10.2f\n", tstep);
    logTxt += QString(outstr);
    sprintf(outstr,"#2Tau    %10.2f  %10.2f  %10.2f  %10.2f\n", tau->dbl(0), tau->dbl(1), tau->dbl(2), tau->dbl(3));
    logTxt += QString(outstr);
    fprintf(f2,"%s", logTxt.toLatin1().data());
    fprintf(f2,"#Time = %s\n",tt_out->time(0).getISOstring().c_str());
    fprintf(f2,"# \n");
    for (i=0; i<Ntags; i++)  {
      fprintf(f2, "%s  %10.4f  %10.4f  %10.4f \n", tt_out->time(i).getISOstring().c_str(),
	     v_cva_curve->dbl(i),v_cta_curve->dbl(i),v_com_curve->dbl(i) );
    }
    fclose(f2);
  
  return (n);

}


// -------------------------------------------------------------------
/*!
  \brief returns the vector sum (length) of a vector
  \returns length of V
  \param V - 3 component vector
*/
// -------------------------------------------------------------------
double vecsum(DvObject_var V)
{
  double sum=0.0;
  
  for (int i=0; i<3; i++) sum += V->dbl(i) * V->dbl(i);

  return(sqrt(sum));
}

