// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : EDI filter routine
// 
// This plugin takes an EDI vector (V or E) as inputs and
// filters it according to the settings in the given filter
// specifications file
//
// This file 'EDI_filter.cc' contains the QSAS-wrapper + minor
// error checking. Reading inputs from the filter settings
// file is handled by code in 'readfilterfile'. The actual
// filtering/sorting code can be found in 'DataOk'
//
// Inputs :
// 	edi_vector     	- vector time series with either E or V from EDI
//      edi_status      - the EDI status matrix
//      edi_chisqr      - EDI reduced ChiSquare results
//
// Outputs
//	edi_filtered    - filtered version of above
//
//
// SEH, 23 Oct, 2001
//   27. Oct 01 SEH : RedChiSqr is now floating point number
//   17. Nov 01 SEH : Output object named after input objects 
//   27. Nov 01 SEH : corrected naming in *metstr
//   29. Nov 01 SEH : Frame to vector>gse_xyz
//   02. Mar 02 SEH : New method RMT : expanded *,ethstr
//                    (see also DataOk + defs in EDI_filter.h)
//   08. Apr 02 SEH : Added support for TriErr, AngErr, Triout as
//                    signed bytes
//   28. Nov 03 SEH : QSAS 2 version; Changed header + 'external'
//   31. Oct 06 SEH : replaced 'Qu_is3Vector()' with cast and .is_nil() check
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "EDI_filter.h"


// comment out for QSAS 1.xx 
extern "C" QplugReturnStatus  DvEDI_filter  (vector <DvObject_var>&, vector <DvObject_var>&);

// ----- Prototypes ----

 int DataOk(sbytes stat, sbytes settings);
 sbytes readfilterfile(const char *filename);
using namespace QSAS;

 QplugReturnStatus  DvEDI_filter  ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
 {

 DvObject_var  edi_vector;
 DvObject_var  edi_temp;
 DvObject_var  edi_filtered;
 DvObject_var  edi_status;
 DvObject_var  edi_chisqr;
 DvString      filename_input;
 char       txtbuf[200];
 sbytes     stat,settings;
 vec        EDI;
 long       n_vecs,n_stat,n_chis,n_left=0,i,j;
 int        REMOVE=0, ChiCheck=1;
 double     chisqr,fillval = -1.0E31;
 char       *str_out,*str_add,str_tot;
                        
// --- get EDI vector -----

 edi_vector = inputs[0];
 DvObject_var tt_in = edi_vector->getTimeTags();

 if ( edi_vector.is_nil() || !edi_vector->isThreeVector()) {	
   QplugAppendTextDisplay ( "ERROR - invalid EDI vector. Input not a Vector Time Series!\n");
   return QPLUG_FAILURE; 
 }        
 
 

 n_vecs = edi_vector->seqSize();
 DvObject_var tt_temp = new DvObject(DvTime(), n_vecs);


// --- get Status matrix - a n x 7 matrix of bytes -----

 edi_status = inputs[1];
 DvString fieldname = edi_status->getXrefText("FIELDNAM");
 if ( !fieldname.contains("EDI status")){
   QplugAppendTextDisplay ( "ERROR - invalid Status words! Input is not EDI status word \n");
   return QPLUG_FAILURE; 
 }

 n_stat = edi_status->seqSize();
 printf("Testing ChiSqr ...\n");
 fflush(stdout);

// --- get Reduced ChiSqr scalar TS -----

 edi_chisqr = inputs[2];
 if (!edi_chisqr.is_nil() && edi_chisqr->getXrefText("FIELDNAM") == "EDI Reduced Chi-squared") {
   n_chis = edi_chisqr->seqSize();
   printf("ChiSqr exists...\n");
   fflush(stdout);
 } 
 else {
   QplugAppendTextDisplay ( "WARNING - invalid or missing ChiSqr. Ignoring this filter \n");
   ChiCheck = 0;
   n_chis = n_vecs;
   printf("ChiSqr does not exists...\n");
   fflush(stdout);
 }

 if (n_vecs != n_stat && n_vecs != n_chis) {
   QplugAppendTextDisplay ( "ERROR - different number of records in VEC, Status and ChiSqr !\n");
   return QPLUG_FAILURE; 
 }        


filename_input = inputs[3]->asStr();
if( filename_input.empty() ) {
  QplugAppendTextDisplay("ERROR - no filter file specified\n");
  return QPLUG_FAILURE;
}


// ---- Read filter definition file ----

 settings.quality = GOOD_CAUT_BAD;
 settings.method  = TRI_PMT_SMT;
 settings.ambigu  = YES_NO ;
 settings.angerr  = 255;
 settings.trierr  = 200;
 settings.triout  = 100;
 settings.redchi  = 1.0E6;

 settings = readfilterfile(filename_input.c_str());
 {
   printf(" Quality       :%d \n",settings.quality);
   printf(" Method        :%d \n",settings.method);
   printf(" Ambigu        :%d \n",settings.ambigu);
   printf(" TriErr        :%d \n",settings.trierr);
   printf(" AngErr        :%d \n",settings.angerr);
   printf(" TriOut        :%d \n",settings.triout);
   printf(" Chisqr        :%5.2f \n",settings.redchi);
   printf(" Filterfile OK :%d \n",settings.OK);
   fflush(stdout);
 }

 if (!settings.OK) {
   QplugAppendTextDisplay("\nWARNING >> Filter file not found - using defaults\n");
 }
 else {
   sprintf(txtbuf,"\nFilter file   : %s\n",filename_input.c_str());
   QplugAppendTextDisplay(txtbuf);
 }


/* ----------------------------------------  
     declare a temporary object to store the 
     filtered recs. Since we don't know the
     final lenght of the we have to allow
     place for all records
-------------------------------------------*/


 edi_temp = new DvObject(0.0, DIM3, n_vecs);


// ---- Scan through the records -------

 for(i=0; i<n_vecs; i++) {  

   stat.quality = edi_status->asInt(i,0) &   3;
   stat.method  = edi_status->asInt(i,3) &  96;
   stat.ambigu  = edi_status->asInt(i,3) & 128;
   stat.trierr  = edi_status->asInt(i,5) & 255;                // update 08 Apr 2002
   stat.angerr  = edi_status->asInt(i,6) & 255;                //        "
   stat.triout  = edi_status->asInt(i,4) & 255;                //        "
   stat.redchi  = -1.0;

   if (ChiCheck) {
     chisqr = edi_chisqr->asDouble(i);
     if (chisqr >= 0.0 ) stat.redchi  = (float) edi_chisqr->asDouble(i);
   }

   if (VERBOSE_EDI)
   {
	printf("Rec : %5ld  ",i); 
	fflush(stdout);
   }
   REMOVE = DataOk(stat,settings);

   if (!REMOVE) {
     for (j=0; j<3; j++) edi_temp->dbl(n_left,j) = edi_vector->dbl(i,j);
     tt_temp->time(n_left) = tt_in->time(i);
     n_left++;
   } 
 }

 sprintf(txtbuf,"Number of vectors input  : %ld \n",n_vecs);
 QplugAppendTextDisplay ( txtbuf);
 sprintf(txtbuf,"Number of vectors output   : %ld \n\n",n_left);
 QplugAppendTextDisplay ( txtbuf);

 if (n_left == 0) {	
   QplugAppendTextDisplay ( "ERROR >>  no vectors fit filters !\n");
   return QPLUG_FAILURE; 
 }        


 // ---- transfer the n_left records to output object ----


 DvObject_var tt_out = new DvObject(DvTime(), n_left);
 edi_filtered = new DvObject(0.0, DIM3, n_left);
 
 for (i=0; i<n_left; i++) {
   tt_out->time(i) = tt_temp->time(i);
   for (j=0; j<3; j++) edi_filtered->dbl(i,j) = edi_temp->dbl(i,j);
 }


 // ----- generate name of output object (SEH 17 Nov. 2001) ----------

 char str1[127]="";
 const char *qsetstr[7]  = {"BAD","CAUTBAD","CAUT","ALL","GOODBAD","GOODCAUT","GOOD" };
 const char *methstr[15] = {"SMT","PMTSMT","PMT","TRISMTPMT","TRISMT","TRIPMT","TRI",
                      "SMTRMT","PMTSMTRMT","PMTRMT","ALL","TRISMTRMT",
		      "TRIPMTRMT","TRIRMT","RMT"};
 const char *ambistr[3]  = {"ALL","NO","YES" };

 strncat(str1,edi_vector->getXrefText("ObjectName").c_str(),1);
 strcat(str1,"_edi_C");
 strncat(str1,edi_vector->getXrefText("SC_id").c_str(),1);
 strcat(str1,"_");
 strncat(str1, tt_out->time(0).getISOstring().c_str(),10); 	
 strcat(str1,"_");
 if (settings.quality < 7) strcat(str1,qsetstr[settings.quality]);
 strcat(str1,"_");
 if (settings.method < 15) strcat(str1,methstr[settings.method]);
 strcat(str1,"_");
 if (settings.ambigu < 3) strcat(str1,ambistr[settings.ambigu]);

 DvString slotname = GetSlotText(0);
 if (slotname.empty()) SetSlotText(0, str1);



 // ----- transfer timetags, copy attributes and push back ------

 edi_filtered->copy_xrefs_from(edi_vector);
 edi_filtered->change_xref(DEPEND_0,tt_out);

 outputs.push_back(edi_filtered);

 QplugAppendTextDisplay ("EDI_filter>> data object created - exiting successfully\n");

 return QPLUG_SUCCESS;
}




int DataOk(sbytes  stat, sbytes settings)
/* -----------------------------------------------------------------
  
  returns 1 if the EDI input data does not satisfy the
  given filter criteria.

  inputs
      stat - extracted status information from EDI cdf variable
      settings - filter options provided by the user

  SEH, 23 Oct 2001
       01 Nov 2001  (reduced ChiSqr now as float + break on Default
                     in switch statements)
       02 Mar 2002  (new method RMT - see mail from M.Foerster, 
                     2. Mar, 2002, 0923 UT)
--------------------------------------------------------------------*/

{

   int REMOVE=0;

   if (VERBOSE_EDI)
   {
	printf(" StatWord : %3d %3d %3d %3d %3d %3d %5.1f  ",stat.quality,stat.method,stat.ambigu,stat.trierr,stat.angerr,stat.triout,stat.redchi);
	fflush(stdout);
   }
   if (stat.quality == 3) {
     REMOVE = 1;
       if (VERBOSE_EDI)
	   {
		printf(" FillVal in QUALITY flag ");
		fflush(stdout);
	   }
   }

   if (!REMOVE) {
     switch (settings.quality) {
     case GOOD:
       if (stat.quality != 2) REMOVE = 1;
       if (VERBOSE_EDI) printf("Quality not GOOD ");
       break;
     case GOOD_CAUT:
       if (stat.quality == 0) REMOVE = 1;
       if (VERBOSE_EDI) printf("Quality not GOOD_CAUT ");
       break;
     case GOOD_BAD:
       if (stat.quality == 1) REMOVE = 1;
       if (VERBOSE_EDI) printf("Quality not GOOD_BAD ");
       break;
     case CAUT:
       if (stat.quality != 1) REMOVE = 1;
       if (VERBOSE_EDI) printf("Quality not CAUT ");
       break;
     case CAUT_BAD:
       if (stat.quality >= 2) REMOVE = 1;
       if (VERBOSE_EDI) printf("Quality not CAUT_BAD ");
       break;
     case BAD:
       if (stat.quality >= 1) REMOVE = 1;
       if (VERBOSE_EDI) printf("Quality not BAD ");
       break;
     default:
       REMOVE = 0;
       if (VERBOSE_EDI) printf("No Quality filtering \n");
       break;
     }
   }
   if(VERBOSE_EDI) fflush(stdout);
   // ----- check Method -----
   //
   // Updated 2. Mar 2002 SEH
   // 
   // stat.method can only attain one of the values
   // 0, 32, 64 or 128. We use the operatos != and == rather 
   // than >= and <= 
   // 

   if (!REMOVE) {
     switch (settings.method) {
     case TRI:
       if (stat.method != 0) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not TRI ");
       break;
     case TRI_PMT:
       if (stat.method != 0 && stat.method !=32) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not TRI_PMT ");
       break;
     case TRI_SMT:
       if (stat.method != 0 && stat.method != 64) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not TRI_SMT ");
       break;
    case PMT:
       if(stat.method != 32) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not PMT ");
       break;
     case PMT_SMT:
       if (stat.method != 32 && stat.method != 64) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not PMT_SMT ");
       break;
     case SMT:
       if (stat.method != 64) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not SMT ");
       break;
     case TRI_PMT_SMT:       // remove if RMT 
       if (stat.method == 96) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not TRI_SMT ");
       break;
     case RMT:              // remove if not RMT 
       if (stat.method != 96) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not RMT ");
       break;
     case TRI_RMT:          // i.e remove if SMT or PMT        
       if (stat.method != 0 && stat.method != 96) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not TRI_RMT ");
       break;
     case TRI_PMT_RMT:      // i.e remove if SMT 
       if (stat.method == 64) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not TRI_PMT_RMT ");
       break;
     case TRI_SMT_RMT:       // i.e remove if PMT      
       if (stat.method == 32) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not TRI_PMT_RMT ");
       break;
     case PMT_RMT:           // remove if TRI or SMT   
       if (stat.method == 0 || stat.method == 64) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not PMT_RMT ");
       break;
     case PMT_SMT_RMT:       // allow all but TRI 
       if (stat.method == 0) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not PMT_SMT_RMT ");
       break;
     case SMT_RMT:           // i.e. remove if TRI or PMT 
       if (stat.method == 0  || stat.method == 32) REMOVE = 1;
       if (VERBOSE_EDI) printf("Method not SMT_RMT ");
       break;

     default:
       REMOVE = 0;
       if (VERBOSE_EDI) printf("No Method filtering \n");
       break;
     }
	     if(VERBOSE_EDI) fflush(stdout);
   }

   // ----- check ambiguity ------

   if (!REMOVE) {
     switch(settings.ambigu) {
     case YES:
       if (stat.ambigu != 128) REMOVE = 1;
       if (VERBOSE_EDI) printf("Ambig not YES ");
       break;
     case NO:
       if (stat.ambigu !=0) REMOVE = 1;
       if (VERBOSE_EDI) printf("Ambig not NO ");
       break;
     default:
       REMOVE = 0;
       if (VERBOSE_EDI) printf("No Ambiguity filtering \n");
       break;
     }
	     if(VERBOSE_EDI) fflush(stdout);
   }
			       

   // ----- check angle error ------

   if(!REMOVE) {
     if(stat.angerr > settings.angerr && stat.angerr != 255) {
       REMOVE = 1;
       if (VERBOSE_EDI) printf(" to high ANGERR ");
     }
   }


   // ----- check magnitude error ------

   if(!REMOVE) {
     if ((stat.trierr > settings.trierr) && (stat.trierr != 255)) {
       REMOVE = 1;
       if (VERBOSE_EDI) printf(" to high MAGERR ");
     }
   }


   // ----- check triangulation outliers (triangulation only) -----

   if (!REMOVE && (stat.method == 0) && (stat.triout != 255)) {
     if (stat.triout > settings.triout) {
       REMOVE = 1;
       if (VERBOSE_EDI) printf(" to many TRI_OUTs ");
     }
   }


   // ----- check reduced chi-sq (triangulation only) -----

   if (!REMOVE && (stat.method == 0)) {
     if (stat.redchi > settings.redchi) {
       REMOVE = 1;
       if (VERBOSE_EDI) printf(" Rchi above limit ");
     }
   }
  
   if (REMOVE && stat.quality !=3 ) {
     if (VERBOSE_EDI) printf(" - record  deleted \n");
   }
   else {
     if (VERBOSE_EDI) printf("\n");
   }
    if(VERBOSE_EDI) fflush(stdout);
   return (REMOVE);
}    


sbytes readfilterfile(const char *filename)
/* ------------------------------------------------------------------------

  reads filter settings from file

  SEH 23 Oct 2001

 ------------------------------------------------------------------------*/
{

  int lineno=0, comment=1;
  FILE* f1;
  char buf[500];
  char *svalue,*token;
  sbytes stat;

  // ---- default values ----

  stat.quality = GOOD_CAUT_BAD;
  stat.method  = TRI_PMT_SMT;
  stat.ambigu  = YES_NO ;
  stat.angerr  = 100;
  stat.trierr  = 100;
  stat.triout  = 100;
  stat.redchi  = 100.0;
  stat.OK      = 0;                                           // No filterfile yet -> OK =0 

  f1 = fopen(filename,"r");
  //  printf("Filter settings file : %s\n",filename);

  token = (char*)" ";                                                // possible separators in input file

  if (f1 != NULL){
    while(fgets(buf,500,f1) != NULL) {
      comment = (buf[0] == '#' || buf[0] == '%' || buf[0] == ';' || buf[0] == '!');
      if (!comment) {
        for (svalue = strtok(buf,token); svalue != 0; svalue=strtok(0,token)) {
	  switch (lineno) {
	  case 0:
	    stat.quality = atoi(svalue);
	    break;
	  case 1:
	    stat.method = atoi(svalue);
	    break;
	  case 2:
	    stat.ambigu = atoi(svalue);
	    break;
	  case 3:
	    stat.angerr = atoi(svalue);
	    break;
	  case 4:
	    stat.triout = atoi(svalue);
	    break;
	  case 5:
	    stat.trierr = atoi(svalue);
	    break;
	  case 6:
	    stat.redchi = atof(svalue);
	    break;
	  }
	}
	lineno++;
      }
    }
    stat.OK = 1;                                             // Filter file found -> OK = 1
    fclose(f1);
  } 
  else {
    printf(" NO FILTER FILE - using defaults \n");
	fflush(stdout);
  }

  return(stat);

}

