// -----------------------------------------------------------------------------------------
//
// Performs a nested generic residue analysis
//
// SEH, Sep 16, 2003
// 
// Inputs :
//     B = magnetic field   (n x 3) 
//     V = plasma velocity  (n x 3) 
//     E = electric field   (n x 3)
//     P = pressure tensor  (n x 3 x 3)
//     Q = heat flux vec    (n x 3)
//     N = density          (n)
//     nestsize = number of points in data sets
//     min_nest = minimum nest size
//     constrain = 1 for Bn=0, 2 for (<V> - U)n etc - see below.
//     alpha = slope of walen correlation coeff (used for Alfvenic constrain)
//
// SEH, 16 Sep 2003
//      14 May 2005, added alpha - the Walenslope for constraining
//      31 May 2006, added MVA dB/dt
//
// -----------------------------------------------------------------------------------------

#include "GenericMinVar.h"
#include "GenericResidue.h"

int NestedAnalysis(DvObject_var &V_, DvObject_var &B_, DvObject_var &dB_, DvObject_var &P_, 
		   DvObject_var &E_, DvObject_var &N_, DvObject_var &H_,
		   DvObject_var &ttJoined,  DvObject_var &ExWeights,
		   int constrained, double alpha, int min_nest, DvString SCs_ID,
		   bool doType[10],int nestsize)
{

  const char *SCc_ID = SCs_ID.c_str();
  
  int i,j,k,nloops=0;
  DvObject_var  B;                                   // B = Magnetic field series
  DvObject_var  dB;                                  // dB = d/dt magnetic field series
  DvObject_var  V;                                   // V = Velocity series
  DvObject_var  E;                                   // E = Electric field
  DvObject_var  P;                                   // P = Pressure tensor
  DvObject_var  H;                                   // H = heat flux
  DvObject_var  N;                                   // N = Plasma density
  DvObject_var tt_out;                                           

  DvObject_var ProjMx  = new DvObject(0.0, DIM33);
  DvObject_var Q  = new DvObject(0.0, DIM33);
  DvObject_var Q_COM  = new DvObject(0.0, DIM33);
  DvObject_var Evec = new DvObject(0.0, DIM33);  
  DvObject_var EvecMVAB = new DvObject(0.0, DIM33);     
  DvObject_var Evalues = new DvObject(0.0, DIM3);
  DvObject_var Dummy =  new DvObject(0.0, DIM3);  
  DvObject_var U = new DvObject(0.0, DIM3);        
  char buf[255];
  double un=0.0, ucis=0.0, deltat=0.0,normfac;

  ttJoined->isRegular(deltat);

  int max_numnest = (nestsize - min_nest + 3) / 2;                               // number of nest to study

  printf("Nested : nestsize = %d\n",nestsize);
  printf("Nested : min_nest = %d\n",min_nest);
  printf("Nested : n loops = %d\n",max_numnest);
  fflush(stdout);

  QDateTime dt = QDateTime::currentDateTime();
    
  DvObject_var q_mvab  = new DvObject(0.0, DIM33, max_numnest);
  DvObject_var q_mvadb  = new DvObject(0.0, DIM33, max_numnest);
  DvObject_var q_mfr  = new DvObject(0.0, DIM33, max_numnest);
  DvObject_var q_mmr  = new DvObject(0.0, DIM33, max_numnest);
  DvObject_var q_mer  = new DvObject(0.0, DIM33, max_numnest);
  DvObject_var q_mter  = new DvObject(0.0, DIM33, max_numnest);
  DvObject_var q_mlmr  = new DvObject(0.0, DIM33, max_numnest);
  DvObject_var q_mvav  = new DvObject(0.0, DIM33, max_numnest);
  DvObject_var q_com  = new DvObject(0.0, DIM33, max_numnest);

  DvObject_var U_mva  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var U_mvadb  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var U_mmr  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var U_mfr  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var U_mer  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var U_mter  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var U_mlmr  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var U_mvav  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var U_com  = new DvObject(0.0, DIM3, max_numnest);

  DvObject_var B_nestave  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var V_nestave  = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var Va_nestave = new DvObject(0.0, DIM3, max_numnest);
  DvObject_var B_doubleave = new DvObject(0.0, DIM3);
  DvObject_var V_doubleave = new DvObject(0.0, DIM3);
  DvObject_var Va_doubleave = new DvObject(0.0, DIM3);

  DvObject_var Valf; //  this is initialized in Q_MVAB_calc

  //  ------ loop over nestsize ------

  DvTime t_start = ttJoined->time(0);
  DvTime t_stop  = ttJoined->time(nestsize-1);
  DvTime t_center = t_start + (t_stop - t_start) / 2.;
  DvEvent trange(t_start, t_stop);
    
  double wl=0.0, cc=0.0, weight_k=0.0, weight_sum=0.0;

  sprintf(buf,"#Date performed : %s\n#",dt.toString().toLatin1().data());
  QplugAppendTextDisplay(buf);
  sprintf(buf,"#nloops = %2d  initial t = %s <-- %s --> %s (%2d samples)\n", max_numnest,
        t_start.getISOstring().c_str(),t_center.getISOstring().c_str(),t_stop.getISOstring().c_str(),nestsize);
  QplugAppendTextDisplay(buf);
    
  //  while (nestsize > min_nest) {
  for (nloops=0; nloops<max_numnest; nloops++) {
      tt_out = ttJoined->subSequence(trange);
       
      sprintf(buf,"\n\n#loop = %2d of %2d  t = %s <-- %s --> %s (%2d samples)\n",
	    nloops+1,max_numnest, t_start.getISOstring().c_str(),t_center.getISOstring().c_str(),t_stop.getISOstring().c_str(),nestsize);
      printf("%s",buf);
      fflush(stdout);


      // ---- do MVAB to get the eigenvectors from unconstrained MVAB ------

      bool doV = V_.is_ok();
      bool doN = N_.is_ok();

      if (doV) V = V_->subSequence(trange);
      if (doN) N = N_->subSequence(trange);

      if (doType[0]) {
          printf("doing initial MVAB... doV = %d doN = %d\n",(int)doV, (int)doN);
          fflush(stdout);
          B = B_->subSequence(trange);
          nestsize = B->seqSize();
          Valf = new DvObject(0.0, DIM3, nestsize);  
          Q = Q_MVAB_calc(B,V,U,Valf,N, &wl, &cc, nestsize, false);
	
          GenericMinVar minvar(0,Q,ProjMx);                                               // unconstrained
          EvecMVAB = minvar.eigenvectors();
     
          printf("Initial MVAB done..\n");
          fflush(stdout);
      }
    

      // ---- 0,2 : do MVAB (HT if doType==2) with selected constrain ------

      if (doType[2]) {
          printf("doing MVAB..\n");
          fflush(stdout);
          Q = Q_MVAB_calc(B,V,U,Valf,N, &wl, &cc,nestsize, doType[2]);
          ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained);

          GenericMinVar minvar(constrained,Q,ProjMx);  
          Evec = minvar.eigenvectors();
          Evalues = minvar.eigenvalues(); 
          printf(" in nested MVAB\n");
          fflush(stdout);

          for (i=0; i<3; i++) {
              for (j=0; j<3; j++) {
                  q_mvab->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
              }
              U_mva->dbl(nloops,i) = U->dbl(i);
          }

          weight_k =  ExWeights->dbl(0)  * weight_k_calc(Evalues);
          printf("weight_MVAB = %10.4f\n",weight_k);
          fflush(stdout);
          weight_sum = weight_k; 
          for (i=0; i<3; i++) {
              for (j=0; j<3; j++) {
                  Q_COM->dbl(0,i,j) = weight_k * Q->dbl(0,i,j) / ( Q->dbl(0,0,0) + Q->dbl(0,1,1) + Q->dbl(0,2,2));         // NB not += here 
              }
              U_com->dbl(nloops,i) += weight_k * U->dbl(i);
          }

 
      // ---- do the double-averages -----------------

      printf("double avg...%d  %d \n", (int)V->seqSize(), (int)B->seqSize());
	  fflush(stdout);
      for (k=0; k<nestsize; k++) {
          for (i=0; i<3; i++) {      
              B_nestave->dbl(nloops,i) += B->dbl(k,i) / (double) nestsize;;
          }
      }

      printf("MVAB done...\n");
	  fflush(stdout);
    }



    // ------- 1 : do MVAdB/dt with this constrain --------

    if (doType[1]) {
      printf("doing MVABdBdt..\n");
	  fflush(stdout);
      dB = dB_->subSequence(trange);
      DvObject_var DummyValf = new DvObject(0.0, DIM3, nestsize);  
      
      Q = Q_MVAB_calc(dB,V,U,DummyValf,N, &wl, &cc,nestsize,doType[2]);
      ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained);
      
      GenericMinVar minvar(constrained, Q, ProjMx);  
	  Evec = minvar.eigenvectors();
	  Evalues = minvar.eigenvalues(); 
            
      
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              q_mvadb->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
          }
          U_mvadb->dbl(nloops,i) = U->dbl(i);
      }
      
      weight_k =  ExWeights->dbl(1)  * weight_k_calc(Evalues);
      printf("weight_MVAB = %10.4f\n",weight_k); fflush(stdout);
      weight_sum = weight_k; 
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              Q_COM->dbl(0,i,j) = weight_k * Q->dbl(0,i,j) / ( Q->dbl(0,0,0) + Q->dbl(0,1,1) + Q->dbl(0,2,2));         // NB not += here 
          }
          U_com->dbl(nloops,i) += weight_k * U->dbl(i);
      }
        
    }


    // ------- 3: do MFR with this constrain ----

    if (doType[3]) {
      printf("Doing MFR...\n");
	  fflush(stdout);
      V = V_->subSequence(trange);
      Q = Q_MFR_calc(B,V,U,nestsize);
      ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize,constrained);

      GenericMinVar minvar(constrained, Q, ProjMx);  
	  Evec = minvar.eigenvectors();
	  Evalues = minvar.eigenvalues(); 
 
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              q_mfr->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
          }
          U_mfr->dbl(nloops,i) = U->dbl(i);
      }

      weight_k =  ExWeights->dbl(2)  * weight_k_calc(Evalues);
      printf("weight_MFR = %10.4f\n",weight_k);
	  fflush(stdout);
      weight_sum += weight_k;
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              Q_COM->dbl(0,i,j) += weight_k * Q->dbl(0,i,j) / ( Q->dbl(0,0,0) + Q->dbl(0,1,1) + Q->dbl(0,2,2));
          }
          U_com->dbl(nloops,i) += weight_k * U->dbl(i);
      }
        
    }


    // ---- 4 : do MMR with selected constrain ------
 
    if (doType[4]) {
      Q = Q_MMR_calc(V, N, U, nestsize);
      ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained);

      GenericMinVar minvar(constrained, Q, ProjMx);  
	  Evec = minvar.eigenvectors();
	  Evalues = minvar.eigenvalues(); 
          
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              q_mmr->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
          }
          U_mmr->dbl(nloops,i) = U->dbl(i);
      }

      weight_k =  ExWeights->dbl(3) * weight_k_calc(Evalues);
      printf("weight_MMR = %10.4f\n",weight_k);
	  fflush(stdout);
      weight_sum += weight_k;
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              Q_COM->dbl(0,i,j) += weight_k * Q->dbl(0,i,j) / ( Q->dbl(0,0,0) + Q->dbl(0,1,1) + Q->dbl(0,2,2));
          }
          U_com->dbl(nloops,i) += weight_k * U->dbl(i);
      }

      // ---- do the double-averages for N (Va) -----------------

        for (k=0; k<nestsize; k++) {
            for (i=0; i<3; i++) {      
                Va_nestave->dbl(nloops,i) += (alpha*Valf->dbl(k,i) / (double) nestsize);
            }
        }
      
    }
            

    // ---- 5: do MLMR with selected constrain ------

    if (doType[5]) {
      printf("doing MLMR..\n");
	  fflush(stdout);
      P = P_->subSequence(trange);
      Q =  Q_MLMR_calc(V,B,P,N,U,nestsize);
      ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained);
	  GenericMinVar minvar(constrained, Q, ProjMx);  
	  Evec = minvar.eigenvectors();
	  Evalues = minvar.eigenvalues(); 

      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              q_mlmr->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
          }
          U_mlmr->dbl(nloops,i) = U->dbl(i);
      }

      weight_k =  ExWeights->dbl(4)  * weight_k_calc(Evalues);
      printf("weight_MLMR = %10.4f\n",weight_k);
	  fflush(stdout);
      weight_sum += weight_k;
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              Q_COM->dbl(0,i,j) += weight_k * Q->dbl(0,i,j) / ( Q->dbl(0,0,0) + Q->dbl(0,1,1) + Q->dbl(0,2,2));
          }
          U_com->dbl(nloops,i) += weight_k * U->dbl(i);
      }
    }


  // ---- 6: do MER with selected constrain ------
 
    if (doType[6]) {
      printf("doing MER..\n");
	  fflush(stdout);
      Q =  Q_MER_calc(V,B,P,N,U,nestsize);
      ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained);
	  GenericMinVar minvar(constrained, Q, ProjMx);  
	  Evec = minvar.eigenvectors();
	  Evalues = minvar.eigenvalues(); 
            
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              q_mer->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
          }
          U_mer->dbl(nloops,i) = U->dbl(i);
      }
      
      
      weight_k =  ExWeights->dbl(6)  * weight_k_calc(Evalues);
      weight_sum += weight_k;
      printf("weight_MER = %10.4f\n",weight_k);
	  fflush(stdout);
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              Q_COM->dbl(0,i,j) += weight_k * Q->dbl(0,i,j) / ( Q->dbl(0,0,0) + Q->dbl(0,1,1) + Q->dbl(0,2,2));
          }
          U_com->dbl(nloops,i) += weight_k * U->dbl(i);
        }
    }


    // ---- 7: do MTER with selected constrain ------

    if (doType[7]) {
     E = E_->subSequence(trange);
     H = H_->subSequence(trange);
     Q =  Q_MTER_calc(V,B,P,E,N,H,U,nestsize);
     ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained);

     GenericMinVar minvar(constrained, Q, ProjMx);  
	 Evec = minvar.eigenvectors();
	 Evalues = minvar.eigenvalues(); 
     
    
     for (i=0; i<3; i++) {
         for (j=0; j<3; j++) {
             q_mter->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
         }
         U_mter->dbl(nloops,i) = U->dbl(i);
     }
 
      weight_k =  ExWeights->dbl(5)  * weight_k_calc(Evalues);
      weight_sum += weight_k;
      printf("weight_MTER = %10.4f\n",weight_k);
	  fflush(stdout);
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              Q_COM->dbl(0,i,j) += weight_k * Q->dbl(0,i,j) / ( Q->dbl(0,0,0) + Q->dbl(0,1,1) + Q->dbl(0,2,2));
          }
          U_com->dbl(nloops,i) += weight_k * U->dbl(i);
      }
    }

  

    // ---- 8 : do MVAV with selected constrain ------ 

    if (doType[8]) {
      printf("Doing MVAV..\n");
	  fflush(stdout);
      Q =  Q_MVAV_calc(V,U,nestsize);
      ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained);
      
	  GenericMinVar minvar(constrained, Q, ProjMx);  
	  Evec = minvar.eigenvectors();
	  Evalues = minvar.eigenvalues(); 
            
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              q_mvav->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
          }
          U_mvav->dbl(nloops,i) = U->dbl(i);
      }
      
      
      weight_k =  ExWeights->dbl(7)  * weight_k_calc(Evalues);
      weight_sum += weight_k;
      printf("weight_MVAV = %10.4f\n",weight_k);
	  fflush(stdout);
      for (i=0; i<3; i++) {
          for (j=0; j<3; j++) {
              Q_COM->dbl(0,i,j) += weight_k * Q->dbl(0,i,j) / ( Q->dbl(0,0,0) + Q->dbl(0,1,1) + Q->dbl(0,2,2));
          }
          U_com->dbl(nloops,i) += weight_k * U->dbl(i);
      }

      // ---- do the double-for V -----------------

	 for (k=0; k<nestsize; k++) {
         for (i=0; i<3; i++) {      
             V_nestave->dbl(nloops<i) += V->dbl(k,i);	    
         }
	 }
	for (i=0; i<3; i++) {
	  V_nestave->dbl(nloops,i) /= (double) nestsize;
	 }
           
    }



    // ---- do COM with selected constrain ------

    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          Q_COM->dbl(0,i,j) /= weight_sum;
      }
    }


    ProjMx = Projection(B, V, U, Valf, EvecMVAB, alpha, nestsize, constrained);
    GenericMinVar minvar(constrained, Q_COM, ProjMx);  
    Evec = minvar.eigenvectors();
    
                         
    for (i=0; i<3; i++) {
       for (j=0; j<3; j++) {
           q_com->dbl(nloops,i,j) = Evec->dbl(0,0,i) * Evec->dbl(0,0,j);
       }
       U_com->dbl(nloops,i) /= weight_sum;
    }

    // ---- add up time -----------------

    t_start += 0.99*deltat;
    t_stop -= 0.99*deltat;
    trange = DvEvent(t_start, t_stop);
  }



  // ---- now, calc doubleaverages and do a Siscoe analysis of the normals -----------

  double nestfac = (double) max_numnest;
  for (k=0; k<max_numnest; k++) {
    for (i=0; i<3; i++) {
      B_doubleave->dbl(i) += B_nestave->dbl(k,i);
      V_doubleave->dbl(i) += V_nestave->dbl(k,i);
      Va_doubleave->dbl(i) += Va_nestave->dbl(k,i);
    }
  }

  for (i=0; i<3; i++) {
    B_doubleave->dbl(i) /= nestfac;
    V_doubleave->dbl(i) /= nestfac;
    Va_doubleave->dbl(i) /= nestfac;
  }
  


  if (doType[0]) ReportSiscoe(q_mvab, U_mva,     B_doubleave, V_doubleave, Va_doubleave, nloops, "MVAB", constrained,alpha, SCc_ID);
  if (doType[1]) ReportSiscoe(q_mvadb, U_mvadb,  B_doubleave, V_doubleave, Va_doubleave, nloops, "MVADB", constrained,alpha, SCc_ID);
  if (doType[3]) ReportSiscoe(q_mfr,  U_mfr,     B_doubleave, V_doubleave, Va_doubleave, nloops, "MFR",  constrained,alpha, SCc_ID);
  if (doType[4]) ReportSiscoe(q_mmr,  U_mmr,     B_doubleave, V_doubleave, Va_doubleave, nloops, "MMR",  constrained,alpha, SCc_ID);
  if (doType[5]) ReportSiscoe(q_mlmr, U_mlmr,    B_doubleave, V_doubleave, Va_doubleave, nloops, "MLMR", constrained,alpha, SCc_ID);
  if (doType[6]) ReportSiscoe(q_mer,  U_mer,     B_doubleave, V_doubleave, Va_doubleave, nloops, "MER",  constrained,alpha, SCc_ID);
  if (doType[7]) ReportSiscoe(q_mter, U_mter,    B_doubleave, V_doubleave, Va_doubleave, nloops, "MTER", constrained,alpha, SCc_ID);
  if (doType[8]) ReportSiscoe(q_mvav, U_mvav,     B_doubleave, V_doubleave, Va_doubleave, nloops, "MVAV",  constrained,alpha, SCc_ID);
  ReportSiscoe(q_com,  U_com,     B_doubleave, V_doubleave, Va_doubleave, nloops, "COM",  constrained,alpha, SCc_ID);

  return (0);
}




