
// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for MVA of any vector Quantity
//
// SEH, Mar 27, 2003
//      Sep 13, 2003  - cloned fromm massflow and modified 
//
// Inputs :
//     B = any vector time series, e.g., magnetic field   (n x 3)     
//
// -----------------------------------------------------------------------------------------

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "GenericResidue.h"



// -----------------------------------------------------------------------------------------
//
// Returns the Q_matrix for MVAB (Minimum Variance of B-field)
//
// SEH, 07 Sep, 2001
//      15 Oct, 2003, Doxygen DOC
//      21 Nov, 2003, Walen Slope + CC as output
//		9 June 2015 migrated to DVOS
//
// Doxygen info :
/*! 
  \brief Calculates the covariance matrix of the magnetic field.
  \param B -  vector time series with the magnetic field in units of nT
  \param V -  vector time series with plasma velocity in units of km/s
  \param U -  holds the calculated velocity on successful return
  \param nestsize - number of elements in B and V.
  \param full - 0 to do MVAB only, 1 to include HT analysis

*/ 

DvObject_var Q_MVAB_calc(DvObject_var &B, DvObject_var &V, DvObject_var &U, DvObject_var &Valf,
			  DvObject_var &N, double *wl, double *cc, int nestsize, bool doHT)
{
  printf("Q_MVAB_calc...nestsize = %d B=NULL : %d   %d\n",nestsize, (int)B.is_nil(), (int)B->seqSize());
  fflush(stdout);
  int i,j,k;
  double normfac = (double) nestsize;
  DvObject_var Q  = new DvObject(0.0, DIM33);
  DvObject_var aveB  = new DvObject(0.0, DIM3);
  DvObject_var dB  = new DvObject(0.0, DIM3, nestsize);
  DvObject_var tt = B->getTimeTags();       // tt of joined V and B
  DvTime timetag;
  bool PO = false;                           // PrintOut to screen ? 

  for (k=0; k<nestsize; k++) {              // find <B>
    for (i=0; i<3; i++) {
      aveB->dbl(i) += B->dbl(k,i) / normfac;
    }
  }

  if (PO)
  {
	printf("Q_MVAB : ave B : [%E  %E  %E]\n", aveB->dbl(0), aveB->dbl(1), aveB->dbl(2));
	fflush(stdout);
  }
  
  for (k=0; k<nestsize; k++) {             // find delta B
    for (i=0; i<3; i++) {
      dB->dbl(k,i) = B->dbl(k,i) - aveB->dbl(i);
    }
  }

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) Q->dbl(0,i,j) += dB->dbl(k,i) * dB->dbl(k,j) / normfac;
    }
   }

// --- SEH 2004-12-04 ---
//
// This gives a Q-matrix as Q=<Bi Bj> instead of <Bi Bj> - <Bi><Bj>
// suggested by BS, tested by SEH/BS 4 Dec 2004, but only minor improvement.
// -> do not implement
//
//    for (k=0; k<nestsize; k++) {
//     for (i=0; i<3; i++) {
//       for (j=0; j<3; j++) (*Q)[i][j] += (double)(*B)[k][i] * (double)(*B)[k][j] / normfac;
//     }
//   }
// --------------

  if (!doHT) return(Q);

  bool doB = B.is_ok() && B->seqSize() > 0;
  bool doV = V.is_ok() && V->seqSize() > 0;
  bool doN = N.is_ok() && N->seqSize() > 0;
  
  // ----- otherwise, proceed with HT analysis ------

  DvObject_var aveV   = new DvObject(0.0, DIM3);
  DvObject_var V_rest = new DvObject(0.0, DIM3, nestsize);
  Valf = new DvObject(0.0, DIM3, nestsize);  

  for (k=0; k<nestsize; k++) {              // find <V>
    for (i=0; i<3; i++) {
      aveV->dbl(i) += V->dbl(k,i) / normfac;;
    }
  }
    if (PO)printf("Q_MVAB : ave V : [%E  %E  %E]\n", aveV->dbl(0), aveV->dbl(1), aveV->dbl(2));

  if (PO)printf("Ready for HT..%d %d %d\n",  (int)B->seqSize(), (int)V->seqSize(), (int)nestsize);
  U = V_HofTel(B, V, nestsize);

  double mu_0 = 1.25664e-06;     // modified mu_0 (due to B in nT and dens in 1/ccm)
  double rho,m_p= 1.67e-27;      // rho = mass dens with proton_mass m_p

  if (doN) {
    for (i=0; i<nestsize; i++) {
      rho = N->dbl(i) * 1.e6 * m_p;
      for (j=0; j<3; j++) {
		V_rest->dbl(i,j) =  V->dbl(i,j) - U->dbl(j)*1.e-3;             // 1e-3 to work in km/s
		Valf->dbl(i,j) =   1.0e-3 * B->dbl(i,j) * 1e-9 / sqrt(mu_0 * rho);   // 1e-3 to work in km/s
      }
       //   printf(" Q_MVAB_calc  Vr=[%g   %g   %g]\n", V_rest->dbl(i,0), V_rest->dbl(i,1), V_rest->dbl(i,2));
       //   printf(" Q_MVAB_calc  V=[%g   %g   %g]\n", V->dbl(i,0), V->dbl(i,1), V->dbl(i,2));
       //   printf(" Valf=[%f   %f   %f]\n", Valf->dbl(i,0), Valf->dbl(i,1), Valf->dbl(i,2));
    }  
    
    
    double cctmp = correlationCoeff(Valf, V_rest, nestsize);
    double wltmp = slope(Valf, V_rest, nestsize);
    
    if (PO)
	{
		printf("Q_MVAB_calc > WalenSlope, Corrcoeff  = %f  %f, U=[%f   %f   %f]\n",wltmp,cctmp,U->dbl(0), U->dbl(1), U->dbl(2));
		fflush(stdout);
	}
    
    *cc = cctmp;
    *wl = wltmp; 
  }
  return (Q);
}






// -----------------------------------------------------------------------------------------
//
// Returns the V_HT frame
//
// SEH, Nov, 2001
//      May, 2002
//      Apr, 2003 (grabbed from HT Plugin, modified here)
//      19 Sep 2003, grabbed from Massflow plugin - simplified and modified here.
//
// Doxygen info :
/*! 
  \brief Returns the DeHoffmann-Teller velocity for the given magnetic 
  field and plasma velocity. The electric field used in the calculation
  is taken from the proxy E = - V x B.

  \param B -  vector time series with the magnetic field in units of nT
  \param V -  vector time series with plasma velocity in units of km/s
  \param nestsize - number of elements in B and V.

*/ 
// -----------------------------------------------------------------------------------------

DvObject_var V_HofTel(DvObject_var B, DvObject_var V, int nestsize)
{
  int i,j,l;
  double bb, time=0.0, normfac=(double) nestsize;

  DvObject_var K = new DvObject(0.0, DIM33);
  DvObject_var Vht = new DvObject(0.0, DIM3);
  DvObject_var B_k = new DvObject(0.0, DIM3);
  DvObject_var K0 = new DvObject(0.0, DIM33);
  DvObject_var KV = new  DvObject(0.0, DIM3);

  for(size_t k=0; k<nestsize; k++)  {
      
    B_k = B->subSequence(k);
    DvObject_var BBk = B_k->dot(B_k);
    bb = BBk->dbl(0);
    

    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
		K->dbl(0,i,j) = (double)(i==j) * bb - (B_k->dbl(i) * B_k->dbl(j));
      }
    }
     
     for (i=0; i<3; i++) {
       for (j=0; j<3; j++) {
            K0->dbl(0,i,j) += K->dbl(0,i,j);
			KV->dbl(i) += K->dbl(0,i,j) * V->dbl(k,j) * 1e3;
       }
     }

  }


  for (i=0; i<3; i++){
    for (j=0; j<3; j++) {
      K0->dbl(0,i,j) /= normfac;                                   // K0 == <K>
    }
    KV->dbl(i) /= normfac;
  }

  DvObject_var Kinv = K0->inverse();


  for  (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      Vht->dbl(i) += Kinv->dbl(0,i,j) * KV->dbl(j); 
    }
  }
  return (Vht);
}





double slope(DvObject_var X_in, DvObject_var Y_in, int n)
// ---------------------------------------------------------------------------
//
//  calculates the regression line, i.e. the coeff b in y(x) = a + bx
//
//
// SEH, 15 Aug, 2002
//      21 Nov 2003, grabbed from HofTel + converted to return a double
//
// ---------------------------------------------------------------------------

{
  int numpts = 3*n;

  vector<double> Xc(numpts, 0.0);
  vector<double> Yc(numpts, 0.0);
  vector<double> Tc(numpts, 0.0);
  
  for (int i=0; i<n; i++) {
    Xc[i] = X_in->dbl(i,0);
    Xc[i+n] = X_in->dbl(i,1);
    Xc[i+2*n] = X_in->dbl(i,2);
    Yc[i] = Y_in->dbl(i,0);
    Yc[i+n] = Y_in->dbl(i,1);
    Yc[i+2*n] = Y_in->dbl(i,2);
  }
  
  double sum_X = 0.0;
  double sum_Y = 0.0;
  for(int i=0; i< numpts; i++){
  	sum_X += Xc[i];
  	sum_Y += Yc[i];
  }
  double ave_X = sum_X / (double) (numpts);
  double ave_Y = sum_Y / (double) (numpts);
  double  b=0.0, tt=0.0;

  for (int i=0; i<numpts; i++) {
    Tc[i] = Xc[i] - ave_X;
    b += Tc[i] * Yc[i];
    tt += Tc[i] * Tc[i];
  }

  b /= tt;
  return(b);
}



double correlationCoeff(DvObject_var X_in, DvObject_var Y_in, int n)
// ---------------------------------------------------------------------------
//
//  calculates the correlation coefficient
//
//
// SEH, 15 Jan 2003 
//      21 Nov 2003, grabbed from HofTel + converted to return a double
//
// ---------------------------------------------------------------------------
{

  double sum_upper=0.0, sum_xx=0.0, sum_yy=0.0;
  int numpts = 3*n;

  vector<double> Xc(numpts, 0.0);
  vector<double> Yc(numpts, 0.0);
  vector<double> Tc(numpts, 0.0);

  for (int i=0; i<n; i++) {
    Xc[i] = X_in->dbl(i,0);
    Xc[i+n] = X_in->dbl(i,1);
    Xc[i+2*n] = X_in->dbl(i,2);
    Yc[i] = Y_in->dbl(i,0);
    Yc[i+n] = Y_in->dbl(i,1);
    Yc[i+2*n] = Y_in->dbl(i,2);
  }

  double sum_x=0;
  for(int i=0; i<numpts; i++) sum_x += Xc[i]; 
  sum_x /= (double) numpts;
  double sum_y=0;
  for(int i=0; i<numpts; i++) sum_y += Yc[i]; 
  sum_y /= (double) numpts;

  for (int i=0; i<numpts; i++) {
    sum_upper += (Xc[i] - sum_x) * (Yc[i] - sum_y);
    sum_xx += (Xc[i] - sum_x)*(Xc[i] - sum_x);
    sum_yy += (Yc[i] - sum_y)*(Yc[i] - sum_y);
  }

  return (double) (sum_upper / (sqrt(sum_xx) * sqrt(sum_yy)));
}

double vecsum(DvObject_var Vector)
{
  // returns the norm of a vector
  double sum=0.0;

  for (int i=0; i<3; i++) sum += Vector->dbl(i) * Vector->dbl(i);

  return(sqrt(sum));
}

double vecsum2(DvObject_var Vector)
{
  // returns the norm^2 of a vector
  double sum=0.0;

  for (int i=0; i<3; i++) sum += Vector->dbl(i) * Vector->dbl(i);

  return(sum);
}
