// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : V_perp and V_par
//
// This plugin takes a magnetic field vector and any vector (e.g., velocity)
// as inputs, and calculates the components of this vector 
// along and perpendicular to the magnetic field
//
// Inputs :
//      bfield          - vector time series with magnetic field
//      invector        - vector time series to be split
//
// Outputs
//      v_perp          - vector perpendicular to B-field
//      v_par           - component of vector along B-field
//
//
// SEH, 19 Sep, 2001
//      11 Mar, 2003    - join V and B in PlugIn
//      26 Dec, 2003    - QSAS 2.0 version; Changed headers + 'external'
//      16 Nov, 2016    - DVOS : make copy of object before manipulating
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

// needed for QSAS 2.0
extern "C" QplugReturnStatus  DvISSI_Vperp  ( vector <DvObject_var>&, vector <DvObject_var>& );


using namespace QSAS;

QplugReturnStatus  DvISSI_Vperp  ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{

QplugReturnStatus status = QPLUG_SUCCESS;


DvObject_var test_B=new DvObject(inputs[0]);
DvObject_var test_V=new DvObject(inputs[1]);

if(test_B.is_nil())
{
	QplugAppendTextDisplay ("ISSI_Vperp >> ERROR - empty obj for bfield\n");
	return QPLUG_FAILURE;  
}

if( !test_B->sameBaseSI("1 > T") )
{
	QplugAppendTextDisplay ("ISSI_Vperp >> ERROR - Wrong units for bfield\n");
	return QPLUG_FAILURE;  
}

if(test_V.is_nil())
{
	QplugAppendTextDisplay ("ISSI_Vperp >> ERROR - empty obj for input vector\n");
	return QPLUG_FAILURE;  
}


// declare and calculate the output objects
// Joining and metadata are handled by DVOS on the fly

DvObject_var Bmag = test_B->abs();

test_B /= Bmag; // Normalised B direction

DvObject_var v_par = test_V->dot(test_B);
DvObject_var v_par_vec = test_B * v_par;
DvObject_var v_perp =test_V - v_par_vec;


for(int i=0; i<v_perp->nError(); i++)  {
   QplugAppendTextDisplay(v_perp->getError(i).c_str());
   status = QPLUG_WARNING;
}

for(int i=0; i<v_par->nError(); i++)  {
   QplugAppendTextDisplay(v_par->getError(i).c_str());
   status = QPLUG_WARNING;
}


outputs.push_back(v_perp);
outputs.push_back(v_par);

QplugAppendTextDisplay ("ISSI_Vperp>> data object created - exiting\n");

return status;

}



