// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Solar Wind Parameter calculator
// 
// This plugin takes a magnetic field vector and plasma data as inputs
// and calculates the solar wind epsilon paramtere, accumulated
// energy, dynamic pressure and the magnetopause standoff distance.
//
// Inputs :
// 	bfield		- vector time series with magnetic field
// 	velocity	- time series with plasma velocity
// 	density	        - scalar time series with plasma density
// 	timeinterval	- time interval
//
// Outputs
//      epsilon parameter - scalar time series with the Akasofu epsilon param
//      energy   - scalar time series with accumulated energy input
//      pdyn  - scalar time series solar wind dynamic pressure
//      Rst  - scalar time series the Magnetopayuse standoff distance
//      clockangle  - Bimf clock angle (0 deg = northward, 180 deg = southward)
//
// SEH, 03 Mar, 2002
//      25 Nov, 2002 - allowed for scalar Vsw + clock angle output.
//      09 Dec, 2002 - V may be vector series, scalarseries or constant
//                     and n may be scalar series or constant.
//      10 Dec, 2002 - allow for time shift + corrected bug when changing time
//      11 Dec, 2002 - V_bulk (shifted) as output 
//      27 Oct, 2004 - B_IMF (shifted) as output + variable time delay option
// AJA  03 Oct, 2005 - Added to QSAS distribution and defensive programing and units conversion added
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"


typedef struct VEC {double x,y,z;} vec;

using namespace QSAS;

extern "C" QplugReturnStatus  DvISSI_Solarwind  ( vector<DvObject_var>& inputs, vector<DvObject_var>& outputs )

{

 DvObject_var bfield;
 DvObject_var epsilon;
 DvObject_var ts_in;
 DvObject_var energy;
 DvObject_var vout;
 DvObject_var pdyn;
 DvObject_var rstandoff;
 DvObject_var clockangle;
 DvObject_var bout;
 DvObject_var tt_out;
 DvObject_var bfield_nest;
 DvObject_var velocity_nest;
 DvObject_var density_nest;

// -- variables ----
// ionmass = 1.67E-6 = 1.67*1E-27 [kg] *(1E6) * (1E9) * (1E6)
//	 ...1E6 = 1/ccm -> 1/m^3 * ...
//       ...1E9 = Pa -> nPa
//       ...1E6 = (km/s)^2 -> (m/s)^2

 long    n_vecs,nestsize;
 double  L0, btot2,theta,deltat,avg_width,Vsw,TotalEnergy=0.0,Bt;
 double  ionMass= 1.67E-6;                                         
 int     nseq =3;
 vec B,V;
 char    txtbuf[200];

 
 // ---- read other inputs ----

 DvObject_var tt_B = inputs[0]->getTimeTags();
 if (tt_B.is_nil()) {
   QplugAppendTextDisplay("Bfield has no time tags\n");
   return(QPLUG_FAILURE);
 }
 long nb = tt_B->seqSize();
 

 DvObject_var b_in = inputs[0];
 if (b_in.is_nil() || b_in->not_dbl() || !b_in->isThreeVector()) {
   QplugAppendTextDisplay("Bfield wrong or not vector time series\n");
   return(QPLUG_FAILURE);
 }
 
 bfield = b_in->changeUnitsTo("1.0e-9>T", "nT"); // ensure units are in Cluster units assumed
 if( !b_in->xref_exists(SI_CONVERSION) ){
    // bfield is B_in
    // ACE does not have CSDS/CAA attributes needed by changeUnitsTo()
    if (b_in->getXrefText(UNITS) == "0.1nT") {
        *bfield *= 0.1;
    }  
 }


 //  --- check if V is vector -----
 
 ts_in = inputs[1];
 if(ts_in.is_nil()) {
    QplugAppendTextDisplay("Cannot find valid Velocity\n");
    return(QPLUG_FAILURE);
 }
 
 if ( ts_in->isThreeVector() ) {
    QplugAppendTextDisplay("Velocity is vector .. calculating |V|...\n");
    
    ts_in = ts_in->getVectorComponent(DV_RMAG);
   
 }
 
 if(ts_in->seqSize() == 1)  {
  
	ts_in = new DvObject(ts_in, nb, true);
    ts_in->change_xref(DEPEND_0,tt_B);  // has none of its own
    sprintf(txtbuf,"NB! Velocity is constant %7.2f\n", (float)ts_in->dbl());
    QplugAppendTextDisplay(txtbuf);
 }

 // Now convert to required units if possible

 ts_in =  ts_in->changeUnitsTo( "1.0e3>m s^-1", "km/s"); // ensure units are in Cluster units assumed
 
 // ---- check density. If not scalar series -> use given input -----
 
 DvObject_var n_in = inputs[2];
 if(n_in.is_nil() ){
    QplugAppendTextDisplay("Cannot find valid density\n");
    return(QPLUG_FAILURE);
 }
 
 if( n_in->seqSize() > 1 ) {
   printf("OK - density is scalar time series\n");
   fflush(stdout);
 } 
 else {
     printf("OK - density is given constant\n");
	 fflush(stdout);
     n_in = new DvObject(n_in, nb, true);
     n_in->change_xref(DEPEND_0, tt_B);
     sprintf(txtbuf,"NB! Velocity is constant %7.2f\n",(float) n_in->dbl());
     QplugAppendTextDisplay(txtbuf);
 }

 n_in = n_in->changeUnitsTo( "1.0e6>m^-3", "/cc"); // ensure units are per cc
 
 //  --- Join B measurements to plasma measurements -----

 b_in->isRegular(deltat); // need spacing but not regular tt
 avg_width = 1.5 * deltat;
 DvJoinList ToJoinList;
 DvJoinList JoinedList;

 ToJoinList.append("B", bfield);
 bfield->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
 bfield->change_xref(DV_GAP_WIDTH, avg_width);
 bfield->change_xref(DV_GAP_OPTION, DV_LINEAR);
 
 ToJoinList.append("V", ts_in);
 ts_in->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
 ts_in->change_xref(DV_GAP_WIDTH, avg_width);
 ts_in->change_xref(DV_GAP_OPTION, DV_LINEAR);
 
 ToJoinList.append("N", n_in);
 n_in->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
 n_in->change_xref(DV_GAP_WIDTH, avg_width);
 n_in->change_xref(DV_GAP_OPTION, DV_LINEAR);


 printf("delta t = %f  ->  gap_fill if t > %f\n", deltat, avg_width);
 fflush(stdout);


 //  ---- get Time interval and mask ----------------

 DvEvent start_end = inputs[3]->getTimeRange();
 
 tt_out = tt_B->subSequence(start_end);
 if (tt_out.is_nil()) {    
   QplugAppendTextDisplay("No time tags in the selected interval\n");
   return(QPLUG_FAILURE);
 }

   
 ToJoinList.MultiJoin(tt_out, JoinedList);


 DvNode *Bnode = JoinedList.find("B");     
 if(Bnode) bfield_nest = Bnode->obj();
   
 if(bfield_nest.is_nil() || bfield_nest->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No B field in range ");
     return (QPLUG_FAILURE);
 }

 nestsize = bfield_nest->seqSize();
 printf("Nestsize B = %ld\n",nestsize);
 fflush(stdout);

 DvNode *Vnode = JoinedList.find("V");     
 if(Vnode) velocity_nest = Vnode->obj();
   
 if(velocity_nest.is_nil() || velocity_nest->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No V data in interval ");
     return (QPLUG_FAILURE);
 }


 DvNode *Nnode = JoinedList.find("N");     
 if(Nnode) density_nest = Nnode->obj();
   
 if(density_nest.is_nil() || density_nest->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No Density data in interval ");
     return (QPLUG_FAILURE);
 }


 // ----- check if time shift is given -------

 bool shift_given = false;
 double shift = 0.0;
 DvObject_var shift_ws = inputs[4];
 if (shift_ws.is_ok()) {
   shift_given = true;
   shift = shift_ws->asDouble();
 }


 // ---- declare and calculate the output objects -----

 epsilon    = new DvObject(0.0, nestsize);
 energy     = new DvObject(0.0, nestsize);
 pdyn       = new DvObject(0.0, nestsize);
 rstandoff  = new DvObject(0.0, nestsize);
 clockangle = new DvObject(0.0, nestsize);
 vout       = new DvObject(0.0, nestsize);
 bout       = new DvObject(0.0, DIM3, nestsize);

 L0 = 7.0 * 6370000;  
 double ACEpos = 1.582e6 - 63700.0;  // ACE position rel to BS at 10Re
 double aveshift = 0.00;

 // calculate epsilon and Pdyn
 // NB ! we use SI units B [T], V [km/s]
 // epilon [W], pdyn [nPa]
 // TotalEnergy in [10^12 Ws = 10^12 J]
 // rstandoff = MP Rst [Re] after (Walker & Russell, 1995)

 float dir = 1.00;

 for(int i=0; i<nestsize; i++)  {
    B.x = bfield_nest->dbl(i,0) * 1.0E-9;
    B.y = bfield_nest->dbl(i,1) * 1.0E-9;
    B.z = bfield_nest->dbl(i,2) * 1.0E-9;
    Vsw = velocity_nest->dbl(i);
    
    theta = acos(B.z / sqrt(B.y*B.y + B.z*B.z));
    btot2 = (B.x*B.x + B.y*B.y + B.z*B.z);
    Bt = sqrt(B.y*B.y + B.z*B.z);
    dir = 1.0;
    if (B.y <= 0.0) dir = -1.0;
    bout->dbl(i,0) = bfield_nest->dbl(i,0);
    bout->dbl(i,1) = bfield_nest->dbl(i,1);
    bout->dbl(i,2) = bfield_nest->dbl(i,2);
    epsilon->dbl(i) = 1.0E7 * L0*L0 * btot2 * pow(sin(theta/2.0),4.0) * Vsw*1E3;   
    energy->dbl(i) = TotalEnergy + epsilon->dbl(i) * (double)deltat * 1.0E-12; 
    pdyn->dbl(i) = density_nest->dbl(i) * ionMass * Vsw * Vsw;
    rstandoff->dbl(i) = 107.4 * pow((density_nest->dbl(i) * Vsw*Vsw),-0.16667);
    clockangle->dbl(i) = dir * 57.29 * acos(B.z/Bt);
    vout->dbl(i) = Vsw;
    if (!shift_given) shift = (ACEpos/Vsw) + 300.0;
    tt_out->time(i) += shift;
    if (i > 0) {
      if (tt_out->time(i) < tt_out->time(i-1) ) tt_out->time(i) = tt_out->time(i-1) + 0.1;
    }
    TotalEnergy = energy->dbl(i);
    aveshift += shift;
    //    printf(" i, Eps, Pdyn, Rst, TotE %d %10.2f %10.2f %10.2f = %10.2f\n",
    //	     i,epsilon->dbl(i),pdyn->dbl(i),rstandoff->dbl(i),TotalEnergy);
 }    

 aveshift /= (double) nestsize;

 // printf("Completed calculations..\n");

 // --- send some output to QSAS window ---
 sprintf(txtbuf,"\nNest size (# of data points) = %ld\n",nestsize);
 QplugAppendTextDisplay(txtbuf);

 sprintf(txtbuf,"Total accumulated energy in interval = %8.2f x 10^12 J\n\n",TotalEnergy);
 QplugAppendTextDisplay(txtbuf);


 // --- insert timetags + attributes and push to WL --- 
 
 sprintf(txtbuf,"Solar wind bulk speed |V|, shifted %d seconds",(int) aveshift);
 vout->change_xref(DEPEND_0, tt_out);
 vout->change_xref("SI_conversion","1.0e3>m s^-1");
 vout->change_xref("UNITS","km/s");
 vout->change_xref("LABLAXIS","V_bulk");
 vout->change_xref("CATDESC",txtbuf);
 if (shift != 0.0) {
   sprintf(txtbuf,"V_SW_Bulk_sh_%ds",(int) aveshift);
   SetSlotText(0,txtbuf);
 }

 sprintf(txtbuf,"BIMF, shifted %d seconds",(int) aveshift);
 bout->change_xref(DEPEND_0, tt_out);
 bout->change_xref("SI_conversion","1.0e-9>T");
 bout->change_xref("UNITS","nT");
 bout->change_xref(FRAME, "vector>gse_xyz");
 bout->change_xref("LABLAXIS","B_IMF");
 bout->change_xref("CATDESC",txtbuf);
 if (shift != 0.0) {
   sprintf(txtbuf,"B_IMF_sh_%ds",(int) aveshift);
   SetSlotText(1,txtbuf);
 }


 sprintf(txtbuf,"Epsilon parameter, shifted %d seconds",(int) aveshift);
 epsilon->change_xref(DEPEND_0, tt_out);
 epsilon->change_xref("SI_conversion","1>W");
 epsilon->change_xref("UNITS","W");
 epsilon->change_xref("LABLAXIS","Epsilon");
 epsilon->change_xref("CATDESC",txtbuf);
 if (shift != 0.0) {
   sprintf(txtbuf,"Epsilon_sh_%ds",(int) aveshift);
   SetSlotText(2,txtbuf);
 }


 sprintf(txtbuf,"Accumulated energy , shifted %d seconds",(int) aveshift);
 energy->change_xref(DEPEND_0, tt_out);
 energy->change_xref("SI_conversion","1.0e12>J");
 energy->change_xref("UNITS","10^12 J");
 energy->change_xref("LABLAXIS","Acc_energy");
 energy->change_xref("CATDESC",txtbuf);
 if (shift != 0.0) {
   sprintf(txtbuf,"Energy_sh_%ds",(int) aveshift);
   SetSlotText(3,txtbuf);
 }

 sprintf(txtbuf,"Solar wind dynamic pressure , shifted %d seconds",(int) aveshift);
 pdyn->change_xref(DEPEND_0, tt_out);
 pdyn->change_xref("SI_conversion","1e-09>Pa");
 pdyn->change_xref("UNITS","nPa");
 pdyn->change_xref("LABLAXIS","Pdyn");
 pdyn->change_xref("CATDESC",txtbuf);
 if (shift != 0.0) {
   sprintf(txtbuf,"P_dyn_sh_%ds",(int) aveshift);
   SetSlotText(4,txtbuf);
 }

 sprintf(txtbuf,"magnetopause standoff distance Rst = 107.4*[nV^2]^(-1/6), shifted %d seconds",(int) aveshift);
 rstandoff->change_xref(DEPEND_0, tt_out);
 rstandoff->change_xref("UNITS","Re");
 rstandoff->change_xref("LABLAXIS","Rst");
 rstandoff->change_xref("CATDESC",txtbuf);
 if (shift != 0.0) {
   sprintf(txtbuf,"Rst_sh_%ds",(int) aveshift);
   SetSlotText(5,txtbuf);
 }

 sprintf(txtbuf,"IMF clock angle, shifted %d seconds",(int) aveshift);
 clockangle->change_xref(DEPEND_0, tt_out);
 clockangle->change_xref("SI_conversion","1>degree");
 clockangle->change_xref("UNITS","Degrees");
 clockangle->change_xref("LABLAXIS","Theta");
 clockangle->change_xref("CATDESC",txtbuf);
 if (shift != 0.0) {
   sprintf(txtbuf,"Theta_sh_%ds",(int) aveshift);
   SetSlotText(6,txtbuf);
 }

 outputs.push_back(vout);
 outputs.push_back(bout);
 outputs.push_back(epsilon);
 outputs.push_back(energy);
 outputs.push_back(pdyn);
 outputs.push_back(rstandoff);
 outputs.push_back(clockangle);

 QplugAppendTextDisplay ("ISSI_Solarwind >> data object created - exiting successfully\n");

 return QPLUG_SUCCESS;

}




