/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	MaxEntSpectrum.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Maximum Entropy Power spectrum wrapper
*
* PURPOSE:      Wrapper for NR 'evlmem' power spectrum.
* 
* DESCRIPTION:  Plug In module checking input parameters.
*
* ARGUMENTS:
*    IN:       inputs  list of plug in objects containing
*                obj vector or scalar regular time series object
*                obj interval object for subsetting
*                int  npoles  number of poles
*                int  nfreq   number of frequencies
*                int  linlog  linear or logarithmic frequency spacing
*                float minfreq minimal frequency
*                float maxfreq maximal frequency
*
*   OUT:
*         float *frequency   frequency tags		
*         float *power_mag       power spectral density of magnitude
*         vector series *power_vec       power spectral density of components
*
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED:	Qcoef Qevlp.c
*
* ALGORITHM:  see Press et al. 'Numerical Recipes in C' (1992)
*
* HISTORY: created  24/11/97 Markus Fraenz 
* Migrated to QDOS 2 and C arrays 09/2000 A.J.Allen 
* indexing repaired 16/1/2001 Markus Fraenz
* vector input and subsetting 18/01/2001 Markus Fraenz
*
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

extern "C" QplugReturnStatus  DvMaxEntSpectrum(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;

static float sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)

void Qcoef(float data[],unsigned long n,unsigned long m,float *xms,float d[]);
float Qevlp(float fdt, float d[], unsigned long m, float xms);


QplugReturnStatus	DvMaxEntSpectrum( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{

// define some C variables

long	i, j;

int	    linlog,vectorset=0;
unsigned long	Ndata, nfreq, npoles;
float           minfreq,maxfreq;

char *attr_value, str_out[150];

float	xms,fdt=0.0,freqstep=0.0,freqwidth=0.0;
float  *d,*freq;
double deltat=0.0,deltat2=0.0;


// define QSAS data objects

DvObject_var ts_out;
DvObject_var ts_inr;
DvObject_var ts_inm;

DvObject_var  out_freq_series;
DvObject_var  out_spmag_series;
DvObject_var  out_spvec_series;

// ...first item is input Time Series

DvObject_var ts_in = inputs[0];


if ( !ts_in->isRegular(deltat) ) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - time series not evenly spaced\n");
	return QPLUG_FAILURE; 
}

//	..2nd item is option 'time_interval'

 DvObject_var dobj1 = inputs[1];
 DvEvent  start_end = dobj1->getTimeRange();
 
//	..3rd item is option 'npoles'

DvObject_var option_np = inputs[2];

if ( option_np.is_nil()  ) { 
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - npoles missing.\n");
	return QPLUG_FAILURE; 
}

npoles = option_np->asInt();
 
if (npoles <= 0) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - npoles must be greater than 0.\n");
	return QPLUG_FAILURE; 
}


//	..3rd item is option 'nfreq'

DvObject_var option_nf = inputs[3];

if ( option_nf.is_nil()  ) { 
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - nfreq missing\n");
	return QPLUG_FAILURE; 
}

nfreq = option_nf->asInt();
if (nfreq <= 0) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - nfreq must be greater than 0.\n");
	return QPLUG_FAILURE;
}


//	..4th item is option 'linlog'

DvObject_var option_ll = inputs[4];

if ( option_ll.is_nil()  ) { 
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - linlog missing\n");
	return QPLUG_FAILURE; 
}

linlog = option_ll->asInt();
if (linlog != 0 && linlog != 1) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - linlog must be 0 or 1.\n");
	return QPLUG_FAILURE; 
}


//	..5th item is option 'minfreq'

DvObject_var option_mf = inputs[5];

if ( option_mf.is_nil()  ) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - minfreq missing\n");
	return QPLUG_FAILURE; 
}
minfreq = (float)option_mf->asDouble();

if (linlog > 0 && minfreq <= 0) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - for log spacing minfreq must be > 0.\n");
	return QPLUG_FAILURE; 
}



//	..6th item is option 'maxfreq'

DvObject_var option_Mf = inputs[6];

if ( option_Mf.is_nil() ) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - maxfreq missing\n");
	return QPLUG_FAILURE; 
}
maxfreq = (float)option_Mf->asDouble();


// now take the subset for the time interval chosen  ###SUBSETTING START

QplugAppendTextDisplay ("MaxEntSpectrum>> Subsetting Time Series...\n");
  

ts_out = ts_in->subSequence(start_end);
    
if(ts_out.is_nil() || ts_out->seqSize() < 2 ){
    QplugAppendTextDisplay ("MaxEntSpectrumFFT>> ERROR - no data in range?\n");
    return QPLUG_FAILURE; 
}


//	..extract the time tags from the data series object ###SUBSETTING END

Ndata = ts_out->seqSize();

if (Ndata <= 1) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - Series must have at least 2 data points.\n");
	return QPLUG_FAILURE; 
}

if (npoles > Ndata) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - npoles must be smaller than number of data points.\n");
	return QPLUG_FAILURE; 
}

if (nfreq > Ndata) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - nfreq must be smaller than number of data points.\n");
	return QPLUG_FAILURE; 
}



//	..extract the time tags and data from the data series object

double *tags = new double[Ndata];
float *data = new float[Ndata];

if ( ts_out->isThreeVector() ) {
    vectorset=1;	
    QplugAppendTextDisplay ("MaxEntSpectrum>> Decomposing Vector Timeseries...\n");
    ts_inm = ts_out;
    ts_inr = ts_out->getVectorComponent(DV_RMAG); 
}
else if(ts_out->arraySize() == 1 ){
    ts_inr = ts_out;
}
else{
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - only scalars and vectors handled\n");
	return QPLUG_FAILURE; 
}

// check Nyquist interval

if (fabs(minfreq*deltat) > 0.5) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - |minfreq*deltat| not <0.5 \n");
	return QPLUG_FAILURE; 
}
if (fabs(maxfreq*deltat) > 0.5) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - |maxfreq*deltat| not <0.5 \n");
	return QPLUG_FAILURE; 
}
if (maxfreq <= minfreq) {
	QplugAppendTextDisplay ("MaxEntSpectrum>> ERROR - maxfreq must be greater than minfreq.\n");
	return QPLUG_FAILURE; 
}


DvObject_var tt_out = ts_out->getTimeTags();
for (int i=0; i<Ndata; i++){
  tags[i] = tt_out->time(i) - tt_out->time(0);
  data[i] = (float) ts_inr->dbl(i);
}

QplugAppendTextDisplay ("MaxEntSpectrum>> Calculating Spectrum.\n");
//	call the NR memcof power spectrum calculation
//      (notice that NR routines start indices with 1!!) 

d = new float[npoles];
freq = new float[nfreq];

if (linlog > 0) 
  freqstep=(log(maxfreq)-log(minfreq))/nfreq;
else 
  freqstep=(maxfreq-minfreq)/nfreq;

for(i=0;i<nfreq;i++) { 
 if (linlog > 0) 
  freq[i]=exp(freqstep*i+log(minfreq));
 else
  freq[i]=freqstep*i+minfreq;
}

Qcoef(data, Ndata, npoles, &xms, d);

//	..create the output data series objects

out_freq_series  = new DvObject(0.0, nfreq );
out_spmag_series  = new DvObject(0.0, nfreq );
for(int i=0; i<nfreq; i++) {
  out_freq_series->dbl(i) =  (double) freq[i];
  out_spmag_series->dbl(i) =  (double) Qevlp(freq[i]*deltat, d, npoles, xms);
}


if(vectorset) { 

 out_spvec_series = new DvObject(0.0, DIM3, nfreq);
 
 for(long i=0; i<Ndata; i++)  data[i] = (float) ts_inm->dbl(i,0);
 Qcoef(data, Ndata, npoles, &xms, d);
 for (long i =0; i<nfreq; i++) out_spvec_series->dbl(i,0) = (double) Qevlp(freq[i]*deltat, d, npoles, xms);

 for(long i=0; i<Ndata; i++)  data[i] = (float) ts_inm->dbl(i,1);
 Qcoef(data, Ndata, npoles, &xms, d);
 for (long i =0; i<nfreq; i++) out_spvec_series->dbl(i,1) = (double) Qevlp(freq[i]*deltat, d, npoles, xms);

 for(long i=0; i<Ndata; i++)  data[i] = (float) ts_inm->dbl(i,2);
 Qcoef(data, Ndata, npoles, &xms, d);
 for (long i =0; i<nfreq; i++) out_spvec_series->dbl(i,1) = (double) Qevlp(freq[i]*deltat, d, npoles, xms);
 
}


//	...set some attributes

out_spmag_series->change_xref(LABLAXIS, "Maximum Entropy Power");
out_spmag_series->change_xref(FIELDNAM, "Maximum Entropy Spectrogram");
sprintf(str_out,"(%s)^2\n", ts_in->getXrefText(UNITS).c_str());
out_spmag_series->change_xref(UNITS, str_out);
out_spmag_series->change_xref(DEPEND_0, out_freq_series);


if(vectorset) {

    out_spvec_series->change_xref(LABLAXIS, "MaxEnt Spectrogram");
    out_spvec_series->change_xref(FIELDNAM, "MaxEnt Spectrogram");
    out_spvec_series->change_xref(UNITS, str_out);
    out_spvec_series->change_xref(DEPEND_0, out_freq_series);
}

out_freq_series->change_xref(LABLAXIS, "frequency");
out_freq_series->change_xref(FIELDNAM, "frequency");
out_freq_series->change_xref(SI_CONVERSION, "1>Hz");
out_freq_series->change_xref("UNITS", "Hz");

//	..add them to the list

  outputs.push_back(out_freq_series);     
  outputs.push_back(out_spmag_series);     
  outputs.push_back(out_spvec_series); 
  
//	..some messages to the analysis interface window..

sprintf(str_out, "MaxEntSpectrum>> %ld data points\n", Ndata);
QplugAppendTextDisplay (str_out);
sprintf(str_out,"MaxEntSpectrum>> %ld frequencies\n", nfreq);
QplugAppendTextDisplay (str_out);
sprintf(str_out,"MaxEntSpectrum>> mean square discrepancy : %f \n", xms);
QplugAppendTextDisplay (str_out);
QplugAppendTextDisplay ("MaxEntSpectrum>> finished...\n");


//	..clean up and exit

delete[] data;
delete[] tags;
delete[] d;
delete[] freq;

return QPLUG_SUCCESS;


}


float Qevlp(float fdt, float d[], unsigned long m, float xms)
{
	int i;
	float sumr=1.0,sumi=0.0;
	double wr=1.0,wi=0.0,wpr,wpi,wtemp,theta;

	theta=6.28318530717959*fdt;
	wpr=cos(theta);
	wpi=sin(theta);
	for (i=0;i<m;i++) {
		wr=(wtemp=wr)*wpr-wi*wpi;
		wi=wi*wpr+wtemp*wpi;
		sumr -= d[i]*wr;
		sumi -= d[i]*wi;
	}
	return xms/(sumr*sumr+sumi*sumi);
}

void Qcoef(float data[], unsigned long n, unsigned long m, float *xms, float d[])
{
	int k,j,i;
	float p=0.0,*wk1,*wk2,*wkm;

	wk1= (float *) malloc( sizeof(float) * n );
	wk2= (float *) malloc( sizeof(float) * n );
	wkm= (float *) malloc( sizeof(float) * m );
	for (j=0;j<n;j++) p += SQR(data[j]);
	*xms=p/n;
	wk1[0]=data[0];
	wk2[n-2]=data[n-1];
	for (j=1;j<n-1;j++) {
		wk1[j]=data[j];
		wk2[j-1]=data[j];
	}
	for (k=0;k<m;k++) {
		float num=0.0,denom=0.0;
		for (j=0;j<(n-k-1);j++) {
			num += wk1[j]*wk2[j];
			denom += SQR(wk1[j]);
			denom += SQR(wk2[j]);
		}
		d[k]=2.0*num/denom;
		*xms *= (1.0-SQR(d[k]));
		for (i=0;i<k;i++)
			d[i]=wkm[i]-d[k]*wkm[k-i-1];
		if (k == m-1) {
			free(wk1);
			free(wk2);
			free(wkm);
			return;
		}
		for (i=0;i<=k;i++) wkm[i]=d[i];
		for (j=0;j<(n-k-2);j++) {
			wk1[j] -= wkm[k]*wk2[j];
			wk2[j]=wk2[j+1]-wkm[k]*wk1[j+1];
		}
	}
	printf("never get here in Qcoef.\n");
}



