/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	Q4SC_Normal.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS 4 SC Determination of Boundary Normal and Velocity wrapper
*
* PURPOSE:      Calculate the normal and boundary velocity based on the
*               crossing of the boundary (assumed planar) by 4 spacecraft.
*               Based on the methd in Ch 10.4.3 of the ISSI book.
*
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*   INPUT   VECTOR    dr1  pos of s/c1 as const vec or time ser
    INPUT   VECTOR    dr2 ,
    INPUT   VECTOR    dr3 ,
    INPUT   VECTOR    dr4 ,
    INPUT   TIME      tic,  time used if dr's are time series
    INPUT   ANY_DATA_OBJECT    ti1 , event time as s/c 1 in secs 
    INPUT   ANY_DATA_OBJECT    ti2 ,
    INPUT   ANY_DATA_OBJECT    ti3 ,
    INPUT   ANY_DATA_OBJECT    ti4 ,

    Out:

    OUTPUT  MATRIX             n   , normal vector with same frame as dr's
                                     n_x forced positive
    OUTPUT  FLOAT              Vn  , velocity along normal

* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: Hacked from Boundary Normal routine
*
* HISTORY: November 2003 Steve Schwartz 
*          January 2004 AJA added error traps
*
***********************************************************************/

//#define Q4SC_DIAG

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>
#include "qplug_if.h"
#include "qrtn.h"

#include "qdutil.h"
#include "DvObj.h"
#include "Xrefs.h"

//const double Re = 6.3712e6; //Earth radius [m]  - a global	

using namespace QSAS;  

// local prototypes

extern "C" QplugReturnStatus DvQ4SC_Normal(vector<DvObject_var>&, vector<DvObject_var>&);

QplugReturnStatus Q4SCPosition(RtnVector *vecp,
                              DvObject_var &pos_in,
			                  DvObject_var &time_c);

QplugReturnStatus Q4SCConstPosition(RtnVector *vecp,
                                    DvObject_var &pos_in,
				                    DvObject_var &time_c);

QplugReturnStatus Q4SCCentrePosition(RtnVector *vecp,
                                    DvObject_var &pos_in_seq,
				                    DvObject_var &time_c);

void RtnVector_diag_print(RtnVector *vecp);
 
#define MAT_INV_EPS 10*DBL_EPSILON

typedef enum MAT_INV_RET_E{
MAT_INV_OK,
MAT_INV_SINGULAR,
MAT_INV_UNKNOWN
} MAT_INV_RET;

MAT_INV_RET mat_inv(double a[3][3], double inverse[3][3]);
double determinant(double mat[3][3]);
void multiply_mat(double a[3][3],double b[3][3],double prod[3][3]);
void multiply_mat_vec(double a[3][3],double b[3],double prod[3]);
void print_mat(double a[3][3]);
 
  
/**********************************************************************/
QplugReturnStatus DvQ4SC_Normal( vector<DvObject_var>& inputs, vector<DvObject_var>& outputs){
/**********************************************************************/

// define some C variables

int i,j,sign_nx;
double dt[4];
double  A[3][3], A_INV[3][3], B[3];
double  XX[3],n[3];
double drvec[4][3], signedmodX;
MAT_INV_RET ret_mat_inv;

// call-list arguments
 
DvObject_var        drin[4];        //  0-3
DvObject_var        tic;           //  4
DvObject_var        ti_tvar[4];    //  5-8 alt
DvObject_var        n_var;          //  9
DvObject_var        Vn_var;         //  10

// define data objects

RtnVector  dr[4];

QplugReturnStatus ret_status;

#ifdef Q4SC_DIAG
   printf("Q4SC Plugin Diagnostics On; undefine in .cc file and recompile\n");
   fflush(stdout);
#endif

// Ensure no input slots are empty

 if(inputs.size() != 9 ){
    QplugAppendTextDisplay("At least one input slot is empty");
    return QPLUG_FAILURE;
 }
 
// ...4 items are the s/c positions

for (i=0; i<4; i++){
  drin[i] = inputs[i];
  if (drin[i].is_nil()){
    char pos_err_string[200]; 
    sprintf(pos_err_string,"Q4SC>> Error getting pos dr[%d] object\n",i+1);
    QplugAppendTextDisplay(pos_err_string);
  }
  #ifdef Q4SC_DIAG
     printf("Q4SC got input s/c %d position\n",i);
	 fflush(stdout);
  #endif
}

tic = inputs[4];
if( tic.is_nil() || tic->not_time()) {
  QplugAppendTextDisplay ("Q4SC>> Time object for tic invalid.\n");
  return QPLUG_FAILURE;
}

#ifdef Q4SC_DIAG
   printf("Q4SC got input time interval\n",i);
   fflush(stdout);
#endif


for (i=0; i<4; i++){
  ret_status = Q4SCPosition(&(dr[i]), drin[i], tic);
  if (ret_status != QPLUG_SUCCESS){
    char pos_err_string[200]; 
    sprintf(pos_err_string,"Q4SC>> Q4SCPosition failed for sc[%d]\n",i);
    QplugAppendTextDisplay(pos_err_string);
    return QPLUG_FAILURE;
  }
  #ifdef Q4SC_DIAG
     printf("Q4SC calculated s/c %d position...\n",i);
	 fflush(stdout);
  #endif

  drvec[i][0] = dr[i].ptr->getVectorComponent(DV_X)->dbl();
  drvec[i][1] = dr[i].ptr->getVectorComponent(DV_Y)->dbl();
  drvec[i][2] = dr[i].ptr->getVectorComponent(DV_Z)->dbl();
  #ifdef Q4SC_DIAG
     printf("and filled drvec = (%g, %g,%g)\n",
               drvec[i][0],drvec[i][1],drvec[i][2]);
	fflush(stdout);
  #endif
}

// dr[i]'s now have a sequence with 1 element, the position
// of the crossing at s/c i with one timetag, and drvecs are C-arrays 

for (i=0; i<4; i++){
	ti_tvar[i]  = inputs[i+5];
	if (!ti_tvar[i].is_nil()) {
        if(ti_tvar[i]->is_dbl()) {
            dt[i] = ti_tvar[i]->dbl();
        }
        else if(ti_tvar[i]->is_time()){
            dt[i] = ti_tvar[i]->time() - tic->time();
        }
        else if(ti_tvar[i]->is_event()) {
            dt[i] = ti_tvar[i]->event().duration();
        }
        else{
            dt[i] = ti_tvar[i]->asDouble();
        }
  	} 
  	else {
    	char time_err_string[200];
    	sprintf(time_err_string, "Q4SC ERROR>> t[%d] not time or scalar object\n",i+1);
    	QplugAppendTextDisplay(time_err_string);
    	return QPLUG_FAILURE;
  	}
  #ifdef Q4SC_DIAG
     printf("Q4SC calculated s/c %d dt as %g\n",i,dt[i]);
	 fflush(stdout);
  #endif
}
// dt[i] now contains times in secs for all 4 s/c

// Now fill dr matrix and right hand sides; s/c 4 is the reference
// [it doesn't matter which s/c is the ref; there are only 3 indep pairs]

for(i=0; i<3; i++){
  for(j=0; j<3; j++){
    A[i][j] = drvec[i][j] - drvec[3][j];
  }
  B[i]      = dt[i] - dt[3];
}

#ifdef Q4SC_DIAG
   printf("Q4SC A matrix is\n");
   print_mat(A);
   fflush(stdout);
#endif
    
ret_mat_inv = mat_inv(A,A_INV);
    
switch (ret_mat_inv) {
   case MAT_INV_OK:
     break;
   case MAT_INV_SINGULAR:
     QplugAppendTextDisplay("Q4SC ERROR>> dr matrix singular\n");
     return QPLUG_FAILURE;
     break;
   case MAT_INV_UNKNOWN: default:
     QplugAppendTextDisplay("Q4SC ERROR>> Matrix Inv unknown error\n");
     return QPLUG_FAILURE;
     break;
}
#ifdef Q4SC_DIAG
   printf("Q4SC A inv matrix found successfully as\n");
   print_mat(A_INV);
   fflush(stdout);
#endif


multiply_mat_vec(A_INV, B, XX);
    
sign_nx = (XX[0] < 0 ) ? -1 : 1;  
Vn_var = new DvObject(0.0);
    
signedmodX = (float)sign_nx*sqrt(XX[0]*XX[0] + XX[1]*XX[1] + XX[2]*XX[2]);
Vn_var->dbl() = 1.0/signedmodX;

n_var  = new  DvObject(0.0, DIM3);
    
n_var->copy_xrefs_from( dr[3].ptr); // copy  frame, etc.


#ifdef Q4SC_DIAG
   printf("Q4SC normal velocity is %g\n", (float)*Vn_var);
   fflush(stdout);
#endif

for (i=0; i<3; i++) {
  n[i]  = XX[i]/signedmodX;
  n_var->dbl(i) = n[i];
}
#ifdef Q4SC_DIAG
   printf("Q4SC normal  is (%g, %g, %g)\n", n[0],n[1],n[2]);
   fflush(stdout);
#endif

// set the attributes for the output objects

n_var->change_xref(FIELDNAM, "Normal");
Vn_var->change_xref(FIELDNAM, "Vn");


n_var->change_xref("SI_conversion", "1>()");
DvObject_var sic = drin[3]->get_xref("SI_conversion");
Vn_var->change_xref("SI_conversion",sic);


n_var->change_xref(UNITS, " ");
DvObject_var units = drin[3]->get_xref(UNITS);
Vn_var->change_xref(UNITS,units);


char crossingt_str[200];
sprintf(crossingt_str,"%s", dr[3].t->time().getISOstring().c_str());
n_var->change_xref("Crossing Time",crossingt_str);
Vn_var->change_xref("Crossing Time",crossingt_str);

DvString frame = drin[3]->getFrameAttr();
n_var->setFrameAttr(frame);


outputs.push_back( n_var);

outputs.push_back( Vn_var);

// put some results to plug-in text window
char outstr[200];
sprintf(outstr,"Q4SC>> Crossing at %s\n",  crossingt_str);
QplugAppendTextDisplay(outstr);
sprintf(outstr,"Q4SC>> Normal is (%9.3g, %9.3g, %9.3g) ",
   n[0], n[1], n[2]);
QplugAppendTextDisplay(outstr);
sprintf(outstr,"Q4SC>> Velocity along normal is %g %s\n", 
    Vn_var->dbl(), Vn_var->getXrefText(UNITS).c_str() );
QplugAppendTextDisplay(outstr);

QplugAppendTextDisplay("Q4SC>> finished successfully\n");

return QPLUG_SUCCESS;

}

/**********************************************************************/
QplugReturnStatus Q4SCPosition(RtnVector *vecp,
                              DvObject_var &pos_in,
							  DvObject_var &time_c){
/**********************************************************************/

long  check_rtn;

QplugReturnStatus ret_status;

// sort out position of crossing. Input can either be a time series or
// a constant vector, so this a bit complicated. Separate routines
// return in vecp a sequence of 1 cartesian vector whose timetag
// is the time  time_c, and whose frame is
// that of pos_in.

if( pos_in->seqSize() > 1) {
  QplugAppendTextDisplay ("Q4SCCoords>> Vector Sequence. Interpolating ...\n");
  ret_status = Q4SCCentrePosition(vecp, pos_in, time_c);
  if (ret_status != QPLUG_SUCCESS){
   QplugAppendTextDisplay("Q4SCCoords>> Failed to calculate crossing position\n");
   return QPLUG_FAILURE;
  }
} 
else {
  ret_status = Q4SCConstPosition(vecp, pos_in, time_c);
  if (ret_status != QPLUG_SUCCESS){
    QplugAppendTextDisplay("Q4SC>> Failed to calc const. position\n");
    return QPLUG_FAILURE;
  }
}

return QPLUG_SUCCESS;
}

/**********************************************************************/
QplugReturnStatus  Q4SCConstPosition(RtnVector *vecp,
                                    DvObject_var &pos_in,
				                    DvObject_var &tt){
/**********************************************************************/

QplugReturnStatus ret_status;

// grab constant vector and fill vecp.

if (!pos_in->isThreeVector()){
   QplugAppendTextDisplay("Q4SC>> Const pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}

vecp->ptr = pos_in;
vecp->n_recs = 1;
vecp->t = tt;

vecp->in_frame = vecp->QrGetFrameFromAttr(pos_in);
if (vecp->in_frame == UNKNOWN){
  QplugAppendTextDisplay("Q4SC>> Const pos_in unknown frame\n");
  return QPLUG_FAILURE;
}

return QPLUG_SUCCESS;
}


/**********************************************************************/
QplugReturnStatus  Q4SCCentrePosition(RtnVector *vecp,
                                    DvObject_var &pos_in_seq,
				                    DvObject_var &time_c){
/**********************************************************************/

int i,nttin,ntrue,il,ih;
QplugReturnStatus ret_status;
double cl,ch, frac;
const char * xyz_comp[] = {"x", "y", "z"};

if (!pos_in_seq->isThreeVector()){
   QplugAppendTextDisplay("Q4SC>> Sequence pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}
// prepare vecp 
    
vecp->ptr = pos_in_seq->Join(time_c);
vecp->t = time_c;
vecp->n_recs = 1;
vecp->in_frame = vecp->QrGetFrameFromAttr(pos_in_seq);
if (vecp->in_frame == UNKNOWN){
  QplugAppendTextDisplay("Q4SC>> Sequence pos_in unknown frame\n");
  return QPLUG_FAILURE;
}


return QPLUG_SUCCESS;
}


/**************************************************************************/
void RtnVector_diag_print(RtnVector *vecp){
  printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
        vecp->IsPositionVector,vecp->in_frame,vecp->out_frame, vecp->n_recs);
  printf("First Vector: (%9.3g, %9.3g, %9.3g) at %s\n",
     vecp->ptr->dbl(0), vecp->ptr->dbl(1), vecp->ptr->dbl(2),
     vecp->t->time().getISOstring().c_str());
	 fflush(stdout);
  return;
}



MAT_INV_RET mat_inv(double a[3][3], double inverse[3][3]){

double det;


det = determinant(a);
if (fabs(det) < MAT_INV_EPS) return MAT_INV_SINGULAR;

inverse[0][0] =  ( a[1][1]*a[2][2] - a[1][2]*a[2][1] )/det;
inverse[0][1] = -( a[0][1]*a[2][2] - a[0][2]*a[2][1] )/det;
inverse[0][2] =  ( a[0][1]*a[1][2] - a[0][2]*a[1][1] )/det;
inverse[1][0] = -( a[1][0]*a[2][2] - a[1][2]*a[2][0] )/det;
inverse[1][1] =  ( a[0][0]*a[2][2] - a[0][2]*a[2][0] )/det;
inverse[1][2] = -( a[0][0]*a[1][2] - a[0][2]*a[1][0] )/det;
inverse[2][0] =  ( a[1][0]*a[2][1] - a[1][1]*a[2][0] )/det;
inverse[2][1] = -( a[0][0]*a[2][1] - a[0][1]*a[2][0] )/det;
inverse[2][2] =  ( a[0][0]*a[1][1] - a[0][1]*a[1][0] )/det;

return MAT_INV_OK;


}



double determinant(double m[3][3]){

double d;

d  = m[0][0]*m[1][1]*m[2][2] + m[0][1]*m[1][2]*m[2][0] +m [0][2]*m[1][0]*m[2][1];
d -= m[2][0]*m[1][1]*m[0][2] + m[2][1]*m[1][2]*m[0][0] + m[2][2]*m[1][0]*m[0][1];

return d;
}

void multiply_mat(double a[3][3],double b[3][3], double prod[3][3]){
int i,j,k;

for (i=0; i<3; i++){
  for (j=0; j<3; j++){
    for (prod[i][j]=0.0,k=0; k<3; k++) prod[i][j]+=a[i][k]*b[k][j];
  }
}

return;

}

void multiply_mat_vec(double a[3][3],double b[3], double prod[3]){
int i,j;

for (i=0; i<3; i++){
    for (prod[i]=0.0,j=0; j<3; j++) prod[i]+=a[i][j]*b[j];
}


return;

}

void print_mat(double a[3][3]){

int i,j;

for (i=0; i<3; i++){
   for (j=0; j<3; j++) printf("%10.4g ", a[i][j]);
   printf("\n");
   fflush(stdout);
}
return;
}


