/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    makeVectorXYZ
* 
* MODULE:	QSAS construct Vector Time Series from 3 scalars
*
* LANGUAGE:	C++
*
* FUNCTION:
*
* PURPOSE: 
* 
* DESCRIPTION:
*
* ARGUMENTS:
*    IN:
*     SCALAR_TIME_SERIES    x_in                 input x time series 
*     SCALAR_TIME_SERIES    y_in                 input y time series 
*     SCALAR_TIME_SERIES    z_in                 input z time series 
*     STRING               Frame_in          Frame attribute  
*    OUT:
*        VECTOR_TIME_SERIES ts_out
* GLOBALS:
*
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
*
* FUNCTIONS_USED:	 
*
* ALGORITHM:  
*
* HISTORY:	Tony Allen Oct 2005
*
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "DvObj.h"
#include "qdutil.h"
#include "makeVectorXYZ.h"

using namespace QSAS;

QplugReturnStatus  DvMakeVectorXYZ  ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{

DvObject_var  x_in;
DvObject_var  y_in;
DvObject_var  z_in;
DvObject_var  frame;


//	..first 3 items are input Time Series

x_in = inputs[0];
if( x_in.is_nil() || x_in->arraySize() > 1 || x_in->not_dbl()) {
   QplugAppendTextDisplay ("makeVectorXYZ>> ERROR - input x not a Scalar Time Series!\n");
	return QPLUG_FAILURE; 
}

y_in = inputs[1];
if( y_in.is_nil() || y_in->arraySize() > 1 || y_in->not_dbl()) {
   QplugAppendTextDisplay ("makeVectorXYZ>> ERROR - input y not a Scalar Time Series!\n");
	return QPLUG_FAILURE; 
}

z_in = inputs[2];
if( z_in.is_nil() || z_in->arraySize() > 1 || z_in->not_dbl()) {
   QplugAppendTextDisplay ("makeVectorXYZ>> ERROR - input z not a Scalar Time Series!\n");
	return QPLUG_FAILURE; 
}

DvObject_var x_unit = new DvObject(0.0, DIM3, 1);
x_unit->dbl(0) = 1.0;
DvObject_var y_unit = new DvObject(0.0, DIM3, 1);
y_unit->dbl(1) = 1.0;
DvObject_var z_unit = new DvObject(0.0, DIM3, 1);
z_unit->dbl(2) = 1.0;

// convert to vectors
x_in = x_in * x_unit;
y_in = y_in * y_unit;
z_in = z_in * z_unit;

// add together

DvObject_var ret = x_in + y_in;
ret += z_in;

if(ret->nError() > 0){
	for(size_t n=ret->nError(); n>0; n--) cout << ret->getError(n-1) << endl;
}


frame = inputs[3];
if( frame.is_nil() || frame->not_str()) {
   QplugAppendTextDisplay ("makeVectorXYZ>> ERROR - input frame not a String!\n");
	return QPLUG_FAILURE; 
}

// load all the attributes, and the half spacing/sample period:

 ret->change_xref("FIELDNAM", "composed vector");
 ret->setFrameAttr(frame->str());

//	..add  the output series to the list

  outputs.push_back(ret);     

// Clean up:


QplugAppendTextDisplay ("makeVectorXYZ>> data object created - exiting successfully\n");

return QPLUG_SUCCESS;

}

//-----------------------

QplugReturnStatus  DvLat_Long2UnitXYZ ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{
    
    DvObject_var  Lat_in;
    DvObject_var  Long_in;
    DvObject_var  frame;
    
    
    //	..first 3 items are input Time Series
    
    Lat_in = inputs[0];
    if( Lat_in.is_nil() || Lat_in->arraySize() > 1 || Lat_in->not_dbl()) {
        QplugAppendTextDisplay ("DvLat_Long2UnitXYZ>> ERROR - input Lat not a Scalar Time Series!\n");
        return QPLUG_FAILURE; 
    }
    
    Long_in = inputs[1];
    if( Long_in.is_nil() || Long_in->arraySize() > 1 || Long_in->not_dbl()) {
        QplugAppendTextDisplay ("DvLat_Long2UnitXYZ>> ERROR - input Long not a Scalar Time Series!\n");
        return QPLUG_FAILURE; 
    }
        
    DvObject_var R_unit = new DvObject(0.0, DIM3, 1);
    R_unit->dbl(0) = 1.0;
    DvObject_var L_unit = new DvObject(0.0, DIM3, 1);
    L_unit->dbl(1) = 1.0;
    DvObject_var P_unit = new DvObject(0.0, DIM3, 1);
    P_unit->dbl(2) = 1.0;
    
    // convert to vectors & add together

    DvObject_var ret = Lat_in * L_unit; 
    ret += Long_in * P_unit;
    ret += R_unit;
    

    ret->RLPtoXYZThis();


    if(ret->nError() > 0){
        for(size_t n=ret->nError(); n>0; n--) cout << ret->getError(n-1) << endl;
    }
    
    frame = inputs[2];
    if( frame.is_nil() || frame->not_str()) {
        QplugAppendTextDisplay ("DvLat_Long2UnitXYZ>> ERROR - input frame not a String!\n");
        return QPLUG_FAILURE; 
    }
    
    // load all the attributes, and the half spacing/sample period:
    
    ret->change_xref("FIELDNAM", "unit vector");
    ret->setFrameAttr(frame->str());
   
    
    //	..add  the output series to the list
    
    outputs.push_back(ret);     
    
    // Clean up:
    
    
    QplugAppendTextDisplay ("DvLat_Long2UnitXYZ>> data object created - exiting successfully\n");
    
    return QPLUG_SUCCESS;
    
}

//-----------------------

QplugReturnStatus  DvCoLat_Long2UnitXYZ ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{
    
    DvObject_var  CoLat_in;
    DvObject_var  Long_in;
    DvObject_var  frame;
    
    
    //	..first 3 items are input Time Series
    
    CoLat_in = inputs[0];
    if( CoLat_in.is_nil() || CoLat_in->arraySize() > 1 || CoLat_in->not_dbl()) {
        QplugAppendTextDisplay ("DvCoLat_Long2UnitXYZ>> ERROR - input Lat not a Scalar Time Series!\n");
        return QPLUG_FAILURE; 
    }
    
    Long_in = inputs[1];
    if( Long_in.is_nil() || Long_in->arraySize() > 1 || Long_in->not_dbl()) {
        QplugAppendTextDisplay ("DvCoLat_Long2UnitXYZ>> ERROR - input Long not a Scalar Time Series!\n");
        return QPLUG_FAILURE; 
    }
    
    DvObject_var R_unit = new DvObject(0.0, DIM3, 1);
    R_unit->dbl(0) = 1.0;
    DvObject_var T_unit = new DvObject(0.0, DIM3, 1);
    T_unit->dbl(1) = 1.0;
    DvObject_var P_unit = new DvObject(0.0, DIM3, 1);
    P_unit->dbl(2) = 1.0;
    
    // convert to vectors & add together
    
    DvObject_var ret = CoLat_in * T_unit; 
    ret += Long_in * P_unit;
    ret += R_unit;
    
    
    ret->RTPtoXYZThis();
    
    
    if(ret->nError() > 0){
        for(size_t n=ret->nError(); n>0; n--) cout << ret->getError(n-1) << endl;
    }
    
    frame = inputs[2];
    if( frame.is_nil() || frame->not_str()) {
        QplugAppendTextDisplay ("DvCoLat_Long2UnitXYZ>> ERROR - input frame not a String!\n");
        return QPLUG_FAILURE; 
    }
    
    // load all the attributes, and the half spacing/sample period:
    
    ret->change_xref("FIELDNAM", "unit vector");
    ret->setFrameAttr(frame->str());
    
    
    //	..add  the output series to the list
    
    outputs.push_back(ret);     
    
    // Clean up:
    
    
    QplugAppendTextDisplay ("DvCoLat_Long2UnitXYZ>> data object created - exiting successfully\n");
    
    return QPLUG_SUCCESS;
    
}

