
// qpui_trace.cc
//
// Plot Interface Implementation - Trace Object
//
// Janet Barnes
//
// July 2002
//
//  Tony Allen
//  Jan 2008, Feb 2013

#include "qpui.h"
#include "qwui.h"
#include "qdutil.h"

using namespace QSAS;

extern QPalette QPUI_normal_palette;
extern int qsas_restore_release;

// local wrapper for plpoin() for speed
void Q_plpoin(PLINT n, PLFLT *x, PLFLT *y, PLINT code, const char *ch, PLFLT size);


// ampersand #[0x0040] ok, large
// paragraph #[0x00a7] ok, large
// filled diamond #[0x25c6], close
const char * symbolUnicodeConversion(int index) // Index in combo box to plplot index
{
	switch(index)
	{
		case 1:
			return (const char*)Q_UNICODE_POINT;
			break;
		case 2:
			return (const char*)Q_UNICODE_PLUS;
			break;
		case 3:
			return (const char*)Q_UNICODE_ASTERISK;
			break;
		case 4:
			return (const char*)Q_UNICODE_CIRCLE;
			break;
		case 5:
			return (const char*)Q_UNICODE_CROSS;
			break;
		case 6:
			return (const char*)Q_UNICODE_SQUARE;
			break;
		case 7:
			return (const char*)Q_UNICODE_TRIANGLE;
			break;
		case 8:
			return (const char*)Q_UNICODE_CIRCLED_POINT;
			break;
		case 9:
			return (const char*)Q_UNICODE_FILLED_DIAMOND;
			break;
		case 10:
			return (const char*)Q_UNICODE_DIAMOND;
			break;
		case 11:
			return (const char*)Q_UNICODE_STAR;
			break;
		case 12:
			return (const char*)Q_UNICODE_WIDECROSS;
			break;
		case 13:
			return (const char*)Q_UNICODE_FILLEDSQUARE;
			break;
		case 14:
			return (const char*)Q_UNICODE_FILLEDCIRCLE;
			break;
		case 15:
			return (const char*)Q_UNICODE_FILLEDSTAR;
			break;
		case 16:
			return (const char*)Q_UNICODE_LEFT_ARROW;
			break;
		case 17:
			return (const char*)Q_UNICODE_RIGHT_ARROW;
			break;
		case 18:
			return (const char*)Q_UNICODE_UP_ARROW;
			break;
		case 19:
			return (const char*)Q_UNICODE_DOWN_ARROW;
			break;
		case 20:
			return (const char*)Q_UNICODE_CUSTOM_SHAPE;
			break;
		case 0:
		default:
			return (const char*)Q_UNICODE_NOSYMBOL;
			break;
	}
}


// --------------------------
//     TraceItem CLASS
//     ------------------

PlotItem::PlotItem(const QString& name, PanelItem* parent, const QString& WLcontext, int position) : QpuiItem(name, parent)
{
	if(parent==NULL)
	{
		std::cerr << "PlotItem::PlotItem: creating PlotItem with NULL parent" << std::endl;
		return;
	}
	
	if(name.isEmpty())
	{
		setName(QString("Plot %1").arg(parent->childCount()));
	}
	
	tabWidget=parent->tabWidget;

	if(WLcontext.isEmpty())
	{
		if( validObjName(name.toLatin1().data()) ) this->WLcontext=name;
		else this->WLcontext="";
	}
	else
	{
		this->WLcontext=WLcontext;
	}
	
}

PlotItem::PlotItem(PlotItem* plot, PanelItem* parent, int position) : QpuiItem("", parent)
{
	if(parent==NULL)
	{
		std::cerr << "PlotItem::PlotItem: creating PlotItem with NULL parent" << std::endl;
		return;
	}
	
	setName(plot->getName());
	
	tabWidget=parent->tabWidget;

	WLcontext=plot->WLcontext;
}

PlotItem::~PlotItem()
{

	foreach(PlotItem* i, linkedPlots)
	{
		i->linkedPlots.removeAll(this);
	}
}


void PlotItem::setLinks(QList <PlotItem*>& linkList)
{
	linkedPlots = linkList;
}

// ----------------------------------

void PlotItem::removeLink(PlotItem* plot_item)
{
	linkedPlots.removeOne(plot_item);	
}

// ------------------------------------------

QpFrameType PlotItem::frameType()
{
	if(panel()) return panel()->frameType();
	return QP_TS;
}

// ---------------TraceItem class--------------

TraceItem::TraceItem(const QString& name, PanelItem* parent, const QString& WLcontext, QpItemType plotType, int position):
		PlotItem(name, parent, WLcontext, position)
{
	if(QuSlotDebug()) cout << "TraceItem::TraceItem '" << name.toLatin1().data() << "'" << endl;

	
	C1_colour = QColor(0,0,0);
	C2_colour = QColor(255,0,0);
	C3_colour = QColor(0,130,0);
	C4_colour = QColor(150,0,255);
	R_colour  = QColor(0,0,0);
	X_colour  = QColor(255,0,0);
	Y_colour  = QColor(0,180,0);
	Z_colour  = QColor(0,0,180);
	T_colour  = QColor(150,0,0);
	P_colour  = QColor(0,130,0);
	L_colour  = QColor(150,0,150);
	userLineColour = false;
	userSymbolColour = false;
	I_Type = plotType;

	if(name.isEmpty()) {
		setNextName();
	}

	DataTab = new QpDataTab(tabWidget, "DataTab", this, plotType);
	DataTab->hide();
	
	TraceTab = new QpTraceTab(tabWidget, "TraceTab", this);
	TraceTab->hide();

	
	QColor black = QColor(0, 0, 0);
	if(WLcontext.contains("_C1_") || WLcontext.contains("_dr1_")|| WLcontext.contains("_at1_")){
		TraceTab->setLineColour(C1_colour);
		TraceTab->setSymbolColour(C1_colour);
	}
	else if(WLcontext.contains("_C2_") || WLcontext.contains("_dr2_")|| WLcontext.contains("_at2_")){
		TraceTab->setLineColour(C2_colour);
		TraceTab->setSymbolColour(C2_colour);
	}
	else if(WLcontext.contains("_C3_") || WLcontext.contains("_dr3_")|| WLcontext.contains("_at3_")){
		TraceTab->setLineColour(C3_colour);
		TraceTab->setSymbolColour(C3_colour);
	}
	else if(WLcontext.contains("_C4_") || WLcontext.contains("_dr4_")|| WLcontext.contains("_at4_")){
		TraceTab->setLineColour(C4_colour);
		TraceTab->setSymbolColour(C4_colour);
	}
	else{
		TraceTab->setLineColour(black);
		TraceTab->setSymbolColour(black);
	}


	// set Minimum Size for TabWidget
	tabWidget->adjustSize();
   
	_y_obj_id_getYRange = 0;
	_x_obj_id_getXRange = 0;
	
	setIcon(I_Type);

}

TraceItem::TraceItem(TraceItem* trace, PanelItem* parent, int position):
		PlotItem(trace, parent, position)
{
	C1_colour = QColor(0,0,0);
	C2_colour = QColor(255,0,0);
	C3_colour = QColor(0,130,0);
	C4_colour = QColor(150,0,255);
	R_colour  = QColor(0,0,0);
	X_colour  = QColor(255,0,0);
	Y_colour  = QColor(0,180,0);
	Z_colour  = QColor(0,0,180);
	T_colour  = QColor(150,0,0);
	P_colour  = QColor(0,130,0);
	L_colour  = QColor(150,0,150);

	I_Type = trace->I_Type;

	userLineColour = trace->userLineColour;
	userSymbolColour = trace->userLineColour;
	DataTab = new QpDataTab(trace->DataTab, this);

    DataTab->timeAsSec_toggle->setVisible(frameType()==QP_XY);
    DataTab->timeAsSec_toggle->setChecked(trace->DataTab->timeAsSec_toggle->isChecked());
	DataTab->hide();

	
	_line_colour = trace->_line_colour;
	_symbol_colour = trace->_symbol_colour;
	
	// set Minimum Size for TabWidget
	tabWidget->adjustSize();
   
	_y_obj_id_getYRange = 0;
	_x_obj_id_getXRange = 0;
	
	setIcon(I_Type);
    
    TraceTab = new QpTraceTab(trace->TraceTab, this);
    TraceTab->hide();
    TraceTab->_c_data_fld->reValidate(true);

}

TraceItem::~TraceItem()
{
	if(DataTab) delete DataTab;
	if(TraceTab) delete TraceTab;
	foreach(PlotItem* i, linkedPlots)
	{
		i->linkedPlots.removeAll(this);
	}
}

// -------------------------------------------

void TraceItem::setDataYObj(QString txt)
{
	if(txt.isEmpty()) return;

	DataTab->_y_data_fld->setText(txt);
	DataTab->_y_data_fld->reValidate(true);
}

// -------------------------------------------

void TraceItem::setDataXObj(QString txt)
{
	if(txt.isEmpty()) return;

	DataTab->_x_data_fld->setText(txt);
	DataTab->_x_data_fld->reValidate(true);
}

// ----------------------------------

void TraceItem::setDefaultOptions()
{
	DataTab->timeAsSec_toggle->setChecked(false);
	DataTab->_show_legend_fld->setChecked(true);
	
	TraceTab->_line_style_fld->setCurrentIndex(QP_SOLID);
	TraceTab->_line_width_fld->setValue(1);
	TraceTab->_connect_style_fld->setCurrentIndex(QPUI_STRAIGHT);
	
	QPalette line_palette = QPalette(QPUI_normal_palette);
	line_palette.setColor(QPalette::Active ,QPalette::ButtonText, Qt::black);
	line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, Qt::black);
	TraceTab->_line_colour_fld->setPalette(line_palette);
	_line_colour = Qt::black;

	TraceTab->_symbol_shape_fld->setCurrentIndex(QPUI_NOSYMBOL);
	TraceTab->_symbol_size_fld->setValue(1.0);
	TraceTab->_symbol_colour_fld->setPalette(line_palette);
	_symbol_colour = Qt::black;
	
	updateAxes();
}

// ------------------------------------

void TraceItem::updateAxes()
{
	if(QuSlotDebug()) cout << "TraceItem::updateAxes " << endl;
	
	frame()->updateAxes();

}

// ------------------------------------

void TraceItem::clearObj_id()
{
   // called when we need to force a range update
	_y_obj_id_getYRange = 0;
	_x_obj_id_getXRange = 0;
}

// -------------------------------------------

void TraceItem::showTabs()
{
	if(QuSlotDebug()) cout << "TraceItem::showTabs '" << this->getName().toLatin1().data() << "'" << endl;
	
	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--)
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}


	tabWidget->addTab(DataTab, "Data");

	tabWidget->addTab(TraceTab, "Trace");
	// value return is always shown
	tabWidget->addTab(page()->ValueReturnTab, "Value Return");
 	if(Tree()->I_LastTab == QP_TRACE_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(TraceTab));
 	else tabWidget->setCurrentIndex(tabWidget->indexOf(DataTab));
	
}

// -------------------------------------------

void TraceItem::replaceData(QString obj_name)
{
	if(obj_name.isEmpty()) return;
	if(QuSlotDebug()) cout << "TraceItem::replaceData '" <<  obj_name.toLatin1().data() << "'" << endl;

	DvString oldCompText = compText( DataTab->_y_data_fld->text().toLatin1().data() ); 

	if(oldCompText.empty() && !(DataTab->_y_data_fld->text().contains('?')) )
	{
		if(this->frameType() == QP_TS)
		{
			DataTab->_y_data_fld->setDefaultComp(QS_0D, QS_SEQ, obj_name); // set data slot
		}
		else
		{
			DataTab->_y_data_fld->setDefaultComp(QS_1D, QS_NOT_SEQ, obj_name); // set data slot
		}
	}
	else
	{
		// setTextKeepComp can change text for comp or substitute string
		DataTab->_y_data_fld->setTextKeepComp(obj_name);  // set data slot	
	}

	this->setName(DataTab->_y_data_fld->text());               // set item name

	this->WLcontext = baseObjectText(DataTab->_y_data_fld->text().toLatin1().data()).c_str();
	DataTab->_y_data_fld->reValidate(true);

}

// -------------------------------------------

void TraceItem::replaceXData(QString obj_name)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceXData" << endl;
	if(obj_name.isEmpty()) return;

	DataTab->_x_data_fld->setTextKeepComp(obj_name);  // set data slot
	// setTextKeepComp can change text for comp or substitute string

}

// -------------------------------------------
void TraceItem::replaceTIdata(QString tivl)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceTI with '" << tivl.toLatin1().data() << "'" << endl;
	if(tivl.isEmpty()) return;
	
	DataTab->_mask_data_fld->setText(tivl);  // set TI slot
	DataTab->_mask_data_fld->reValidate(true);


}

// -------------------------------------------
void TraceItem::replaceTIcomp(QString comp)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceTI with '" << comp.toLatin1().data() << "'" << endl;
	if(comp.isEmpty()) return;
	
	DvString newName = baseObjectText(DataTab->_mask_data_fld->text().toLatin1().data());
	newName += comp.toLatin1().data();
	DataTab->_mask_data_fld->setText(newName.c_str());  // set TI slot

	DataTab->_mask_data_fld->reValidate(true);

}

// -------------------------------------------

void TraceItem::replaceComponent(QString comp)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceComponent with '" << comp.toLatin1().data() << "'" << endl;
	if(comp.isEmpty()) return;
	
	DvString newName = baseObjectText(DataTab->_y_data_fld->text().toLatin1().data());
	newName += getNewComp(DataTab->_y_data_fld->text().toLatin1().data(), comp.toLatin1().data());
	
	DataTab->_y_data_fld->setText(newName.c_str());  // set data slot
	DataTab->_y_data_fld->reValidate(true);
	this->setName(newName.c_str());               // set item name 

}

// -------------------------------------------

void TraceItem::replaceXComponent(QString comp)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceXComponent" << endl;
	if(comp.isEmpty()) return;
	
	DvString newName = baseObjectText(DataTab->_x_data_fld->text().toLatin1().data());
	newName += comp.toLatin1().data();
	DataTab->_x_data_fld->setText(newName.c_str());  // set data slot
	
}
// -------------------------------------------

void TraceItem::replace_C_Component(QString comp)
{
	if(QuSlotDebug()) cout << "TraceItem::replace_C_Component" << endl;
	if(comp.isEmpty()) return;
	
	DvString newName = baseObjectText(TraceTab->_c_data_fld->text().toLatin1().data());
	newName += comp.toLatin1().data();
	TraceTab->_c_data_fld->setText(newName.c_str());  // set data slot
	TraceTab->_c_data_fld->reValidate(true);
	
}

// -------------------------------------------
void TraceItem::replaceCdata(QString base)
{
	if(QuSlotDebug()) cout << "TraceItem::replaceCdata with '" << base.toLatin1().data() << "'" << endl;
	if(base.isEmpty()) return;
	
	TraceTab->_c_data_fld->setText(base);  
	TraceTab->_c_data_fld->reValidate(true);


}

// -------------------------------------------

void TraceItem::refreshTabs()
{
	if(QuSlotDebug()) cout << "QpTraceItem::refreshTabs" << endl;

	DataTab->_y_data_fld->reValidate();

}


// ------------------------------------------

void TraceItem::changeFrameType(int frameType)
{
	// called when Frame changes type XY <-> TS
	
	if(frameType == QP_TS)
	{
	
		DataTab->_x_label_fld->setVisible(false);
		DataTab->_x_data_fld->setVisible(false);
		DataTab->timeAsSec_toggle->setVisible(false);
        DataTab->_gap_detection_fld->setVisible(true);
        DataTab->_gap_params_container->setVisible(true);

		if(DataTab->I_PlotType == QP_CONST_XT)
		{
			// set to const t trace

			DataTab->_y_label_fld->setText("Time Constant: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_TIME, QS_0D, QS_NOT_SEQ, QS_ANY));
			DataTab->_show_legend_fld->setChecked(false);
		}
		else if(DataTab->I_PlotType == QP_CONST_Y)
		{
			DataTab->_y_label_fld->setText("Y Constant: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
			DataTab->_show_legend_fld->setChecked(false);
		}
		else if(DataTab->I_PlotType == QP_REG_LINE)
		{
			DataTab->_y_label_fld->setText("Y: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_show_legend_fld->setChecked(false);
		}
		else
		{
			DataTab->_y_label_fld->setText("Y: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_TT));
			DataTab->_show_legend_fld->setChecked(true);
		}
	}
	else
	{
		DataTab->timeAsSec_toggle->setVisible(true);
        DataTab->_gap_detection_fld->setVisible(false);
        DataTab->_gap_params_container->setVisible(false);
		
		if(DataTab->I_PlotType == QP_CONST_XT)
		{
			DataTab->_y_label_fld->setText("X Constant: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));

			DataTab->_x_label_fld->hide();
			DataTab->_x_data_fld->hide();
			DataTab->timeAsSec_toggle->hide();
			DataTab->_show_legend_fld->setChecked(false);
		}
		else if(DataTab->I_PlotType == QP_CONST_Y)
		{
			DataTab->_y_label_fld->setText("Y Constant: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_NOT_SEQ, QS_ANY));
			
			DataTab->_x_label_fld->hide();
			DataTab->_x_data_fld->hide();
			DataTab->timeAsSec_toggle->hide();
			DataTab->_show_legend_fld->setChecked(false);
		}
		else if(DataTab->I_PlotType == QP_REG_LINE)
		{
			DataTab->_y_label_fld->setText("Y: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_y_data_fld->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
			
			DataTab->_x_label_fld->show();
			DataTab->_x_data_fld->show();
			DataTab->_x_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_y_data_fld->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
			DataTab->timeAsSec_toggle->show();
			DataTab->_show_legend_fld->setChecked(false);
		}
		else
		{
			DataTab->_y_label_fld->setText("Y: ");
			DataTab->_y_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_y_data_fld->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));

			DataTab->_x_label_fld->show();
			DataTab->_x_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_ANY));
			DataTab->_y_data_fld->addAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_1D, QS_NOT_SEQ, QS_ANY));
			DataTab->_x_data_fld->show();
			DataTab->timeAsSec_toggle->show();
			DataTab->_show_legend_fld->setChecked(false);
		}
		

	}
	DataTab->setDataTypeTitle();
	
	// force reevaluation
	DataTab->_y_data_fld->reValidate(true);

}

// -------------------------------------------------

void TraceItem::getLabel(const QpAxes& axis, DvString& label)
{
	// fetch label for axis

	DvObject_var obj;
	DvString default_label="";
	switch(axis)
	{
		case X_AXIS:
			// if time series plot default label is Universal Time
			if(frameType() == QP_TS)
			{
				label = "Universal Time";
				return;
			}
			if(DataTab->timeAsSec_toggle->isChecked())
			{
				label="Seconds";
				return;
			}
			// otherwise fetch x object and check if object needs refreshing
			label = "X Axis";
			obj = _xdo;
			if(obj.is_nil() )
			{
				if( !_ydo.is_nil() && _ydo->xref_exists(DEPEND_1) )  obj = _ydo->get_xref(DEPEND_1);
			}
			break;

		case Y_AXIS:
			// fetch y object and check if object needs refreshing
			label="Y Axis";
			obj = _ydo;
			break;

		default:
			return;
	}
	if(obj.is_nil())  return;

  
	// if object found set default label from LABLAXIS attribute if provided
	if(obj->xref_exists(LABLAXIS))
	{
  	  	DvObject_var xref=obj->get_xref(LABLAXIS);

		if(xref->arraySize() == 1){
  	  		default_label = xref->asStr();
		}
		else {
  	  		valarray <DvString> ms_data = xref->toStr();
			default_label = ms_data[0];
			for( int i=1;i<ms_data.size();i++){
				default_label = default_label.common(ms_data[i]);
			}
		}
	}
	else default_label = "";
	
	
	// if not use FIELDNAM attribute
	if(default_label.empty()) default_label = obj->getXrefText(FIELDNAM);
	
	// last resort use object name
	if(default_label.empty()) default_label = obj->getXrefText(OBJECT_NAME);
	if( !default_label.empty()) label = default_label.c_str();
  
	// attach units
	DvString units = obj->getXrefText(UNITS);
	units.trimmed();
	
	if( !units.empty() )
	{
		label += " (";
		label += units;
		label += ")";
	}
	
	// convert theta and phi
	label = label.replace("theta","#gh");
	label = label.replace("phi","#gf");

}


DvString TraceItem::getLegendText()
{
	// fetch legend text
	DvString legend_text = " "; // ensure ret string is safe
	
	if(_ydo.is_nil())  return legend_text;
  
	if(DataTab->I_PlotType == QP_REG_LINE){

		if(_xdo.is_nil()) return legend_text;
		
		if(frameType() == QP_TS){
			linFit_var = _ydo->regression(_xdo);
			DataTab->_legend_text_fld->setToolTip("Slope m, Intercept b (at plot start time), Correlation Coeff r (0 bad, 1 good) ");
		}
		else{
			linFit_var = _ydo->regression(_xdo);
			DataTab->_legend_text_fld->setToolTip("Slope m, Intercept b, Correlation Coeff r (0 bad, 1 good) ");
		}
		
		if(linFit_var.is_nil()) return legend_text;
		
		QString params;
		
		DvObject_var mObj = linFit_var->get_xref("Slope");
		if( !mObj.is_nil() ) {
			params = QString("m=%1 ").arg(mObj->asDouble(0), 0, 'e', 2);
			if(mObj->xref_exists(UNITS)) {
				DvString units = mObj->getXrefText(UNITS);
				params += units.c_str();
			}
		}
		DvObject_var bObj = linFit_var->get_xref("Intercept");
		if( !bObj.is_nil() ) {
			params += QString(",  b=%1 ").arg(bObj->asDouble(0), 0, 'e', 2);
			if(bObj->xref_exists(UNITS)) {
				DvString units = bObj->getXrefText(UNITS);
				params += units.c_str();
			}
		}
		
		DvObject_var rObj = linFit_var->get_xref("CoeffCorrel");
		if( !rObj.is_nil() ) params += QString(",  r=%1").arg(fabs(rObj->asDouble(0)), 0, 'e', 2);

		legend_text = string(params.toLatin1().data());

		return legend_text;
	}
	
	DataTab->_legend_text_fld->setToolTip(" ");

	
	DvString default_label = _ydo->getXrefText(LABLAXIS);
	// if not use FIELDNAM attribute
	if(default_label.empty()) default_label = _ydo->getXrefText(FIELDNAM);
	// else use object name
	if(default_label.empty()) default_label = _ydo->getXrefText(OBJECT_NAME);

	if(!default_label.empty()) legend_text = default_label;
	else legend_text = "";
	
	// attach units
	DvString units = _ydo->getXrefText(UNITS);
	units.trimmed();
	if(!units.empty())
	{
		legend_text += " (";
		legend_text += units;
		legend_text += ")";
	}

	// and spacecraft id
	DvString scid = _ydo->getXrefText(SC_ID);
	scid.trimmed();
	if(!scid.empty())
	{
		legend_text += " SC=";
		legend_text += scid;
	}
	return legend_text;
}

void TraceItem::getFrameName(string& name)
{
	// fetch name of trace object (for display in listview)

	name = DataTab->_y_data_fld->text().toLatin1().data();
	int plotType = I_Type;
	
	if(plotType == QP_CONST_XT && frameType() == QP_TS)
	{
		DvTime time_val;
		if(time_val.setFromISOstring(name.c_str()))
			name="T="+string(time_val.getISOstring(0));
		else
			name="Const T Trace";
		return;
	}
	if(plotType == QP_CONST_Y )
	{
		bool okay;
		double d = QString(name.c_str()).toDouble(&okay);
		if(okay)
		{
			name = "Constant=";
			name += name.c_str();
		}
		else
			name="Const Y Trace";
		return;
	}
	if(plotType == QP_CONST_XT && frameType() == QP_XY )
	{
		bool okay;
		double d = QString(name.c_str()).toDouble(&okay);
		if(okay)
		{
			name = "Constant=";
			name += name.c_str();
		}
		else
			name="Const X Trace";
		return;
	}
	if(plotType == QP_TRACE && frameType() == QP_TS)
	{
		if(name=="")  name="TS-Trace";
		return;
	}
	if(plotType == QP_TRACE && frameType() == QP_XY)
	{
		if(name=="") name="XY-Trace";
		return;
	}
	if(plotType == QP_REG_LINE && frameType() == QP_XY)
	{
		if(name=="") name="XY-Linear Fit";
		return;
	}
	if(plotType == QP_REG_LINE && frameType() == QP_TS)
	{
		if(name=="") name="TS-Linear Fit";
		return;
	}
}

DvString TraceItem::getTitle()
{
	// fetch title for current trace object

	int plotType = I_Type;

	DvString title ="";


	DvObject_var dobj = getWLObject( DataTab->_y_data_fld->text().toLatin1().data() );
	if(!dobj.is_nil()) title = dobj->getXrefText(FIELDNAM);

	
	if(title.empty()) {
		 title = DataTab->_y_data_fld->text().toLatin1().data();
		 title = title.afterR('/');
	}
	
	if(frameType() == QP_XY && plotType == QP_TRACE)
	{
		if(DataTab->timeAsSec_toggle->isChecked() ) return title;
		
		DvString x_title;
		if(!_xdo.is_nil())  x_title = _xdo->getXrefText(FIELDNAM);
  
		if(x_title.empty()) x_title += DataTab->_x_data_fld->text().toLatin1().data();
		if(!x_title.empty() && !title.empty() )  title += " vs ";
		title += x_title;
	}
	return title;
}

// --------------------------------

int TraceItem::getXmetaIsLog()
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::getXmetaIsLog \n" );
		fflush(stdout);
	}
	
	int isLog=0;
	// find SCALETYP
		
	if (!_xdo.is_nil() && _xdo->xref_exists(SCALETYP) )
	{
		DvString scaleType = _xdo->getXrefText(SCALETYP);
		if(!scaleType.empty() )
		{
			if(scaleType == "LOG")  isLog = 1;
		}
	}
	return isLog;
}

// --------------------------------

void TraceItem::get_dobjs()
{
	if ( QuSlotDebug() )
	{
		printf( " TraceItem::get_dobjs \n" );
		fflush(stdout);
	}
	_ydo = DataTab->_y_data_fld->getObject();
	if(_ydo.is_nil()) return;

	
	if(frameType() == QP_TS)
	{
		// if time series fetch timetags
		_xdo = _ydo->getTimeTags();
	}
	else
	{
		if(DataTab->timeAsSec_toggle->isChecked())
		{
			DvObject_var tt = _ydo->getTimeTags();
			DvObject_var tOffset = DataTab->_x_data_fld->getObject();
			if(tt.is_nil() || tOffset.is_nil()) _xdo = tt;
			else
			{
				DvTime t0 = tOffset->asTime(0);
				DvObject_var ss = new DvObject(0.0, tt->seqSize());
				for(int i=0; i<tt->seqSize(); i++)
				{
					ss->dbl(i) = tt->time(i).getDiffSecs(t0);
				}
				_xdo = ss;
			}
		}
		else
		{
			_xdo = DataTab->_x_data_fld->getObject();

			if( _xdo.is_nil() )
			{
				if( _ydo->seqSize() > 1)
				{
					// use DEPEND_0 for data sequence
					if(_ydo->xref_exists(DEPEND_0) )  _xdo = _ydo->get_xref(DEPEND_0);

				}
				else
				{
					// try Depend_1 for 1D array
					if( _ydo->xref_exists(DEPEND_1) )  _xdo = _ydo->get_xref(DEPEND_1);
				}
			}
		}
	}
	
	// Apply Time Interval Mask if set
	DvObject_var TImaskObj = DataTab->_mask_data_fld->getObject();
	if( TImaskObj.is_ok()) {
		DvEvent range = TImaskObj->getTimeRange();
		_xdo = _xdo->subSequence(range);
		_ydo = _ydo->subSequence(range);
	}
	
	return;
}

// --------------------------------

void TraceItem::updateColourBar()
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::updateColourBar() \n" );
		fflush(stdout);
	}
	_cdo = TraceTab->_c_data_fld->getObject();
	if(_cdo.is_nil() || _ydo.is_nil() ) return;

	
	// Apply Time Interval Mask if set
	DvObject_var TImaskObj = DataTab->_mask_data_fld->getObject();
	if( TImaskObj.is_ok()) {
		DvEvent range = TImaskObj->getTimeRange();
		_cdo = _cdo->subSequence(range);
	}

	// is object joined
	if( _ydo->isJoined(*_cdo) ) {
		_cdo = _cdo->Join(_ydo, false); // no xrefs (arg = false)
		if( _cdo.is_nil() ) return;
	}

	
	if( TraceTab->c_range_auto->isChecked() ){
		// set min/max
		double min, max;
	
		bool res = _cdo->getDataRange(min, max);
		
		if(res)
		{
			// store values 
			TraceTab->_c_min_fld->setText(QString("%1").arg(min));
			TraceTab->_c_max_fld->setText(QString("%1").arg(max));
		}
	}
	
	// set coour array
	
	// set Label
	if(TraceTab->userColLab == false) {
		DvString zlabel;
	
		if (_cdo->xref_exists("LABLAXIS")) {
			zlabel = _cdo->getXrefText("LABLAXIS");
		}
		else zlabel = "z";
	
		if (_cdo->xref_exists("UNITS")){
			zlabel += " (";
			zlabel += _cdo->getXrefText("UNITS");
			zlabel += ")";
		}

		TraceTab->colLabSlot->setText(zlabel.c_str());
	}

	
	return;
}

// --------------------------------

bool TraceItem::getXmetaRange(double& min, double& max)
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::getXmetaRange \n" );
		fflush(stdout);
	}

  // gets x scalemin and scalemax values for this trace object
  // returns true if found
  // call with xmin and xmax set to values of range fields
  
	int plotType = I_Type;
	bool rangeFound = false;
	if(plotType == QP_CONST_Y || plotType == QP_CONST_XT)  return false;
	
	if(frameType() == QP_TS)
	{
		// if time series fetch timetags
		DvObject_var tdo = _ydo->getTimeTags();
		// Use 1st and last for speed, getXRange finds true min/max if non-monotonic (slower)
		if(tdo.is_nil() || tdo->seqSize() < 2) return false;
		min = tdo->time(0).getEpoch2000();
		max = tdo->time(tdo->seqSize()-1).getEpoch2000();
		rangeFound = true;
	}
	else if( frameType() == QP_XY)
	{
		if( !(DataTab->timeAsSec_toggle->isChecked()) )
		{
			if(_xdo.is_nil() ) return false;
			if(_xdo->xref_exists(SCALEMIN) && _xdo->xref_exists(SCALEMAX))
			{
				DvObject_var minObj = _xdo->get_xref(SCALEMIN);
				DvObject_var maxObj = _xdo->get_xref(SCALEMAX);
				if(!minObj.is_nil() && !maxObj.is_nil() )
				{
					min = minObj->asDouble(0);
					max = maxObj->asDouble(0);
					rangeFound =  true;
				}
			}
		}
	}

	if (rangeFound)
	{
		// set current choice for use in y range
		_x_range_min = min;
		_x_range_max = max;
		return rangeFound;
	}
	else	return getXRange(min, max);

}

// --------------------------------

bool TraceItem::getXRange(double& min, double& max)
{
	if ( QuSlotDebug() )
	{
		fflush(stdout);
	}

  // gets x min and max values for this trace object
  // returns true if range calculated
  // call with xmin and xmax set to values of range fields
  
	int plotType = I_Type;

	if(plotType == QP_CONST_XT && frameType() == QP_XY)
	{
		// In this case ydo is x value
		// ----------------------------
		if(!_ydo.is_nil())
		{
			min=_ydo->asDouble(0);
			max=min;
			return true;
		}
		else return false;
		
	}
	else if(plotType == QP_CONST_XT && frameType() == QP_TS)
	{
		if(_ydo->is_time())
		{
			min = _ydo->time(0).getEpoch2000();
			max = min;
			return true;
		}
		else return false;
	}
		
	if(_xdo.is_nil()) return false;

	int obj_id = _xdo->get_id();

	// check if object id has changed
	if(_x_obj_id_getXRange == obj_id)
	{
		// range has already been calculated just get previously stored values for x min and x max
		min = _x_from;
		max = _x_to;
		return true;
	}
	else
	{
        // if yes, refresh id
        _x_obj_id_getXRange = obj_id;

		// recalculate range for x data object
		
		if( _xdo->getDataRange(min, max) )
		{
			// store values 
			_x_from = min;
			_x_to = max;
			return true;
		}
		
		return false;
	}
  
}

// --------------------------------

int TraceItem::getYmetaIsLog()
{
	if ( QuSlotDebug() )
	{
		printf( " QpTraceItem::getYmetaIsLog \n" );
		fflush(stdout);
	}
	
	int isLog=0;
	// find SCALETYP
		
	if (!_ydo.is_nil() && _ydo->xref_exists(SCALETYP) )
	{
		DvString scaleType = _ydo->getXrefText(SCALETYP);
		if(!scaleType.empty() )
		{
			if(scaleType == "LOG")  isLog = 1;
		}
	}
	return isLog;
}

//--------------------------

bool TraceItem::getYmetaRange(double& ymin, double& ymax)
{
	// gets y min and max values for all y values
	if ( QuSlotDebug() )
	{
		printf( "QpTraceItem::getYmetaRange (yobj is nil? %d)\n" ,  _ydo.is_nil() );
		fflush(stdout);
	}
	
	bool rangeFound = false;

	int plotType = I_Type;

	if( _ydo.is_nil() ) return false; // can go no further
	
	if(_ydo->xref_exists(SCALEMIN) && _ydo->xref_exists(SCALEMAX) )
	{
		DvObject_var minObj = _ydo->get_xref(SCALEMIN);
		DvObject_var maxObj = _ydo->get_xref(SCALEMAX);
		
		if( !minObj.is_nil() && !maxObj.is_nil() )
		{
			ymin = minObj->asDouble(0);
			ymax = maxObj->asDouble(0);

			rangeFound = true;
		}
	}

	// choose default spacecraft colour
	if (_ydo->xref_exists(SC_ID) )
	{
		DvString sc_id_str = _ydo->getXrefText(SC_ID);
		if(!sc_id_str.empty() )
		{
			if(sc_id_str == "1" || sc_id_str == "C1")
			{
				if(!userLineColour) TraceTab->setLineColour(C1_colour);
				if(!userSymbolColour) TraceTab->setSymbolColour(C1_colour);
			}
			if(sc_id_str == "2" || sc_id_str == "C2")
			{
				if(!userLineColour) TraceTab->setLineColour(C2_colour);
				if(!userSymbolColour) TraceTab->setSymbolColour(C2_colour);
			}
			if(sc_id_str == "3" || sc_id_str == "C3")
			{
				if(!userLineColour) TraceTab->setLineColour(C3_colour);
				if(!userSymbolColour) TraceTab->setSymbolColour(C3_colour);
			}
			if(sc_id_str == "4" || sc_id_str == "C4")
			{
				if(!userLineColour) TraceTab->setLineColour(C4_colour);
				if(!userSymbolColour) TraceTab->setSymbolColour(C4_colour);
			}
		}
		
	}
	if (rangeFound) return rangeFound;
	else{
		bool autoFound = getYRange(ymin, ymax);
		if (!panel()->yAxisTab->_axis_y_nudge_fld->isChecked()) return autoFound;
		
		// range is not nudged by panel getYmetaRange, so do it here
		// This is not effective, however, if each trace in a panel 
		// has a small range relative to panel range
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			double nudgeUp = pow(fabs(ymax / ymin), 0.015);
			double nudgeDown = pow(fabs(ymax / ymin), 0.03);
			ymax *= nudgeUp;
			ymin /= nudgeDown;
		}
		else
		{
			double nudge = fabs(ymax - ymin)*0.03;
			ymax += nudge;
			ymin -= nudge;
		}
		return autoFound;
	}
}


//--------------------------

bool TraceItem::getYRange(double& ymin, double& ymax)
{
	// gets y min and max values for all y values in range xmin - xmax
	// call with ymin and ymax values set to current values of y range fields
	if ( QuSlotDebug() ) printf( " QpTraceItem::getYRange \n" );
	double xmin, xmax;
	
	if(frameType() == QP_TS)
	{
		xmin=frame()->tAxisTab->_axis_time_fld->getStart().getEpoch2000();
		xmax = frame()->tAxisTab->_axis_time_fld->getEnd().getEpoch2000();
	}
	else
	{
		xmin=frame()->xAxisTab->_axis_xy_fld->getStart();
		xmax = frame()->xAxisTab->_axis_xy_fld->getEnd();
	}

	bool recalc=false;
		
	int plotType = I_Type;

	if(plotType == QP_CONST_XT ) return false;
	if( _ydo.is_nil() ) return false; // can go no further
	  
	if(plotType == QP_CONST_Y)
	{
		ymax = _ydo->asDouble(0)*1.05;
		ymin = _ydo->asDouble(0)*0.95;
		return true;
		
	}
	
	// check if y object has changed
	int obj_id = _ydo->get_id();

	if(  _y_obj_id_getYRange != obj_id)
	{
		// if yes, store id
		recalc = true;
		_y_obj_id_getYRange = obj_id;
	}
 
	if(_xdo.is_nil()) return false;
	
	obj_id = _xdo->get_id();
	  
	// check if x range has changed since last call
	if(_x_range_min != xmin || _x_range_max != xmax)
	{
		_x_range_min = xmin;
		_x_range_max = xmax;
		recalc=true;
	}

	// if nothing has changed fetch range calculated last time
	if( !recalc )
	{
		ymin = _y_from;
		ymax = _y_to;
		return true;
	}
	
	// refresh y range

	int from = 0;
	int to = _ydo->seqSize()-1;
	
	if(_xdo->is_dbl()) {
		int recL=0;
		int recU=_xdo->seqSize()-1;
		_xdo->getBounds(recL, recU, xmin);	
		from = recL;
		
		recL=0;
		recU=_xdo->seqSize()-1;
		_xdo->getBounds(recL, recU, xmax);
		to = recU;
		
	}
	else if(_xdo->is_time())
	{
		DvEvent range(xmin, xmax);
		_xdo->getEventRecBound(from, to, range);
	}
	
	bool ok = false;

	ok = _ydo->getDataRange(from, to, ymin, ymax);
	
	// store results if min and max found
	if(ok)
	{
		_y_from = ymin;
		_y_to = ymax;
	}
	return ok;
}


//--------------------------

QpuiReturnStatus TraceItem::plot()
{
	if(QuSlotDebug())
	{
		printf("QpTraceItem::plot \n");
		fflush(stdout);
	}
	int plotType = I_Type;
		
	// plots the Trace Object
	PLFLT *x_data=0;
	PLFLT *y_data=0;
	PLFLT *c_data=0;
	
	valarray <double> x_diff;
	int n_data=0;
	
	PLINT line_colour_index = getPGPColourIndex(_line_colour);
	PLINT symbol_colour_index = getPGPColourIndex(_symbol_colour);
	
	// Refresh y data object
	if(_ydo.is_nil())
	{
		plmtex("lv",-10.0,0.5,0.0,"Trace Plot ERROR: Null Y Object");
		return QPUI_FAILURE;
	}
	
	if(!DataTab->_y_data_fld->isValid())
	{
		plmtex("lv",-10.0,0.5,0.0,"Trace Plot ERROR: Invalid Y Slot");
		return QPUI_FAILURE;
	}
	
	// find end of panel for text direction
	PLFLT x_end;
	DvTime t_plot_end;
	DvTime t0;
	
	if(frameType() == QP_TS)
	{
		t0=frame()->tAxisTab->_axis_time_fld->getStart();
		t_plot_end=frame()->tAxisTab->_axis_time_fld->getEnd();

		x_end = (PLFLT)t_plot_end.getDiffSecs(t0);
	}
	else
	{
		x_end = frame()->xAxisTab->_axis_xy_fld->getEnd();
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			x_end = SafeLog10(x_end);
		}
	}

	if(plotType == QP_CONST_XT && frameType() == QP_TS)
	{				
		if(_ydo->is_time() && _ydo->seqSize() > 1)
		{
			DvEvent tint = _ydo->getTimeRange();
			x_data  =new PLFLT[4];
			y_data = new PLFLT[4];
			y_data[0] = y_data[3] = panel()->yAxisTab->_axis_xy_fld->getEnd(); 
			y_data[1] = y_data[2] = panel()->yAxisTab->_axis_xy_fld->getStart();
			if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
			{
				y_data[0] = SafeLog10(y_data[0]);
				y_data[1] = SafeLog10(y_data[1]);
				y_data[2] = SafeLog10(y_data[2]);
				y_data[3] = SafeLog10(y_data[3]);
			}
			n_data = 4;

			x_data[0] = x_data[1] = (PLFLT)tint.start().getDiffSecs(t0);
			x_data[2] = x_data[3] = (PLFLT)tint.end().getDiffSecs(t0);
		}
		else
		{
			DvTime time_val = _ydo->time(0);
		 
			PLFLT t_constant = (PLFLT)time_val.getDiffSecs(t0);
			x_data = new PLFLT[2];
			x_data[0] = x_data[1] = t_constant;
			y_data = new PLFLT[2];
			y_data[0] = panel()->yAxisTab->_axis_xy_fld->getStart();
			y_data[1] = panel()->yAxisTab->_axis_xy_fld->getEnd();
			if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
			{
				y_data[0] = SafeLog10(y_data[0]);
				y_data[1] = SafeLog10(y_data[1]);
			}
			n_data=2;
		}
	}
	else if(plotType == QP_CONST_Y)
	{
		PLFLT y_constant;
		y_constant = (PLFLT)_ydo->asDouble(0);
		
		n_data=2;
		y_data = new PLFLT[2];
		y_data[0] = y_constant;
		y_data[1] = y_constant;
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			y_data[0] = SafeLog10(y_data[0]);
			y_data[1] = SafeLog10(y_data[1]);
		}
		x_data = new PLFLT[2];
		if(frameType() == QP_TS)
		{
			x_data[0] = (PLFLT)0.0;
			x_data[1] = (PLFLT)t_plot_end.getDiffSecs(t0);
		}
		else
		{
			x_data[0] = frame()->xAxisTab->_axis_xy_fld->getStart();
			x_data[1] = frame()->xAxisTab->_axis_xy_fld->getEnd();
			if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
			{
				x_data[0] = SafeLog10(x_data[0]);
				x_data[1] = SafeLog10(x_data[1]);
			}
		}
	}
	else if(plotType == QP_CONST_XT && frameType() == QP_XY)
	{
		PLFLT x_constant;
		x_constant = (PLFLT)_ydo->asDouble(0);
		
		n_data=2;
		x_data = new PLFLT[2];
		x_data[0] = x_data[1] = x_constant;
		y_data = new PLFLT[2];
		y_data[0] = panel()->yAxisTab->_axis_xy_fld->getStart();
		y_data[1] = panel()->yAxisTab->_axis_xy_fld->getEnd();
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			x_data[0] = SafeLog10(x_data[0]);
			x_data[1] = SafeLog10(x_data[1]);
		}
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			y_data[0] = SafeLog10(y_data[0]);
			y_data[1] = SafeLog10(y_data[1]);
		}
	}
	else if(plotType == QP_REG_LINE && frameType() == QP_XY)
	{
		if(linFit_var.is_nil()) linFit_var = _ydo->regression(_xdo);
		
		if(linFit_var.is_nil()) return QPUI_FAILURE;
		
		DvObject_var dep0 = linFit_var->getDep0();

		if(dep0.is_nil()) return QPUI_FAILURE;

		n_data=2;
		x_data = new PLFLT[2];
		x_data[0] = (PLFLT) dep0->asDouble(0);
		x_data[1] = (PLFLT) dep0->asDouble(1);
		y_data = new PLFLT[2];
		y_data[0] = (PLFLT) linFit_var->asDouble(0);
		y_data[1] = (PLFLT) linFit_var->asDouble(1);
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			x_data[0] = SafeLog10(x_data[0]);
			x_data[1] = SafeLog10(x_data[1]);
		}
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			y_data[0] = SafeLog10(y_data[0]);
			y_data[1] = SafeLog10(y_data[1]);
		}
		
		
	}
	else if(plotType == QP_REG_LINE && frameType() == QP_TS)
	{
		if(linFit_var.is_nil()) linFit_var = _ydo->regression(_xdo);
		if(linFit_var.is_nil()) return QPUI_FAILURE;
		
		DvObject_var tt = linFit_var->getTimeTags();
		if(tt.is_nil() || tt->seqSize() < 2) return QPUI_FAILURE;
		 
		n_data=2;
		x_data = new PLFLT[2];
		x_data[0] = (PLFLT) tt->time(0).getDiffSecs(t0);
		x_data[1] = (PLFLT) tt->time(1).getDiffSecs(t0);
			
		y_data = new PLFLT[2];
		y_data[0] = (PLFLT) linFit_var->dblS(0);
		y_data[1] = (PLFLT) linFit_var->dblS(1);
		
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			y_data[0] = SafeLog10(y_data[0]);
			y_data[1] = SafeLog10(y_data[1]);
		}
	}
	else if(plotType == QP_TRACE && frameType() == QP_TS)
	{
		if(_xdo->not_time() )
		{
			plmtex("lv",-10.0,0.5,0.0,"Trace Plot ERROR: Null T Object");
			return QPUI_FAILURE;
		}
		// fetch number of elements in each sequence/array
		n_data = _ydo->seqSize();
		if(n_data == 1) n_data = _ydo->arraySize();
		
		int nt_data = _xdo->seqSize();

		if(nt_data != n_data)
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Y and T series of different lengths");
			return QPUI_FAILURE;
		}
		
		// create float c-style arrays for plotting
		x_data = new PLFLT[n_data];
		y_data = new PLFLT[n_data];
		x_diff.resize(n_data);
		
		// get y object data
		if(!QuSet_Array(n_data, &(y_data[0]), _ydo, t0))
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Y Object not a Scalar, ScalarSeq or 1D matrix");
			return QPUI_FAILURE;
		}
		
		// loop through data and set x and y arrays
		double x_diff_sum = 0.;
		for(int i=0; i<n_data; i++)
		{
			// set time in number of seconds from t_midnite
			double s_i = _xdo->time(i).getDiffSecs(t0);
			if(i==0) x_diff[0] = s_i;
			else x_diff[i] = s_i - x_diff_sum;
			x_diff_sum += x_diff[i];
			x_data[i] = (PLFLT)s_i;
		}
		
		// calculate logarithmic values if logs selected
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) 
		{
			for(int i=0; i<n_data; i++) y_data[i] = (PLFLT) SafeLog10((double)y_data[i]);
		}
	}
	else if(plotType == QP_TRACE && frameType() == QP_XY)
	{
		// Data Series (XY trace)

		// fetch number of elements in each sequence/array
		n_data = _ydo->seqSize();
		if(n_data == 1) n_data = _ydo->arraySize();

		int nx_data = _xdo->seqSize();
		if(nx_data == 1) nx_data = _xdo->arraySize();

		if(nx_data != n_data)
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Y and X series of different lengths");
			return QPUI_FAILURE;
		}

		// create float c-style arrays for plotting
		x_data = new PLFLT[n_data];
		y_data = new PLFLT[n_data];

		// get y data
		if(!QuSet_Array(n_data, &(y_data[0]), _ydo))
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Y Object not a Scalar, ScalarSeq or 1D matrix");
			return QPUI_FAILURE;
		}
		
		// get x data
		if(!QuSet_Array(n_data, &(x_data[0]), _xdo))
		{
			plmtex("lv",-10.0,0.5,0.0,"ERROR: X Object not an Scalar, ScalarSeq or 1D Matrix");
			return QPUI_FAILURE;
		}

		// loop through data and set x logs 
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) 
		{
			for(int i=0; i<n_data; i++) x_data[i] = SafeLog10(x_data[i]);
		
		}
				
		// loop through data and set y logs 
		if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) 
		{
			for(int i=0; i<n_data; i++) y_data[i] = SafeLog10(y_data[i]);
		
		}

	} // End if else over trace types

	// Set colour data if available
	if( !_cdo.is_nil() ){

		c_data = new PLFLT[n_data];
		TraceTab->colourTable->getCurrentTable().apply();

		
		bool OKmin, OKmax;
		PLFLT min = (PLFLT) TraceTab->_c_min_fld->text().toFloat(&OKmin);
		PLFLT max = (PLFLT) TraceTab->_c_max_fld->text().toFloat(&OKmax);
		
		PLFLT range = max - min;
		
		if( _cdo->seqSize() != _ydo->seqSize() ){
			plmtex("lv",-10.0,0.5,0.0,"ERROR: Colour Object not compatible length");
			for(int i=0; i<n_data; i++){
				c_data[i] = 0.5; 
			}
		}
		else{
						
			if(!OKmin || !OKmax || range <= 0.){
				// find range from data as fallback
				min = 1.e10;
				max = -1.e10;
				for(int i=0; i<n_data; i++){
					if(min > _cdo->asDouble(i)) min = _cdo->asDouble(i);
					if(max < _cdo->asDouble(i)) max = _cdo->asDouble(i);
				}
				range = max - min;
				if (range <= 0.) {
					plmtex("lv",-10.0,0.5,0.0,"ERROR: Colour Object range negative");
					for(int i=0; i<n_data; i++){
						c_data[i] = 0.5; 
					}
				}
			}

			if(TraceTab->c_range_log->isChecked()){
				min = SafeLog10(min);
				max = SafeLog10(max);
				range = max - min;
				for(int i=0; i<n_data; i++){
					c_data[i] = (SafeLog10(_cdo->asDouble(i)) - min) / range;
					// out of range is supposed to be safe, but is not always
					if(c_data[i] < 0.) c_data[i] = 0.;
					if(c_data[i] > 1.) c_data[i] = 1.;
				}
			}
			else{
				for(int i=0; i<n_data; i++){
					c_data[i] = (_cdo->asDouble(i) - min) / range;
					// out of range is supposed to be safe, but is not always
					if(c_data[i] < 0.) c_data[i] = 0.;
					if(c_data[i] > 1.) c_data[i] = 1.;
				}
			}
		}
		// set colour bar
		if(TraceTab->_show_colour_bar->isChecked())
		{
			PLFLT x1win_save, x2win_save, y1win_save, y2win_save;
			PLFLT x1vp_save, x2vp_save, y1vp_save, y2vp_save;
			PLFLT x[4], y[4];

			plgvpw(&x1win_save, &x2win_save, &y1win_save, &y2win_save);
			plgvpd(&x1vp_save, &x2vp_save, &y1vp_save, &y2vp_save);
	
			PLFLT chsize_ndc;
		
			chsize_ndc=getCharHeight();
	
			x[0] = x2vp_save + TraceTab->_colour_bar_disp->value() * chsize_ndc*1.5;
			x[2] = x[0] +  TraceTab->_colour_bar_width->value() * chsize_ndc;
			float ht = (y2vp_save - y1vp_save) ;
			y[0] = y1vp_save + ht * TraceTab->_colour_bar_voffset->value();
			y[1] = y[0] + ht * TraceTab->_colour_bar_height->value();
	
			plvpor(x[0],x[2], y[0], y[1]);
		
			plwind(0., 1., min, max);
	
			x[0]=0.;
			x[1]=0.;
			x[2]=1.;
			x[3]=1.;

			PLFLT zstep=(max-min)/200.;
			PLFLT i=0.;
			for(PLFLT col=min;col<max;col+=zstep, i+=1./200.)
			{
				y[0] = col;
				y[3] = col;
				y[1] = col+zstep;
				y[2] = y[1];
				if(i>=1.) i=0.999;
				plcol1(i);
				plfill(4, x, y);
			}
	
			plcol0(1);
			page()->setSmooth(true);
			string yopt = "bcmtsv";
			if(TraceTab->c_range_log->isChecked()) yopt += "l";
			plschr(0., frame()->DisplayTab->_ticks_char_size_fld->value());
		
			plwid(QtWidth2PlWidth(frame()->DisplayTab->_line_width_fld->value()));

			PLFLT _ticks_length = TraceTab->_colourscale_ticklength->value();
		
			plsmaj(0., _ticks_length);
			plsmin(0., _ticks_length*0.6);

			PLLabelDefaults y_label_opt;
			y_label_opt.exp_label_disp = 2.0 + (PLFLT) TraceTab->_exp_label_offset_y->value();
			y_label_opt.exp_label_pos = 1.02 +(PLFLT) TraceTab->_exp_label_offset_x->value();
			y_label_opt.exp_label_just = 0.0;
			plslabelfunc(NULL, &y_label_opt);
			plbox("bc", 0.0, 0, yopt.c_str(), 0.0, 0);
			plslabelfunc(NULL, NULL);
        
			string label="";
	
			plwid(1);
		
			label = TraceTab->colLabSlot->text().toLatin1().data();
		
			Q_plmtex("l", 1., 0.5, 0.5, SafePGPString(label).c_str());
			plvpor(x1vp_save, x2vp_save, y1vp_save, y2vp_save);
			plwind(x1win_save, x2win_save, y1win_save, y2win_save);
			page()->setSmooth(false);
		}

	}

	// set up plot symbol and size outside of data loop
	PLINT index=symbolIndexConversion(TraceTab->_symbol_shape_fld->currentIndex());
	const char* ch=symbolUnicodeConversion(TraceTab->_symbol_shape_fld->currentIndex());
	PLFLT chSz = Q_plsize(index, ch, TraceTab->_symbol_size_fld->value());
	int i=0;
	while (i < n_data)
	{
		int j=i;
		i++;
		if(plotType == QP_TRACE && frameType() == QP_TS)
		{
			// break into sub-sequences on gaps
            if(DataTab->gap_threshold>0)
            {
                while(i < n_data && (double)x_diff[i] <= DataTab->gap_threshold && x_diff[i] > 0. )
			    {
				    i++;  // increment i
			    }
            }
            else
            {
                i = n_data;
            }
		}
		else if( plotType == QP_CONST_XT && frameType() == QP_TS)
		{
			i++;
		}
		else i = n_data; // most trace types ignore gaps
	
		plwid(1); // force reset
		// calculate no. points in this sub-sequence with no gaps
		int n_pts = i-j;

        if(TraceTab->_line_style_fld->currentIndex() != QP_NOLINE)
		{
			plwid(QtWidth2PlWidth(TraceTab->_line_width_fld->value()));
			setDashStyle((QpLineStyle)TraceTab->_line_style_fld->currentIndex(), TraceTab->_line_scale_fld->value());
			plcol0(line_colour_index);
			// draw line between points of sub-sequence of appropriate time
			switch(TraceTab->_connect_style_fld->currentIndex())
			{
				case QPUI_STRAIGHT: default:
					plline(n_pts, x_data+j, y_data+j);
					break;
				case QPUI_HIST:
                    if(x_end<0){
                        plotHist(n_pts, x_data+j, y_data+j);
                        
                    }
                    else{
                        plbin(n_pts, x_data+j, y_data+j, PL_BIN_CENTRED|PL_BIN_NOEXPAND);
                    }
					break;
				case QPUI_NOCONNECT:
					break;
			}
			setDashStyle(QP_SOLID, 1.0);

		}

        if(TraceTab->_symbol_shape_fld->currentIndex() == QPUI_NOSYMBOL && 
                ( n_pts==1 || fabs((x_data+j)[1] - (x_data+j)[0]) < 1e-20) )
		{
			// ensure single point is visible!
			plssym(0., TraceTab->_line_width_fld->value());
			plcol0(line_colour_index);
			// plot symbols in sub-sequence
			PLFLT size = Q_plsize((PLINT)QPUI_POINT, Q_UNICODE_POINT, 2.0*TraceTab->_line_width_fld->value());
			Q_plpoin(1, x_data+j, y_data+j, (PLINT)QPUI_POINT, Q_UNICODE_POINT, size);
		}
		
		if(TraceTab->_symbol_shape_fld->currentIndex() != QPUI_NOSYMBOL)
		{
			
//			if(index==QPUI_CUSTOM_SHAPE)
//			{
//				plschr(0., TraceTab->_symbol_size_fld->value());
//				plcol0(symbol_colour_index);
				
//				for(int i=0; i<n_pts; ++i)
//				{
//					if( !(TraceTab->_symbol_colour_fld->isEnabled())  && TraceTab->_colour_on->isChecked() && c_data) plcol1(c_data[j+i]);
//					 if(TraceTab->_size_on->isChecked() && c_data) 
//					 	plschr(0., TraceTab->_symbol_size_fld->value()*2.*c_data[j+i]);

//					plptex(x_data[j+i], y_data[j+i], x_end, 0., 0.5, TraceTab->_custom_symbol_fld->text().toLatin1().data());
//				}
//			}
//			else
//			{

				plcol0(symbol_colour_index);
				// plot symbols in sub-sequence
				for(int i=0; i<n_pts; ++i)
				{
                    if( !(TraceTab->_symbol_colour_fld->isEnabled()) && TraceTab->_colour_on->isChecked() && c_data ) plcol1(c_data[j+i]);
					if(TraceTab->_size_on->isChecked() && c_data ) {
                        chSz *= 2.*c_data[j+i];
//                        plssym(0., chSz); // not used by Q_plpoin now, uses plstring()
                    }
					if(index==QPUI_CUSTOM_SHAPE){
						Q_plpoin(1, x_data+j+i, y_data+j+i, index, TraceTab->_custom_symbol_fld->text().toLatin1().data(), chSz);
					}
					else {
						Q_plpoin(1, x_data+j+i, y_data+j+i, index, ch, chSz);
					}
				}
//			}
		}
	} // repeat until all data exhausted..

	plcol0(0);
	delete x_data;
	delete y_data;
	if(c_data != 0) delete c_data;
	return QPUI_SUCCESS;
}


void TraceItem::plotHist(int n_pts, PLFLT *x_data, PLFLT *y_data ){
    
    if (n_pts < 2) return;
    
    // used for inverted time line
    PLFLT y_pts[2];
    PLFLT x_pts[2];
    for(int i=0; i<n_pts-1; i++){
        // horizontal bar centred on x
        y_pts[0] = y_data[i];
        y_pts[1] = y_data[i];
        PLFLT diff = x_data[i+1] - x_data[i];
        diff /= 2.;
        x_pts[0] = x_data[i] - diff;
        x_pts[1] = x_data[i] + diff;
        plline(2, &(x_pts[0]), &(y_pts[0]));
        
        // verticals
        y_pts[0] = y_data[i];
        y_pts[0] = 0;
        
        x_pts[0] = x_data[i] - diff;
        x_pts[1] = x_data[i] - diff;
        plline(2, &(x_pts[0]), &(y_pts[0]));
        
        x_pts[0] = x_data[i] + diff;
        x_pts[1] = x_data[i] + diff;
        plline(2, &(x_pts[0]), &(y_pts[0]));
        

    }
    // last point
    // horizontal
    y_pts[0] = y_data[n_pts-1];
    y_pts[1] = y_data[n_pts-1];
    PLFLT diff = x_data[n_pts-1] - x_data[n_pts-2];
    diff /= 2.;
    x_pts[0] = x_data[n_pts-1] - diff;
    x_pts[1] = x_data[n_pts-1] + diff;
    plline(2, &(x_pts[0]), &(y_pts[0]));
    
    // verticals
    y_pts[0] = y_data[n_pts-1];
    y_pts[0] = 0;
    
    x_pts[0] = x_data[n_pts-1] - diff;
    x_pts[1] = x_data[n_pts-1] - diff;
    plline(2, &(x_pts[0]), &(y_pts[0]));
    
    x_pts[0] = x_data[n_pts-1] + diff;
    x_pts[1] = x_data[n_pts-1] + diff;
    plline(2, &(x_pts[0]), &(y_pts[0]));
    
}

QpuiReturnStatus TraceItem::plotLegend()
{
    if( !(DataTab->_show_legend_fld->isChecked()) || !(panel()->LegendTab->_show_legend_fld->isChecked()) ) return QPUI_SUCCESS;
 
   
	PLINT line_colour_index = getPGPColourIndex(_line_colour);
	PLINT symbol_colour_index = getPGPColourIndex(_symbol_colour);
    
	PLFLT x0win_save, x1win_save, y0win_save, y1win_save;
	PLFLT x0vp_save, x1vp_save, y0vp_save, y1vp_save;
    
	plgvpw(&x0win_save, &x1win_save, &y0win_save, &y1win_save);
	plgvpd(&x0vp_save, &x1vp_save, &y0vp_save, &y1vp_save);
    
//     QPlot_window->save();
    
	int row_num, col_num, num_cols, num_rows;
    
    row_num = panel()->_legend_entry / panel()->LegendTab->_legend_num_cols_fld->value();
    col_num = panel()->_legend_entry - row_num * panel()->LegendTab->_legend_num_cols_fld->value();
    
//     PLFLT xch, ych; 
    plschr(0., panel()->LegendTab->_legend_char_size_fld->value());
//     plgchr(&xch, &ych);
    
    float legend_width, legend_height;
    legend_width = panel()->LegendTab->_legend_width_fld->value() * (x1vp_save - x0vp_save);
    legend_height = panel()->LegendTab->_legend_height_fld->value() * (y1vp_save - y0vp_save);

    num_cols = panel()->LegendTab->_legend_num_cols_fld->value();
    num_rows = panel()->LegendTab->_legend_num_rows_fld->value();
    float _legend_cell_width, _legend_cell_height;
    _legend_cell_width = legend_width / num_cols;
    _legend_cell_height = legend_height / num_rows;
    
    float x0_vp, x1_vp, y0_vp, y1_vp;
    x0_vp = x0vp_save + panel()->LegendTab->_legend_loffset_fld->value() * (x1vp_save - x0vp_save);
    x0_vp += col_num * _legend_cell_width ;
    x1_vp = x0_vp + _legend_cell_width;
    y1_vp = y1vp_save - panel()->LegendTab->_legend_toffset_fld->value() * (y1vp_save - y0vp_save);
    y1_vp -= row_num * _legend_cell_height;
    y0_vp = y1_vp - _legend_cell_height -0.001; 

    plvpor(x0_vp, x1_vp, y0_vp, y1_vp);
    plwind(0.,1.,0.,1.);
	plscol0(0, 255, 255, 255);
	plscol0(1, 0, 0, 0);
    plcol0(0);
    if((QpuiLegendBackground)panel()->LegendTab->_legend_background_fld->currentIndex() == QPUI_OPAQUE)
    {
		plwid(1);
	    PLFLT x_fill[]={0., 0., 1., 1.};
	    PLFLT y_fill[]={0., 1., 1., 0.};
	    plfill(4, x_fill, y_fill);

    }
    plcol0(1);
    
    float len, marg;
    len = panel()->LegendTab->_legend_line_length_fld->value();
    marg = 0.1*len;
    
	PLFLT x_lgd[2];
	x_lgd[0] = marg;
	x_lgd[1] = marg + len;
	PLFLT y_lgd[2] = {0.,0.};
	float txt_lgd = 0.;
	float ych_cell_height = getCharHeight();/*ych;*/
	if( _legend_cell_height > 1e-4) ych_cell_height /=  _legend_cell_height;
	if( ych_cell_height > 1.) ych_cell_height = 1.;
    
    switch((QpuiLegendJust)panel()->LegendTab->_legend_just_fld->currentIndex())
    {
        case QPUI_VCENTRE: default:
            txt_lgd = 0.5;
            break;
        case QPUI_TOP:
            txt_lgd = 1. - ych_cell_height*1.2;
            break;
        case QPUI_BOTTOM:
		txt_lgd = 0.00 + ych_cell_height*1.2;
            break;
    }

    
    y_lgd[1] = txt_lgd;
    y_lgd[0] = y_lgd[1];
	   
	plschr(0., panel()->LegendTab->_legend_char_size_fld->value());
	setDashStyle((QpLineStyle)TraceTab->_line_style_fld->currentIndex(), TraceTab->_line_scale_fld->value());
	if(TraceTab->_line_style_fld->currentIndex() != QP_NOLINE)
	{
		plwid(QtWidth2PlWidth(TraceTab->_line_width_fld->value()));
		plcol0(line_colour_index);
    }
	else
	{
		plwid(1);
		plcol0(0);
	}
    // for some reason (plp bug?) the line must be plotted for the symbol and text to show!
	plline(2, x_lgd, y_lgd);
    
    if(TraceTab->_symbol_shape_fld->currentIndex() != QPUI_NOSYMBOL)
    {
        PLINT index=symbolIndexConversion(TraceTab->_symbol_shape_fld->currentIndex());
        const char * ch=symbolUnicodeConversion(TraceTab->_symbol_shape_fld->currentIndex());
		PLFLT chSize = Q_plsize(index, ch, TraceTab->_symbol_size_fld->value());
		if(index==QPUI_CUSTOM_SHAPE){
			ch = TraceTab->_custom_symbol_fld->text().toLatin1().data();
		}
//			if(index==QPUI_CUSTOM_SHAPE)
//			{
//				plschr(0, TraceTab->_symbol_size_fld->value());
//				plcol0(symbol_colour_index);
//				PLFLT x=0.5*(x_lgd[0]+x_lgd[1]);
//				PLFLT y=y_lgd[0];
//				plptex(x, y, 1., 0., 0.5, TraceTab->_custom_symbol_fld->text().toLatin1().data());
//			}
//			else
//			{
//				plssym(0., TraceTab->_symbol_size_fld->value()); // not needed for point now (using plstring())
				plcol0(symbol_colour_index);
				PLFLT x=0.5*(x_lgd[0]+x_lgd[1]);
				PLFLT y=y_lgd[0];
				Q_plpoin(1, &x, &y, index, ch, chSize);
//			}

    }
    plcol0(1);
    
    int _text_line_width = 1;

	plschr(0., panel()->LegendTab->_legend_char_size_fld->value());
    PLINT family, style, weight;
#if QT_VERSION >= 0x040300
    panel()->LegendTab->_legend_font_fld->getFontParams(family, style, weight);
    plsfont(family, style, weight);
#else
    family=panel()->LegendTab->_legend_font_fld->getFontIndex();
    plfont(family);
#endif

	
	plwid(QtWidth2PlWidth(_text_line_width));
	string legend_text = DataTab->_legend_text_fld->text().toLatin1().data();

    legend_text = SafePGPString(legend_text);

    QString greek = QString(legend_text.c_str());
    greek.replace("theta","#gh");
    greek.replace("phi","#gf");
    legend_text = greek.toLatin1().data();
    
    if(legend_text.size() > panel()->LegendTab->_legend_max_char_length_fld->value())
	    legend_text = legend_text.substr(0, panel()->LegendTab->_legend_max_char_length_fld->value());

    plptex(x_lgd[1]+marg, txt_lgd, 1., 0., 0., legend_text.c_str());
    panel()->_legend_entry++;
    plscol0(0,255,255,255);
    plscol0(1,0,0,0);
    plcol0(1);

//     QPlot_window->unsave();
	plvpor(x0vp_save, x1vp_save, y0vp_save, y1vp_save);
	plwind(x0win_save, x1win_save, y0win_save, y1win_save);
    
	return QPUI_SUCCESS;
}

QpItemType QpRemapPlotType(int _type, QpFrameType &frameType){

	switch(_type){
		case 0:
			frameType = QP_XY;
			return QP_TRACE;
			break;
		
		case 1: default:
			frameType = QP_TS;
			return QP_TRACE;
			break;
		
		case 2:
			frameType = QP_TS;
			return QP_SPECTROGRAM;
			break;

		case 3:
			frameType = QP_XY;
			return QP_SPECTROGRAM;
			break;

		case 4:
			frameType = QP_TS;
			return QP_CONST_XT;
			break;

		case 5:
			frameType = QP_XY;
			return QP_CONST_Y;
			break;
		
		case 6:
			frameType = QP_XY;
			return QP_CONST_XT;
			break;
		
		case 7:
			frameType = QP_XY;
			return QP_TRACE;
			break;
		
		case 8:
			frameType = QP_TS;
			return QP_TRACE;
			break;
		
		case 9:
		// unknown type flag
			return QP_TRACE;
			break;
		
		case 10:
			frameType = QP_XY;
			return QP_EVENT;
			break;
		
		case 11:
			frameType = QP_TS;
			return QP_EVENT;
			break;
   
	}
   
}

// ----------------------------------

QDomElement TraceItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Plot_Item");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));

    switch(I_Type)
    {
        case QP_CONST_Y:
            xml.setAttribute("Type", "Const_Y");
            break;
        case QP_CONST_XT:
            xml.setAttribute("Type", "Const_XT");
            break;
        case QP_REG_LINE:
            xml.setAttribute("Type", "Reg_Line");
            break;
        case QP_TRACE: default:
            xml.setAttribute("Type", "Trace");
            break;
    }

    xml.appendChild(DataTab->toXMLElement(doc, I_Type));
    
    QString linkText=getLinkText();
    if(!linkText.isEmpty())
    {
        linkText += "S"; // to avoid conflict with existing links on restore
    }
    xml.appendChild(createTextXMLNode(doc, "Link_Text", linkText));

    xml.appendChild(TraceTab->toXMLElement(doc));

    xml.appendChild(createTextXMLNode(doc, "User_Line_Colour", userLineColour));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Red", _line_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Green", _line_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Blue", _line_colour.blue()));

    xml.appendChild(createTextXMLNode(doc, "User_Symbol_Colour", userSymbolColour));
    xml.appendChild(createTextXMLNode(doc, "Symbol_Colour_Red", _symbol_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Symbol_Colour_Green", _symbol_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Symbol_Colour_Blue", _symbol_colour.blue()));
    

    return xml;
}

void TraceItem::fromXMLElement(QDomElement& el)
{
    QString s;
	
 	if(QuSlotDebug()) cout <<  "TraceItem restore from XML (>>)" << endl;
  
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }

    QDomElement dataElement=el.firstChildElement("Data_Tab");
    if(!dataElement.isNull())
    {
        DataTab->fromXMLElement(dataElement, I_Type);
    }

    QDomElement traceElement=el.firstChildElement("Trace_Tab");
    if(!traceElement.isNull())
    {
        TraceTab->fromXMLElement(traceElement);
    }
  
    if(getXMLNodeText(el, "User_Line_Colour", s))
    {
        userLineColour=s.toInt();
    }
    
    if(getXMLNodeText(el, "Link_Text", s))
    {
        setLinkText(s);
    }
   

    int r(0),g(0),b(0);
    if(getXMLNodeText(el, "Line_Colour_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Line_Colour_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Line_Colour_Blue", s))
    {
        b=s.toInt();
    }

    _line_colour=QColor(r,g,b);
    QPalette line_palette = QPalette(QPUI_normal_palette);
    line_palette.setColor(QPalette::Active ,QPalette::ButtonText, _line_colour);
    line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _line_colour);
    TraceTab->_line_colour_fld->setPalette(line_palette);

    if(getXMLNodeText(el, "User_Symbol_Colour", s))
    {
        userSymbolColour=s.toInt();
    }
	if(getXMLNodeText(el, "Symbol_Colour_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Symbol_Colour_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Symbol_Colour_Blue", s))
    {
        b=s.toInt();
    }
   
    _symbol_colour=QColor(r,g,b);
    line_palette.setColor(QPalette::Active ,QPalette::ButtonText, _symbol_colour);
    line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _symbol_colour);
    TraceTab->_symbol_colour_fld->setPalette(line_palette);


	DataTab->updateGapThreshold();

}

PLFLT Q_plsize(PLINT code, const char *ch, PLFLT size){

	if(ch == Q_UNICODE_NOSYMBOL) return size;
	
    if(ch == Q_UNICODE_POINT && code == QPUI_POINT){ // point
        size *= 0.2;
    }
    else if(ch == Q_UNICODE_CUSTOM_SHAPE ){
        size *= 0.6;
    }
    else if(ch == Q_UNICODE_STAR || ch == Q_UNICODE_FILLEDSTAR){
        size *= 0.6;
    }
    else if(ch == Q_UNICODE_PLUS || ch == Q_UNICODE_ASTERISK || ch == Q_UNICODE_CROSS || ch == Q_UNICODE_CROSS ){
        size *= 1.;
    }
    else if(ch == Q_UNICODE_TRIANGLE || ch == Q_UNICODE_CIRCLED_POINT || ch == Q_UNICODE_WIDECROSS){
        size *= 0.6;
    }
    else{
		size *= 0.8; // would include Q_UNICODE_POINT as filled circle, but code != QPUI_POINT
	}
	
	return size;
}

void Q_plpoin(PLINT n, PLFLT *x, PLFLT *y, PLINT code, const char *ch, PLFLT size){

/* interface to plpoin to only plot points that lay within viewport
*  This speeds up plotting for some drivers that apparently render
*  slowly, and so can spend a lot of time rendering points that get
*  clipped by the viewport anyway. Given the overheads in plpoin to
*  look up symbols and fonts, reverting to calling plpoin point by point
*  is probably at least as fast if not faster even in cases where n is 
*  quite small.
*
*  Steve Schwartz; June 2010
*  Tony Allen; Oct 2014
*/
    PLFLT chrHt, defHt;
    plgchr(&defHt, &chrHt);
    
	PLFLT xmin, xmax, ymin, ymax, swap, xx, yy;
	int i;

	plgvpw(&xmin, &xmax, &ymin, &ymax);

/* detect if either axis runs backward and swap min & max values */

	if (xmax < xmin){swap = xmax; xmax = xmin; xmin = swap;}
	if (ymax < ymin){swap = ymax; ymax = ymin; ymin = swap;}

    plschr(0., size);

	// plssym(0., TraceTab->_symbol_size_fld->value());
	for (i=0; i<n; i++){
		xx = x[i]; yy = y[i];
		if(xx > xmin && xx < xmax && yy> ymin && yy < ymax){
            //	plpoin(1,&xx,&yy,code);
            //	plsym(1,&xx,&yy,code);
			 plstring(1,&xx,&yy,ch);
		}
	}
    // reset char size
   plschr(0., chrHt);
    
	return;
}

// our version of plmtex which allows for various syntax schemes

void Q_plmtex(const char * side, PLFLT disp, PLFLT pos, PLFLT just, const char * text){

	DvString newText = text;
	
	// 'Granny' strings
	int i=0;
	bool isUp=false;
	bool isDown=false;
	while(i<newText.size()){
		if(newText[i] == '!'){
			i++;
			if(newText[i] == 'E'){
				newText = newText.replace("!E", "#u");
				isUp=true;
				isDown=false;
			}
			else if(newText[i] == 'D'){
				newText = newText.replace("!D", "#d");
				isDown=true;
				isUp=false;
			}
			else if(newText[i] == 'N'){
				if(isUp) newText = newText.replace("!N", "#d");
				else if(isDown) newText = newText.replace("!N", "#u");
				isUp=false;
				isDown=false;
			}
		}
		i++;
	}
	newText = newText.replaceAll("!!", "!");
	
	
	plmtex(side, disp, pos, just, newText.c_str());
	return;
	
}

